-- control indexes
-- color scheme
-- dialog special codes:
--  inner states & exit states

include graphics.e
include screen.e

-----------------------------------------------------------------------------
-- color schemes
global sequence
    OF_TEXT,            -- for text in the MLE
    OF_SELECTED,        -- any selected text
    OF_BORDER,          -- MLE border
    OF_MENUBAR,         -- menu bar
    OF_STATUS,          -- bottom line of screen
    OF_ACCEL,           -- accelerator (hot) keys
    OF_DISABLED,        -- disabled stuff (only use foreground color)
    OF_DIALOG,          -- dialog box
    OF_DIALOG_TITLE,    -- title bar on dialog
    OF_DIALOG_BUTTON,   -- control buttons
    OF_HOTKEY,          -- hot keys
    OF_SCROLL           -- scroll bars on dialogs

-- default settings
    OF_TEXT = { WHITE, BLUE }
    OF_SELECTED = { BRIGHT_WHITE, BLACK }
    OF_BORDER = { WHITE, BLACK }
    OF_MENUBAR = { BLACK, WHITE }
    OF_STATUS = { YELLOW, CYAN }
    OF_ACCEL = { BRIGHT_WHITE, WHITE }
    OF_DISABLED = { GRAY, WHITE }
    OF_DIALOG = { BLACK, WHITE }
    OF_DIALOG_TITLE = { BLACK, BRIGHT_WHITE }
    OF_DIALOG_BUTTON = { BLACK, BRIGHT_WHITE }
    OF_HOTKEY = { RED, BRIGHT_WHITE }
    OF_SCROLL = { BRIGHT_WHITE, BLACK }

-----------------------------------------------------------------------------
-- controls

global constant        

    TYPE            = 1,    -- type of control
    TITLE           = 2,    -- title of text
    HOT             = 3,    -- position of hot key
    ROW             = 4,    -- upper left row on screen
    COL             = 5,    -- upper left row on screen
    DIM             = 6,    -- dimmed?
    DATA            = 7,    -- data in the control
    ITEMS           = 8,    -- LIST items
    HELP_TEXT       = 9,    -- help message text
    WIDE            = 10,   -- how wide the control is
    FOCUS           = 11,   -- true if BUTTON has focus
    CHECKED         = 12,   -- CHECK/RADIO is on
    FIRST_CHAR      = 13,   -- first byte of SLE text displayed
    ITEM            = 14,   -- selected item on LIST or MLE
    CURSOR          = 15,   -- SLE cursor position relative to display
    DEFAULT_BUTTON  = 16,   -- default button
    ANCHOR_COL      = 17,   -- SLE/MLE text anchor column
    ANCHOR_LINE     = 18,   -- MLE text anchor line
    CANCEL_BUTTON   = 19,   -- cancel button
    DEEP            = 20,   -- LIST count of items to display
    REDRAW          = 21,   -- true if title needs to be redrawn
    PERSIST         = 22,   -- true if list highlight stays
    EXIT_STATE      = 23,   -- exit state leaving dialog
    TOP_LINE        = 24,   -- top line of the LIST or MLE
    ELEVATOR        = 25,   -- elevator for scroll bar
    KEY_PRESSED     = 26,   -- key pressed while using dialog
    CLEAN_SCREEN    = 27,   -- background the dialog covered
    DIRTY           = 28,   -- text has been edited
    READ_ONLY       = 29,   -- text cannot be edited
    CURSOR_POS      = 30,   -- cursor position of overlaid text
    COLOR_SYNTAX    = 31,   -- color text
    AUTOCOMPLETE    = 32,    -- true if autocomplete option is selected
    GROUP           = 33,   -- radio button control group
    IS_FOCUS        = 34,   -- true if control is focused on
    MASK            = 35    -- contains edit mask for SLE
    
constant ATTRIBUTE_COUNT = 35

----------------------------------------------------------------------------
-- types of controls
global constant
    MLE         = 1,    -- multiple line edit
    SLE         = 2,    -- single line edit
    LIST        = 3,    -- scrolling list
    BUTTON      = 4,    -- push button
    CHECK       = 5,    -- check box
    RADIO       = 6,    -- radio button
    LABEL       = 7,    -- text label
    MENU        = 8,    -- menu
    DROPDOWN    = 9,    -- menu dropdown
    DIALOG      = 10    -- dialog

-----------------------------------------------------------------------------
-- special key codes
global constant
            GET_FOCUS   = -1,   -- ask focused control to display
            LOSE_FOCUS  = -2,   -- move focus from control
            MOVE_LEFT   = -3,   -- move to next control on left
            MOVE_RIGHT  = -4,   -- move to next control on right
            MOUSE       = -5,   -- catch-all mouse event
            REFRESH     = -6    -- redraw control

-----------------------------------------------------------------------------
-- exit states
global constant
    CANCELLED   = -1,   -- dialog was cancelled
    OKAYED      = 0,    -- dialog was completed
    MORE        = 1     -- more action pending     


-----------------------------------------------------------------------------
-- search/replace
global sequence
    searchFor,          -- for MLE search/replace
    replaceWith,        -- for MLE search/replace
    searchStartedAt     -- for MLE search/replace

global integer
    searchCount,        -- number of finds/replacements
    searchCase,         -- upper/lower case matters
    searchWhole         -- only match on whole word

-- intial values
    searchFor = ""
    replaceWith = ""
    searchCount = 0
    searchCase = 0
    searchWhole = 0

-----------------------------------------------------------------------------
-- close box position
-- negative places is relative to the right side
-- positive relative to the left side
-- zero hides it

global integer CLOSE_BOX_COL
    CLOSE_BOX_COL = 0


-----------------------------------------------------------------------------
global function reset_control( sequence control )

    -- set control to initialized values

    -- alias control

    -- global
        control[REDRAW] = 1     -- redraw the control

    -- default values
    -- case control[TYPE]

    if control[TYPE] =  MLE then 
        control[FIRST_CHAR] = 1 -- first column of MLE
        control[TOP_LINE] = 1   -- first (only) line displayed
        control[CURSOR] = 1     -- cursor position relative to display
        control[ITEM] = 1       -- line 1 is selected
        control[TOP_LINE] = 1   -- first line in list
        control[ELEVATOR] = 0   -- thumb position
        control[DATA] = ""      -- but the data goes here
        control[ITEMS] = { "" } -- empty

    elsif control[TYPE] =  SLE then 
        control[FIRST_CHAR] = 1 -- first byte of SLE text displayed
        control[TOP_LINE] = 1   -- first (only) line displayed
        control[CURSOR] = 1     -- cursor position relative to display
        control[ITEM] = 1       -- line 1 is selected
        control[DEEP] = 1       -- in essence, a 1 line MLE
        control[DATA] = ""      -- but the data goes here

    elsif control[TYPE] =  LIST then 
        control[ITEMS] = {}     -- empty list
        control[TOP_LINE] = 1   -- top item on list
        control[ITEM] = 1       -- selected item on LIST
        control[ELEVATOR] = 0   -- remove elevator item

    elsif control[TYPE] =  BUTTON then 
        -- nothing special

    elsif control[TYPE] =  CHECK then 
        -- nothing special

    elsif control[TYPE] =  RADIO then 
        -- nothing special

    elsif control[TYPE] =  MENU then 
        --nothing special

    elsif control[TYPE] =  DROPDOWN then 
        -- nothing special

    end if

    return control

end function


-----------------------------------------------------------------------------
global function new_control( integer type_of )
    
    -- create a new control of a given type

    sequence control

    -- alias control

    control   = repeat( 0, ATTRIBUTE_COUNT )  -- allot values
    control[TYPE]   = type_of                       -- set type
    control[DATA]   = ""                            -- clear data
    control[REDRAW] = 1                             -- first time being drawn

    -- reset it
    return reset_control( control )

end function


