include screen.e   -- screen i/o
include control.e   -- constants
include xle.e       -- mle/sle
include list.e      -- list
include button.e    -- command buttons
include check.e     -- check boxes
include radio.e     -- radio buttons
include label.e     -- labels

global sequence MAIN_MENU
global sequence MAIN_DIALOG
global sequence DIALOG_STACK

MAIN_MENU = {}
MAIN_DIALOG = {}
DIALOG_STACK = {}

-----------------------------------------------------------------------------
global function focus()
    -- returns index of control with focus
    return MAIN_DIALOG[FOCUS]
end function

-----------------------------------------------------------------------------
global function get_dialog( integer a )
    -- returns attribute a of dialog
    return MAIN_DIALOG[a]
end function

-----------------------------------------------------------------------------
global procedure set_dialog( integer a, object o)
    -- sets attribute a of dialog to o
    MAIN_DIALOG[a] = o
end procedure



-----------------------------------------------------------------------------
global function get_control( integer i )
    -- returns control #1
    return MAIN_DIALOG[ITEMS][i]
end function

-----------------------------------------------------------------------------
global procedure set_control( integer i, sequence s )
    MAIN_DIALOG[ITEMS][i] = s
end procedure


-----------------------------------------------------------------------------
global function get_attrib( integer i, integer a )
    -- returns attrib a of control i
    return MAIN_DIALOG[ITEMS][i][a]
end function

-----------------------------------------------------------------------------
global procedure set_attrib( integer i, integer a, object o )
    MAIN_DIALOG[ITEMS][i][a] = o
end procedure



-----------------------------------------------------------------------------
global function navigation_key( integer keyPress )

    -- returns true if the key could cause another item to be
    -- selected on a list

    return  (   keyPress = UP
                or keyPress = DOWN
                or keyPress = PAGE_UP
                or keyPress = PAGE_DOWN
                or keyPress = HOME
                or keyPress = END
                or keyPress = CTRL_HOME
                or keyPress = CTRL_END
                or keyPress = MOUSE_CLICK 
                or keyPress = MOUSE_HELD
                or keyPress = MOUSE_DOUBLE
                or keyPress <= 255 )

end function


-----------------------------------------------------------------------------

-- alt keys, in alpha order
constant ALPHA_ALT_KEYS = { ALT_A, ALT_B, ALT_C, ALT_D, ALT_E, ALT_F,
                            ALT_G, ALT_H, ALT_I, ALT_J, ALT_K, ALT_L,
                            ALT_M, ALT_N, ALT_O, ALT_P, ALT_Q, ALT_R,
                            ALT_S, ALT_T, ALT_U, ALT_V, ALT_W, ALT_X,
                            ALT_Y, ALT_Z }

-- alt keys, in numeric order
constant NUMERIC_ALT_KEYS = { ALT_0, ALT_1, ALT_2, ALT_3, ALT_4,
                              ALT_5, ALT_6, ALT_7, ALT_8, ALT_9 }

-----------------------------------------------------------------------------
global function normal_to_alt( integer k )

    -- maps a normal key to an alt key, if possible.
    -- if not alpha or numeric, returns -1

    -- case k
    if compare(k, 'A') > -1 and compare( k,  'Z') < 1 then 
        return ALPHA_ALT_KEYS[k-'A'+1]
    elsif compare(k, 'a') > -1 and compare( k,  'z') < 1 then 
        return ALPHA_ALT_KEYS[k-'a'+1]
    elsif compare(k, '0') > -1 and compare( k,  '9') < 1 then 
        return NUMERIC_ALT_KEYS[k-'0'+1]
    else
        return -1
    end if

end function

-----------------------------------------------------------------------------
function accelerate( integer index, integer k, integer altOnly )

    integer hotKey, flag
    sequence title

    -- returns true if k matches control's accelerator key
    -- altFlag controls if key is altKey or normal

    -- false if it has no accelerator key or title
    if get_attrib( index, HOT ) = 0
    or length( get_attrib( index, TITLE ) ) = 0 then
        return 0
    end if

    -- get the hotkey
    title = get_attrib( index, TITLE )          -- get title
    hotKey = title[ get_attrib( index, HOT ) ]   -- get char at hotkey

    -- read as alt key, or normal and alt?
    -- i.e.: pressing 'A' on command button is same as pressing ALT+A
    if altOnly then
        return (normal_to_alt( hotKey ) = k)
    else
        flag = (normal_to_alt( hotKey ) = k)
        if flag then
            return 1
        else
            return (normal_to_alt( hotKey ) = normal_to_alt( k ) )
        end if
    end if

end function

-----------------------------------------------------------------------------
procedure set_default_button()

    -- if none of the buttons have the focus flag set, sets the flag
    -- of the first one, and displays it

    integer buttonCount, defaultButton, buttonSet
    sequence cursorAt

    defaultButton = 0   -- button with default behavior
    buttonCount = 0     -- buttons on dialog
    buttonSet = 0       -- flag if one is set

    for i = 1 to length( get_dialog( ITEMS ) ) do
        if get_attrib( i, TYPE ) = BUTTON then
            -- increment button count
            buttonCount = buttonCount + 1

            -- first button is default unless over-ridden
            if buttonCount = 1 then
                defaultButton = i
            end if

            -- default button?
            if get_attrib( i, DEFAULT_BUTTON ) then
                defaultButton = i
            end if

            -- is it set?
            if get_attrib( i, FOCUS ) then
                buttonSet = 1
            end if
        end if
    end for

    -- are there buttons, but none set?
    if buttonCount > 0 and buttonSet = 0 then
        -- save cursor position
        cursorAt = get_position()

        -- set focus to default button
        set_attrib( defaultButton, FOCUS, 1 )

        -- display it
        set_control( defaultButton,
            draw_button( get_control( defaultButton ) ) )

        -- restore cursor
        position( cursorAt[1], cursorAt[2] )

    end if

end procedure

-----------------------------------------------------------------------------
function can_focus( integer i )

    -- return flag if item can be focused on

    -- Please note that this returns true for labels, although
    -- technically you cannot focus on them.

    if get_attrib( i, DIM ) then
        -- can't focus on dimmed controls
        return 0

    elsif get_attrib( i, TYPE ) = LIST then
        if length( get_attrib( i, ITEMS ) ) = 0 then
            -- can't focus on empty list            
            return 0
        end if
    end if

    -- probably good
    return 1

end function


-----------------------------------------------------------------------------
function mouse_on_item()

    -- return the number of the item the mouse clicked

    integer bottom, right, left, top

    for i = 1 to length( get_dialog( ITEMS ) ) do

        -- calculate positions      
        left =  get_attrib(i,COL)
        top = get_attrib(i,ROW)
        
        right =  get_attrib(i,COL)+get_attrib(i,WIDE)
        bottom = get_attrib(i,ROW)+get_attrib(i,DEEP)

        if get_attrib( i, TYPE ) = LABEL then
            -- can't focus on these

        elsif get_attrib( i, TYPE ) = LIST then

            if  MOUSE_COL >= get_attrib( i, COL )       -- left corner
            and MOUSE_ROW >= get_attrib( i, ROW )-1     -- top
            and MOUSE_COL <= right                      -- right corner
            and MOUSE_ROW <= bottom then                -- bottom

                -- in list
                if can_focus( i ) then
                    return i
                end if

            end if

        elsif get_attrib( i, TYPE ) = MLE then          -- multi-line edit

            if  MOUSE_COL >= get_attrib( i, COL )       -- left column
            and MOUSE_COL <= right                      -- scroll bar
            and MOUSE_ROW >= get_attrib( i, ROW )-1     -- top border
            and MOUSE_ROW <= bottom                     -- bottom border
            then
                -- in mle
                if can_focus( i ) then
                    return i
                end if
            end if


        elsif MOUSE_COL >= get_attrib( i, COL )         -- upper left corner
        and   MOUSE_ROW >= get_attrib( i, ROW )
        and   MOUSE_COL <  right                        -- lower left corner
        and   MOUSE_ROW <= bottom then

            -- in object
            if can_focus( i ) then
                return i
            end if
                                
        end if

    end for

    return 0

end function

-----------------------------------------------------------------------------
function close_box_clicked()
    -- true if close box was clicked

    integer col
    
    -- on title bar?
    if MOUSE_ROW = get_dialog(ROW) then
        -- on close box?
        if CLOSE_BOX_COL > 0 then
            -- box is on the left side
            col = get_dialog(COL)+CLOSE_BOX_COL-1
        else
            -- mouse is on the right side
            col = get_dialog(COL)+get_dialog(WIDE)+CLOSE_BOX_COL+1
        end if

        -- give some margin for error
        if MOUSE_COL > col-2 and MOUSE_COL < col+2 then
            return 1
        end if
    end if

    -- not on box
    return 0

end function

-----------------------------------------------------------------------------
global procedure give_dialog_key( integer event )

    -- hand the dialog a key to process

    integer state, nextFocus, accelFlag, group
    integer loopCount, mouseFlag, altOnly
    integer f, right, bottom

    accelFlag = 0   -- true if moved using accelerator key
    nextFocus = 0   -- which control the accelerator selected
    loopCount = 0   -- make sure the program doesn't get stuck
    mouseFlag = 0   -- key is not a mouse event

    -- exit state
    if event = ENTER
    and get_attrib( focus(), TYPE ) != MLE then
        -- pressing enter on non-MLE exits dialog
        set_dialog( EXIT_STATE, OKAYED )    -- dialog is finished
    else
        set_dialog( EXIT_STATE, MORE )      -- awaiting more keys
    end if

    -- close box?
    if event = MOUSE_CLICK then
        if close_box_clicked() then
            -- cancel 
            set_dialog( EXIT_STATE, CANCELLED )
            return
        end if
    end if

    -- check for mouse
    if event = MOUSE_CLICK
    or event = MOUSE_DOUBLE
    or event = MOUSE_HELD then
        -- did the mouse land on anything?
        nextFocus = mouse_on_item()
        if nextFocus = 0 then
            -- exit without taking action
            return
        end if

        mouseFlag = event   -- hold it
        event = MOUSE       -- generic event

    end if

    while 1 do 

        loopCount = loopCount + 1
        if loopCount > 100 then
            -- abort out
            set_dialog( EXIT_STATE, CANCELLED )
            return
        end if

        -- exit if cancelled
        if event = ESC then
            set_dialog( EXIT_STATE, CANCELLED)
            return
        end if

    -- get focus
    f = focus()

    -- select appropriate widget type
    -- case get_attrib( f, TYPE )

        if get_attrib( f, TYPE ) = MLE then  ------------------------------------------------------------

            -- case event

            if event = REFRESH then 
                set_control( f, edit_xle( get_control(f), REFRESH ) )

            elsif event = GET_FOCUS then 
                set_attrib( f, IS_FOCUS, 1 )
                set_control( f, edit_xle( get_control(f), 0 ) )

            elsif event = LOSE_FOCUS then 
                set_attrib( f, IS_FOCUS, 0 )

            elsif event = TAB then 
                event = MOVE_RIGHT

            elsif event = SHIFT_TAB then 
                event = MOVE_LEFT

            else
                set_control( f, edit_xle( get_control(f), event ) )

            end if


        elsif get_attrib( f, TYPE ) = SLE then  ------------------------------------------------------------

            -- case event

            if event = REFRESH then 
                set_attrib( f, REDRAW, 1 )
                set_control( f, draw_sle( get_control(f) ) )
                set_control( f, focus_on_sle( get_control(f) ) )

            elsif event = GET_FOCUS then 
                set_attrib( f, IS_FOCUS, 1 )
                set_control( f, focus_on_sle( get_control(f) )  )

            elsif event = LOSE_FOCUS then 
                set_attrib( f, IS_FOCUS, 0 )
                set_control( f, focus_off_sle( get_control(f) ) )

            elsif event = TAB then 
                set_control( f, focus_off_sle( get_control(f) ) )
                event = MOVE_RIGHT

            elsif event = SHIFT_TAB then 
                set_control( f, focus_off_sle( get_control(f) ) )
                event = MOVE_LEFT

            else
                set_control( f, edit_xle( get_control(f), event ) )
                set_control( f, draw_sle( get_control(f) )  )

            end if

        elsif get_attrib( f, TYPE ) = LIST then  -----------------------------------------------------------

            -- case event

            if event = REFRESH then 
                set_attrib( f, REDRAW, 1 )
                set_control( f, draw_list( get_control(f) ) )
                draw_selected_list_item( get_control(f) )

            elsif event = GET_FOCUS then 
                set_attrib( f, IS_FOCUS, 1 )
                draw_selected_list_item( get_control(f) )
                event = 0

            elsif event = LOSE_FOCUS then 
                set_attrib( f, IS_FOCUS, 0 )
                -- need to erase the highlight?
                if get_attrib( f, PERSIST ) = 0 then
                    set_control( f, draw_list( get_control(f) ) )
                end if

            elsif event = TAB then 
                -- shift
                event = MOVE_RIGHT

            elsif event = SHIFT_TAB then 
                -- shift
                event = MOVE_LEFT

            else
                -- edit?
                set_control( f, edit_list( get_control(f), event ) )

                right = get_attrib(f, COL) + get_attrib(f, WIDE)
                bottom = get_attrib(f, ROW) + get_attrib(f, DEEP)

                if  event = MOUSE_DOUBLE                    -- double click
                and MOUSE_COL < right                       -- not on scroll 
                and MOUSE_ROW >= get_attrib(f, ROW)         -- not to left
                and MOUSE_ROW < bottom  then                -- not below

                    set_dialog( EXIT_STATE, OKAYED )        -- finished

                end if

            end if

        elsif get_attrib( f, TYPE ) = BUTTON then  --------------------------------------------------------

            set_attrib( f, FOCUS, 1 )
            set_control( f, edit_button( get_control(f), event ) )

            -- case event
            if event = GET_FOCUS then 
                set_attrib( f, IS_FOCUS, 1 )

                -- make sure no other buttons have focus
                for i = 1 to length( get_dialog( ITEMS ) ) do
                    -- different control?
                    if  i != f then
                        -- is it a button with focus?
                        if get_attrib( i, TYPE ) = BUTTON
                        and get_attrib( i, FOCUS ) != 0 then

                            -- remove focus
                            set_attrib( i, FOCUS, 0 )

                            -- re-display button
                            set_control( i, draw_button( get_control( i ) ) )

                        end if
                    end if
                end for

            elsif event = LOSE_FOCUS then 
                set_attrib( f, IS_FOCUS, 0 )
                set_attrib( f, FOCUS, 0 )
                set_control( f, draw_button( get_control(f) ) )

            elsif event =  ' ' or event = ENTER then 
                if get_attrib( f, CANCEL_BUTTON ) then
                    set_dialog( EXIT_STATE, CANCELLED )
                else
                    set_dialog( EXIT_STATE, OKAYED )
                end if
                event = 0

            elsif event = MOUSE_CLICK then 
                -- is the mouse still over the button?
                if mouse_on_item() = get_dialog( FOCUS ) then
                    if get_attrib( f, CANCEL_BUTTON ) then
                        set_dialog( EXIT_STATE, CANCELLED )
                    else
                        set_dialog( EXIT_STATE, OKAYED )
                    end if
                    event = 0
                end if

                

            elsif event = TAB then 
                set_attrib( f, FOCUS, 0 )
                set_control( f, draw_button( get_control(f) ) )
                event = MOVE_RIGHT

            elsif event = SHIFT_TAB then                 
                set_attrib( f, FOCUS, 0 )
                set_control( f, draw_button( get_control(f) ) )
                event = MOVE_LEFT

            end if

        elsif get_attrib( f, TYPE ) = CHECK then  ----------------------------------------------------------

            set_control( f, edit_check( get_control(f), event ) )

            -- case event

            if event = GET_FOCUS then 
                set_attrib( f, IS_FOCUS, 1 )

            elsif event = LOSE_FOCUS then 
                set_attrib( f, IS_FOCUS, 0 )
                set_control( f, draw_check( get_control(f) ) )

            elsif event = TAB then 
                event = MOVE_RIGHT

            elsif event = SHIFT_TAB then 
                event = MOVE_LEFT

            end if

        elsif get_attrib( f, TYPE ) = RADIO then  ----------------------------------------------------------

            set_control( f, edit_radio( get_control(f), event ) )

            -- case event
            if event =  ' ' or event = MOUSE_CLICK or event = MOUSE_DOUBLE then 
                -- clear off the others
                for i = 1 to length( get_dialog( ITEMS ) ) do

                    -- radio button in same group?
                    if  get_attrib( i, TYPE ) = RADIO
                    and get_attrib( i, GROUP ) = get_attrib( f, GROUP ) then
                        -- not the current button
                        if i != f then
                            -- clear the button
                            set_attrib( i, CHECKED, 0 )

                            -- display it
                            set_control( i, draw_radio( get_control( i ) ) )

                        end if
                    end if
                end for

                -- move focus back to this control
                set_control( f, draw_radio( get_control(f) ) )

            elsif event = GET_FOCUS then 
                set_attrib( f, IS_FOCUS, 1 )

            elsif event = LOSE_FOCUS then 
                set_attrib( f, IS_FOCUS, 0 )
                set_control( f, draw_radio( get_control(f) ) )

            elsif event = TAB then 
                event = MOVE_RIGHT

            elsif event = SHIFT_TAB then 
                event = MOVE_LEFT

            end if

        elsif get_attrib( f, TYPE ) = LABEL then 
            -- you should never get focus on a label
                event = MOVE_RIGHT

        end if

        -- handle movement
        -- case event
            if event = 0 then  ----------------------------------------------------------

                -- don't exit if a mouse event is pending
                if mouseFlag != 0 then

                    event = mouseFlag       -- make it the event
                    mouseFlag = 0           -- clear the pending flag

                else

                    -- safe to exit
                    set_default_button()    -- set button focus
                    exit

                end if


            elsif event = MOUSE then  ------------------------------------------------------

                -- find the control the mouse was clicked on
                if nextFocus != f then
                    -- move off the current item
                    event = LOSE_FOCUS
                else
                    -- let exit handle mouse pending flag
                    event = 0
                end if
            
            elsif event = MOVE_LEFT then  --------------------------------------------------
                -- move to next control
                if f = 1 then
                    nextFocus = length( get_dialog( ITEMS ) )
                else
                    nextFocus = f - 1
                end if

                -- skip dimmed and label controls and empty lists
                while not can_focus( nextFocus )
                or get_attrib( nextFocus, TYPE ) = LABEL do

                    -- skip to end/move back
                    if nextFocus = 1 then
                        nextFocus = length( get_dialog( ITEMS ) )
                    else
                        nextFocus = nextFocus - 1
                    end if
                end while

                -- next event
                event = LOSE_FOCUS

            elsif event = MOVE_RIGHT then  -------------------------------------------------
                -- move to next control
                if f = length( get_dialog( ITEMS ) ) then
                    nextFocus = 1 
                else
                    nextFocus = f + 1
                end if

                -- skip dimmed and label controls
                while not can_focus( nextFocus )
                or get_attrib( nextFocus, TYPE ) = LABEL do

                    -- move to front/move ahead
                    if nextFocus = length( get_dialog( ITEMS ) ) then
                        nextFocus = 1 
                    else
                        nextFocus = nextFocus + 1
                    end if
                end while

                -- next event
                event = LOSE_FOCUS

            elsif event = GET_FOCUS then  --------------------------------------------------
                -- default is all done
                event = 0

                -- unless an accelerator key was pressed...
                if accelFlag then
                    -- clear the flag
                    accelFlag = 0

                    -- act as if key were pressed
                    -- case get_attrib( f, TYPE )
                        if get_attrib( f, TYPE ) = RADIO or get_attrib( f, TYPE ) = CHECK then 
                        -- treat as if space bar was pressed
                        event =  ' '

                        elsif get_attrib( f, TYPE ) = BUTTON then 
                        -- treat as if "enter" were pressed
                        event = ENTER
                    end if
                end if

            elsif event = LOSE_FOCUS then  -------------------------------------------------
                if nextFocus > 0 then
                    event = GET_FOCUS
                    set_dialog(FOCUS, nextFocus)
                else
                    -- oops. we should *not* get here!                    
                    exit
                end if

            else -------------------------------------------------------
                -- if sle/mle or list, only ALT keys are hotkeys
                -- otherwise, A is treated as ALT+A

                if get_attrib( f, TYPE ) = MLE
                or get_attrib( f, TYPE ) = SLE
                or get_attrib( f, TYPE ) = LIST then

                    -- only treat alt keys as hot keys
                    altOnly = 1

                else

                    -- treat normal keys as hot keys
                    altOnly = 0

                end if

                -- hot key?
                for i = 1 to length( get_dialog(ITEMS) ) do

                    -- accelerator and not dim?
                    if  accelerate( i, event, altOnly )
                    and can_focus( i ) then

                        accelFlag = 1
                        event = LOSE_FOCUS

                        -- if label, move to next item
                        if get_attrib( i, TYPE ) = LABEL then
                            nextFocus = i + 1
                            -- wrap, if needed
                            if nextFocus = length( get_dialog( ITEMS ) ) then
                                nextFocus = 1
                            end if
                        else
                            nextFocus = i
                        end if
                        -- leave loop
                        exit
                    end if
                end for


                -- no accelerator?
                if accelFlag = 0 then
                    exit
                end if
        end if
        
    end while

    -- return dialog
    return

end procedure

-----------------------------------------------------------------------------
global procedure edit_dialog()

    -- get a key, and feed it to the dialog

    integer keyPressed
    
    -- display cursor
    if VGA_MODE then
        -- only show if in MLE or SLE
        if get_attrib( focus(), TYPE ) = SLE
        or get_attrib( focus(), TYPE ) = MLE then
            show_graphic_cursor()
        end if
    else
        show_cursor()                               -- display the cursor
    end if

    keyPressed = watch_keys()                       -- wait for a key
    set_dialog( KEY_PRESSED, keyPressed )           -- save it
    
    -- hide cursor
    if VGA_MODE then
        -- only show if in MLE or SLE
        if get_attrib( focus(), TYPE ) = SLE
        or get_attrib( focus(), TYPE ) = MLE then
            hide_graphic_cursor()
        end if
    else
        hide_cursor()
    end if
    
    give_dialog_key( keyPressed )                   -- give to dialog

end procedure

-----------------------------------------------------------------------------
global procedure draw_controls()

    -- display controls

    -- the act of drawing controls updates some status information,
    -- such as the DIM, WIDE and REDRAW.

    for i = 1 to length( get_dialog( ITEMS ) ) do
        -- case get_attrib( i, TYPE )
            if get_attrib( i, TYPE ) = MLE then 
                    set_control( i, draw_mle( get_control(i) ) )
            elsif get_attrib( i, TYPE ) = SLE then 
                    set_control( i, draw_sle( get_control(i) ) )
            elsif get_attrib( i, TYPE ) = LIST then 
                    set_control( i, draw_list( get_control(i) ) )
            elsif get_attrib( i, TYPE ) = BUTTON then 
                    set_control( i, draw_button( get_control(i) ) )
            elsif get_attrib( i, TYPE ) = CHECK then 
                    set_control( i, draw_check( get_control(i) ) )
            elsif get_attrib( i, TYPE ) = RADIO then 
                    set_control( i, draw_radio( get_control(i) ) )
            elsif get_attrib( i, TYPE ) = LABEL then 
                    set_control( i, draw_label( get_control(i) ) )
        end if
    end for

end procedure

----------------------------------------------------------------------------
global procedure show_dialog()

    integer withText
    integer row, col, wide, deep
    sequence s

    -- display the dialog box

    -- there are two ways we can describe the dialog. either the dialog
    -- is passed as a sequence in DATA, or the size is specified in WIDE
    -- and DEEP.

    -- save cursor position
    set_dialog(CURSOR_POS, get_position())


    -- do some basic checks on the dialog

    -- are there any controls?
    if compare( get_dialog(ITEMS), 0 ) = 0 then
        clear_screen()
        puts( 1, "Dialog has no controls defined.\n" )
        abort(0)
    end if

    -- are all the controls initialized?
    for i = 1 to length( get_dialog(ITEMS) ) do
        if compare( get_control(i), 0 ) = 0 then
            -- undefined, error
            clear_screen()
            printf( 1, "Item #%g in control array is undefined.\n", {i} )
            abort(0)
        end if
    end for

    -- if no size, calculate from data
    if get_dialog( DEEP ) = 0 then
        withText = 1                            -- text passed flag

        set_dialog( DEEP,                       -- title is part of the size
            length( get_dialog(DATA) ) + 1 )

        s = get_dialog( DATA )                  -- width = line 1 length
        set_dialog( WIDE, length( s[1] ) )

    else
        withText = 0                    -- text not passed flag
    end if

    -- calculate position based on depth screen
    if get_dialog( ROW ) = 0 then
        set_dialog( ROW, floor( ( (SCREEN_DEEP+1) - get_dialog(DEEP) )/2 ) )
    end if

    if get_dialog( COL ) = 0 then
        -- count shadow in centering
        set_dialog( COL, floor( (SCREEN_WIDE -(get_dialog(WIDE)-2) )/2 ) )
    end if

    -- make calculations simpler
    row = get_dialog( ROW )
    col = get_dialog( COL )
    wide = get_dialog( WIDE )
    deep = get_dialog( DEEP )

    -- save screen (and shadow)
    set_dialog( CLEAN_SCREEN, save_clean_image( {row, col},
        {row+deep, col+wide+1} ) )

    -- defer to make "snappier"
    defer_display()

    -- adjust all the items in the dialog
    for i = 1 to length( get_dialog( ITEMS ) ) do
        -- position relative to dialog
            set_attrib( i, ROW, get_attrib( i, ROW ) + row )
            set_attrib( i, COL, get_attrib( i, COL ) + col )

        -- vga lists move up 1 row, and are lengthened
            if  ( get_attrib( i, TYPE ) = LIST
                  or  get_attrib( i, TYPE ) = MLE )
            and VGA_MODE then

                set_attrib( i, DEEP, get_attrib( i, DEEP ) + 1 )
                set_attrib( i, ROW, get_attrib( i, ROW ) - 1 )

            end if

    end for

    -- video or text mode?
    if VGA_MODE then

        -- graphic window
        graphic_window( get_dialog( TITLE ), 1, BRIGHT_BLUE,
            col-1, row, wide+1, deep )

    else

        -- text window

        -- draw the title bar
        position( row, col )
        put_colors( OF_DIALOG_TITLE, repeat(   ' ', get_dialog(WIDE) ) )
    
        -- title bar text
        position( row, col+2 )
        put_colors( OF_DIALOG_TITLE, get_dialog(TITLE) )

        -- close box
        CLOSE_BOX_COL = -4
        position( row, col+wide-4 )
        put_colors( OF_DIALOG_TITLE, "[" & 254 & "]" )

    end if

    -- draw text in the box?
    if withText then
        s = get_dialog( DATA )
        for i = 1 to length( s ) do
            if VGA_MODE then
                -- write onto window
                graphic_text( row+i, col, OF_DIALOG[1], OF_DIALOG[2], s[i] )
            else
                position( row+i, col )
                put_colors( OF_DIALOG, s[i] )
                put_shadow( 2 )
            end if
        end for
    elsif not VGA_MODE then
        for i = 1 to deep-1 do
            position( row+i, col )
            put_colors( OF_DIALOG, repeat(   ' ', wide ) )
            put_shadow( 2 )
        end for
    end if

    -- shadow under the box
    position( row + deep, col + 2 )
    put_shadow( wide )    

    -- populate the dialog
    draw_controls()

    -- set the "more" flag
    set_dialog( EXIT_STATE, MORE )

    -- focus on the first legal item
    for i = 1 to length( get_dialog( ITEMS ) ) do

        -- can focus, and not a lable
        if can_focus( i )
        and get_attrib( i, TYPE ) != LABEL then
            set_dialog( FOCUS, i )
            exit
        end if

    end for

    -- focus on the item
    give_dialog_key( GET_FOCUS )

    -- now display    
    show_deferred()

end procedure

-----------------------------------------------------------------------------
function find_ampersand( sequence s )
    -- find where hot key is in text
    -- it is preceeded by an ampersand "&"

    integer hot
    sequence txt

    hot = 0
    txt = ""
    for i = 1 to length( s ) do
        if s[i] = '&' then
            hot = i
        else
            txt = txt & s[i]
        end if
    end for

    return { txt, hot }

end function


-----------------------------------------------------------------------------
global procedure m_bar( integer i, sequence s )

    -- add an item to MAIN_MENU

    integer col
    sequence m, result

    -- abort if not in sequence
    if length( MAIN_MENU ) != i - 1 then
        clear_screen()
        puts( 1, "Error in m_bar: Items must be declared in sequence.\n" )
        abort( 0 )
    end if

    -- find hotkey
    result = find_ampersand( s )

    -- find column
    if i = 1 then
        -- start in column 4
        col = 4
    else
        -- position 3 characters after prior items
        col = length( MAIN_MENU[i-1][TITLE] ) +
            MAIN_MENU[i-1][COL] + 3
    end if


    -- build control
    m = new_control( DROPDOWN )

    -- unique values
    m[TITLE] = result[1]
    m[HOT] = result[2]
    m[COL] = col-1
    m[ITEMS] = {}


    -- append control
    MAIN_MENU = append( MAIN_MENU, m )

end procedure


-----------------------------------------------------------------------------
global procedure m_item( integer i, integer j, sequence s, sequence h )

    -- define a menu dropdown item for a menu bar

    sequence d, result

    -- make sure menu item exists
    if i > length( MAIN_MENU ) then
        clear_screen()
        printf( 1, "Error in m_item: Menu bar %g does not exist.\n", {i} )
        abort( 0 )
    end if

    -- must be defined in order
    if j-1 != length( MAIN_MENU[i][ITEMS] ) then
        clear_screen()
        printf( 1, "Error in m_item: Item %g must be declared in sequence.\n",
             {j} )
        abort( 0 )
    end if

    -- find hotkey
    result = find_ampersand( s )

    -- define a generic dropdown item
    d = new_control( DROPDOWN )

    -- add unique values
    d[TITLE] = result[1]
    d[HOT] = result[2]
    d[HELP_TEXT] = h


    -- append control
    MAIN_MENU[i][ITEMS] = append( MAIN_MENU[i][ITEMS], d )
    
end procedure


-----------------------------------------------------------------------------
global procedure add_control( integer index,     -- index on dialog
                              integer t,         -- control type
                              sequence title,    -- title
                              integer row,       -- row
                              integer col,       -- column
                              integer wide,      -- width
                              integer deep,      -- depth
                              sequence flags  )  -- list of flags

    -- build a new control
    -- example: control( 1, BUTTON, "&Ok", 2, 3, 8, 0, {DEFAULT_BUTTON} )
    --          creates a button with the text "Ok"
    --          hotkey is char #1
    --          row is 2
    --          column is 3
    --          width is 8
    --          depth is 0
    --          DEFAULT_BUTTON attribute is 1

    sequence s, result

    -- must be in order
    if length( get_dialog(ITEMS) ) != index-1 then
        clear_screen()
        printf( 1, "Err in control: control %g out of sequence.", {index} )
        abort( 0 )
    end if    
        

    -- build control    
        s = new_control( t )

    -- get hotkey
    result = find_ampersand( title )

    -- attributes
    -- don't overwrite values if zero was passed
        if row != 0 then    s[ROW] = row    end if
        if col != 0 then    s[COL] = col    end if
        if wide != 0 then   s[WIDE] = wide  end if
        if deep != 0 then   s[DEEP] = deep  end if
        s[TITLE] = result[1]
        s[HOT] = result[2]

    -- true attributes
    for i = 1 to length( flags ) do
        s[flags[i]] = 1
    end for
        
    set_dialog( ITEMS, append( get_dialog(ITEMS), s ) )
    
end procedure


-----------------------------------------------------------------------------
global procedure new_dialog( sequence title,    -- title
                             integer row,       -- row
                             integer col,       -- column
                             integer wide,      -- width
                             integer deep,      -- depth
                             sequence layout )  -- layout

    -- push dialog
    DIALOG_STACK = append( DIALOG_STACK, MAIN_DIALOG )

    MAIN_DIALOG = new_control( DIALOG )
    set_dialog(TITLE, title)
    set_dialog(ROW, row)
    set_dialog(COL, col)
    set_dialog(WIDE, wide)
    set_dialog(DEEP, deep)
    set_dialog(DATA, layout)
    set_dialog(ITEMS, {})

end procedure



-----------------------------------------------------------------------------    
global procedure run_dialog()    

    -- generic dialog run routine

    -- display the objects
    show_dialog()

    -- loop forever
    while get_dialog(EXIT_STATE) = MORE do
        edit_dialog()    -- process keys
    end while

end procedure



----------------------------------------------------------------------------
global procedure hide_dialog()

    -- hides the dialog
    -- pop the prior dialog into current context

    sequence cursorAt

    -- restore screen
    display_clean_image( {get_dialog(ROW), get_dialog(COL)},
        get_dialog(CLEAN_SCREEN) )

    -- place cursor
    cursorAt = get_dialog( CURSOR_POS )
    position( cursorAt[1], cursorAt[2] )

    -- pop stack
    if length( DIALOG_STACK ) != 0 then
        MAIN_DIALOG = DIALOG_STACK[length(DIALOG_STACK)]
        DIALOG_STACK = DIALOG_STACK[1..length(DIALOG_STACK)-1]
    end if
    
end procedure


