-- Common Dialog
--  colors_dialog()
--  choose colors

global procedure colors_dialog()


    sequence cursorAt
    integer keyPress

    sequence controlList, colorList, scheme
    integer this, fColor, bColor

    -- skip this if VGA Mode
    if VGA_MODE then
        -- warn user
            message_box( "Editor",
                {"Color schemes in Graphic mode are hard coded.",
                 "You can only edit colors in text mode." } )

        -- exit
            return
    end if

    -- the controls
    controlList = {
        "Normal Text", "Selected Text", "Window Border", "Menubar",
        "Status Line", "Keyboard Accelerators", "Disabled Items",
        "Dialogs", "Dialog Titlebar", "Dialog Buttons", "Dialog Hot Keys",
        "Dialog Scrollbars" }

    -- the colors
    colorList = {
        "Black", "Blue", "Green", "Cyan", "Red", "Magenta", "Brown", "White",
        "Gray", "BrBlue", "BrGreen", "BrCyan", "BrRed", "BrMagenta", "Yellow",
        "BrWhite" }

    -- the color schemes
    scheme = {  OF_TEXT,
                OF_SELECTED,
                OF_BORDER,
                OF_MENUBAR,
                OF_STATUS,
                OF_ACCEL,
                OF_DISABLED,
                OF_DIALOG,
                OF_DIALOG_TITLE,
                OF_DIALOG_BUTTON,
                OF_HOTKEY,
                OF_SCROLL }


    -- put stuff in the box
    new_dialog( "Colors", 0, 0, 0 , 0, {
        "                                                              ",
        "   Item:                       Foreground:     Background:    ",
        "                                                              ",
        "                                                              ",
        "                                                              ",
        "                                                              ",
        "                                                              ",
        "                                                              ",
        "                                                              ",
        "                                                              ",
        "                                                              ",
        "              Choose the colors for the item                  ",
        "                                                              ",
        "                                                              ",
        "                                                              " } )

    -- give them handles
    -- enum C_TYPE, FORE, BACK, C_DEFAULT, OK, CANCEL, HELP

    -- controls
    add_control( 1, LIST, "&Item:", 4, 3, 24, 6, {PERSIST} )       
    add_control( 2, LIST, "&Foreground:", 4, 31, 12, 6, {PERSIST} )        
    add_control( 3, LIST, "&Background:", 4, 47, 12, 6, {PERSIST} )        
    add_control( 4, BUTTON, "&Default", 14, 5, 11, 0, {} )
    add_control( 5, BUTTON, "&OK", 14, 21, 8, 0, {DEFAULT_BUTTON} )
    add_control( 6, BUTTON, "&Cancel", 14, 34, 10, 0, {CANCEL_BUTTON} )
    add_control( 7, BUTTON, "&Help", 14, 50, 8, 0, {} )

    -- populate the lists
        set_attrib( 1, ITEMS, controlList )
        set_attrib( 2, ITEMS, colorList )
        set_attrib( 3, ITEMS, colorList )

    -- these should be set by C_TYPE. I'm just being paranoid.
        fColor = BLACK
        bColor = WHITE

    -- display the objects
    show_dialog()

    -- loop forever
    while get_dialog( EXIT_STATE ) = MORE do

        -- save the cursor
        cursorAt = get_position()

        -- make "snappier"
        defer_display()

        -- case focus()
        if focus() = 1 then 

            -- get the colors
            this = list_selected_index( get_control(1) )
            fColor = scheme[this][1]
            bColor = scheme[this][2]

            -- change focus on new foreground color
            set_control( 2, focus_on_list_item( get_control(2), fColor+1 ) )

            -- redraw list
            set_control( 2, draw_list( get_control(2) ) )

            -- draw selected item
            draw_selected_list_item( get_control(2) )

            -- change focus on new background color
            set_control( 3, focus_on_list_item( get_control(3), bColor+1 ) )

            -- redraw list
            set_control( 3, draw_list( get_control(3) ) )

            -- draw selected item
            draw_selected_list_item( get_control(3) )

        elsif focus() = 2 then 
            -- update the selected foreground color
            fColor = list_selected_index( get_control(2) ) - 1
            this = list_selected_index( get_control(1) )
            scheme[this][1] = fColor

        elsif focus() = 3 then 
            -- update the selected background color
            bColor = list_selected_index( get_control(3) ) - 1
            this = list_selected_index( get_control(1) )
            scheme[this][2] = bColor

        end if

        -- display the text in the color
        position( get_dialog(ROW)+12, get_dialog(COL)+13 )
        put_colors( { fColor, bColor }, " Choose the colors for the item " )

        -- display it
        show_deferred()

        -- restore the cursor
        position( cursorAt[1], cursorAt[2] )
        show_cursor()

        edit_dialog()    -- process keys

        -- help key?
        if  get_dialog( FOCUS ) = 7                     -- help key?
        and get_dialog( EXIT_STATE ) = OKAYED then
            message_box( "Colors", {
            "Determines the color of screen elements.                       ",
            "                                                               ",
            " From the Item list, click the item you want to change.       ",
            "                                                               ",
            "  - Select a color from the Foreground list.                   ",
            "  - Select a color from the Background list.                   ",
            "                                                               "} )




                set_dialog( EXIT_STATE, MORE )          -- stay in loop
        end if


    end while

    if get_dialog( EXIT_STATE ) = OKAYED then
        -- save the colors
            OF_TEXT             = scheme[1]
            OF_SELECTED         = scheme[2]
            OF_BORDER           = scheme[3]
            OF_MENUBAR          = scheme[4]
            OF_STATUS           = scheme[5]
            OF_ACCEL            = scheme[6]
            OF_DISABLED         = scheme[7]
            OF_DIALOG           = scheme[8]
            OF_DIALOG_TITLE     = scheme[9]
            OF_DIALOG_BUTTON    = scheme[10]
            OF_HOTKEY           = scheme[11]
            OF_SCROLL           = scheme[12]

        -- set flag to redraw 
        redraw_flag = 1

    end if

    -- restore screen
    hide_dialog()


end procedure


