-- Open File
-- Common Dialog

include files.e
include dialog.e
include d_msgbox.e
include d_yesno.e

-----------------------------------------------------------------------------
global function open_file_dialog( sequence currentDir )

    -- returns
    --  OK:          { file name, current directory }
    --  CANCELLED:   -1


    sequence cursorAt, tmp, fileName
    integer keyPress, f
    object result

    -- put stuff in the box
    new_dialog( "Open File", 0, 0, 47, 22, {} )

    -- enum NAME, DIR, FILES, DIRS, OK, CANCEL, HELP

    -- controls
        add_control( 1, SLE, "File &Name:", 2, 13, 32, 0, {} )
        add_control( 2, LABEL, currentDir, 3, 2, 40, 0, {} )
        add_control( 3, LIST, "&Files:", 6, 2, 21, 12, {} )
        add_control( 4, LIST, "&Directories:", 6, 26, 19, 12, {} )
        add_control( 5, BUTTON, "&OK", 20, 4, 8, 0, {DEFAULT_BUTTON, FOCUS} )
        add_control( 6, BUTTON, "&Cancel", 20, 19, 8, 0, {CANCEL_BUTTON} )
        add_control( 7, BUTTON, "&Help", 20, 35, 8, 0, {} )


    -- values
        set_attrib( 1, DATA, "*.*" )
        set_attrib( 3, ITEMS, file_list( currentDir, "*.*" ) )
        set_attrib( 4, ITEMS, dir_list( currentDir ) )
    

    -- clear file name
    fileName = ""

    -- display the objects
    show_dialog()

    -- loop forever
    while get_dialog( EXIT_STATE ) = MORE do

        -- set to display
        edit_dialog()
        keyPress = get_dialog( KEY_PRESSED )
        f = get_dialog( FOCUS )

        -- select appropriate widget
        -- case f

        if f = 3 then  -- file list
            if navigation_key( keyPress ) then

                -- get the cursor
                cursorAt = get_position()

                -- move the field to the text box                
                set_attrib( 1, DATA, list_selected_text( get_control(3) ) )
                set_control( 1, draw_sle( get_control(1) ) )

                -- replace the cursor
                position( cursorAt[1], cursorAt[2] )

            end if

        elsif f = 4 then  -- directory list
            
            if navigation_key( keyPress ) then

                -- get the cursor
                cursorAt = get_position()

                -- get the selected text
                tmp = list_selected_text( get_control(f) )

                -- drive?
                if tmp[1] = '[' then
                    -- rewrite drive
                    tmp = {tmp[3]} & ":\\*.*"
                else
                    -- rewrite with full path
                    if last_char( currentDir ) = '\\' then
                        tmp = currentDir & tmp & "\\*.*"
                    else
                        tmp = currentDir & "\\" & tmp & "\\*.*"
                    end if
                end if

                -- move the field to the test box
                set_attrib( 1, DATA, tmp )
                set_control( 1, draw_sle( get_control(1) ) )

                -- replace the cursor
                position( cursorAt[1], cursorAt[2] )

            end if

        end if

        -- was HELP button selected?
        if  get_dialog( FOCUS ) = 7
        and get_dialog( EXIT_STATE ) = OKAYED then

            message_box( "Open File", {
            "Opens an existing file or creates a new named file.            ",
            "                                                               ",
            " Use the Files and Directories boxes to select the path and   ",
            "  file.                                                        ",
            "                                                               ",
            " If the file doesn't exist, a new, unsaved empty file is      ",
            "  created.                                                     ",
            "                                                               ",
            " Hint: You can type the path name in the File Name box.       "} )

            -- don't exit
            set_dialog( EXIT_STATE, MORE )

        end if

        -- something other than cancelled?
        if get_dialog( EXIT_STATE ) = OKAYED then

            -- file/directory in text box?
            if compare( get_attrib(1,DATA), "" ) = 0 then

                -- no text
                message_box( "Open File", "You must specify a filename." )

                -- don't exit
                set_dialog( EXIT_STATE, MORE )

            else

                -- text was entered

                -- convert to path & file
                tmp = full_path( get_attrib( 1, DATA ), currentDir )

                if compare( tmp[3], "f" ) = 0 then

                    -- save the file name
                    fileName = tmp[2]

                    -- save the new path
                    currentDir = tmp[1]

                elsif compare( tmp[3], "d") = 0 then

                    -- it's a directory

                    -- change the current directory
                    currentDir = tmp[1]

                    -- new directory                    
                    set_control( 2, reset_control( get_control(2) ) )
                    set_attrib( 2, TITLE, currentDir )

                    -- update the file list
                    set_control( 3, reset_control( get_control(3) ) )
                    set_attrib( 3, ITEMS, file_list( currentDir, "*.*" ) )

                    -- update the directory list
                    set_control( 4, reset_control( get_control(4) ) )
                    set_attrib( 4, ITEMS, dir_list( currentDir ) )

                    -- update the file text
                    set_control( 1, reset_control( get_control(1) ) )
                    set_attrib( 1, DATA, tmp[2] )

                    -- draw all the controls
                    draw_controls()

                    -- focus on the first item
                    set_dialog( FOCUS, 1 )
                    give_dialog_key( GET_FOCUS )

                    -- don't exit
                    set_dialog( EXIT_STATE, MORE )


                elsif compare( tmp[3], "m") = 0 then

                    -- it's a mask

                    -- change the current directory
                    currentDir = tmp[1]

                    -- new directory
                    set_control( 2, reset_control( get_control(2) ) )
                    set_attrib( 2, TITLE, currentDir )

                    -- update the file list
                    set_control( 3, reset_control( get_control(3) ) )
                    set_attrib( 3, ITEMS, file_list( currentDir, tmp[2] ) )

                    -- update the directory list
                    set_control( 4, reset_control( get_control(4) ) )
                    set_attrib( 4, ITEMS, dir_list( currentDir ) )


                    -- update the file text
                    set_control( 1, reset_control( get_control(1) ) )
                    set_attrib( 1, DATA, tmp[2] )


                    -- draw all the controls
                    draw_controls()

                    -- focus on the first item
                    set_dialog( FOCUS, 1 )
                    give_dialog_key( GET_FOCUS )

                    -- don't exit
                    set_dialog( EXIT_STATE, MORE )

                else

                    -- it's an error
                    message_box( "Open File", "Invalid filename or path." )

                    -- don't exit
                    set_dialog( EXIT_STATE, MORE )

                end if


            end if
        end if

    end while

    -- not cancelled
    if get_dialog( EXIT_STATE ) = OKAYED then
        if compare( fileName, "" ) != 0 then
            -- file name
            result = { fileName, currentDir }            
        else
            -- fail, no file name
            result = -1
        end if
    else
        -- cancelled
        result = -1
    end if
    

    -- restore screen
    hide_dialog()

    return result
    
end function

