-- Save File/Save File As
-- Common Dialog

include files.e
include dialog.e
include d_msgbox.e
include d_yesno.e


global function save_as_dialog( sequence fName, sequence currentDir )

    -- returns
    --  OK:          { file name, current directory }
    --  CANCELLED:   -1
        
    sequence cursorAt, tmp, fileName
    integer keyPress
    object result

    -- put stuff in the box
    new_dialog( "Save File", 0, 0, 50, 21, {} )

    -- give them handles
    -- enum NAME, DIR, FILES, DIRS, OK, CANCEL, HELP

    -- controls
    add_control( 1, SLE, "File &Name:", 2, 14, 32, 0, {} )        
    add_control( 2, LABEL, currentDir, 3, 2, 0, 0, {} )
    add_control( 3, LIST, "&Existing Files:", 6, 3, 21, 10, {} )        
    add_control( 4, LIST, "&Directories:", 6, 27, 19, 10, {} )        
    add_control( 5, BUTTON, "&OK", 18, 3, 8, 0, {DEFAULT_BUTTON} )
    add_control( 6, BUTTON, "&Cancel", 18, 18, 10, 0, {CANCEL_BUTTON} )
    add_control( 7, BUTTON, "&Help", 18, 36, 8, 0, {} )

    -- values
    set_attrib( 1, DATA, fName )
    set_attrib( 3, ITEMS, file_list( currentDir, "*.*" ) )
    set_attrib( 4, ITEMS, dir_list( currentDir ) )

    -- display the objects
    show_dialog()

    -- clear file name
    fileName = ""

    -- loop forever
    while get_dialog( EXIT_STATE ) = MORE do

        -- handle a key
        edit_dialog()
        keyPress = get_dialog( KEY_PRESSED )

        -- select appropriate widget
        -- case get_dialog( FOCUS )

        if get_dialog( FOCUS ) = 3 then  -- file list

            if navigation_key( keyPress ) then

                -- get the cursor
                cursorAt = get_position()

                -- move the field to the test box
                set_attrib( 1, DATA, list_selected_text( get_control(3) ) )
                set_control( 1, draw_sle( get_control(1) ) )

                -- replace the cursor
                position( cursorAt[1], cursorAt[2] )

            end if

        elsif get_dialog( FOCUS ) = 4 then  -- directory list
                        
            if navigation_key( keyPress ) then

                -- get the cursor
                cursorAt = get_position()

                -- get the selected text
                tmp = list_selected_text( get_control(4) )

                -- drive?
                if tmp[1] = '[' then
                    -- rewrite drive
                    tmp = {tmp[3]} & ":\\*.*"
                else
                    -- rewrite with full path
                    if last_char( currentDir ) = '\\' then
                        tmp = currentDir & tmp & "\\*.*"
                    else
                        tmp = currentDir & "\\" & tmp & "\\*.*"
                    end if
                end if

                -- move the field to the test box
                set_attrib( 1, DATA, tmp )
                set_control( 1, draw_sle( get_control(1) ) )

                -- replace the cursor
                position( cursorAt[1], cursorAt[2] )

            end if

        end if


        -- was HELP button selected?
        if  get_dialog( FOCUS ) = 7
        and get_dialog( EXIT_STATE ) = OKAYED then

            -- help text
            message_box( "Open File", {
            "Saves and names the file you are working on.                   ",
            "                                                               ",
            " Use this to rename the current file, or name an unnamed file.",
            "                                                               ",
            " You can enter a different filename, directory and drive in   ",
            "  the Directories and Existing Files boxes, or in File Name.   ",
            "                                                               ",
            " Edit will ask if you're sure you want to replace the file if ",
            "  you select the name of an existing file.                     " } )

            -- don't exit
            set_dialog( EXIT_STATE, MORE )

        end if

        -- selection made
        if get_dialog(EXIT_STATE) = OKAYED then

            -- file/directory in text box?
            if compare( get_attrib( 1, DATA ), "" ) = 0 then

                -- no text
                message_box( "Save As", "You must specify a filename." )

                -- don't exit
                set_dialog( EXIT_STATE, MORE )

            else

                -- text entered

                -- convert to path & file
                tmp = full_path( get_attrib( 1, DATA ), currentDir )

                -- file, or new file name?                                
                if compare( tmp[3], "f" ) = 0
                or compare( tmp[3], "n" ) = 0 then

                    -- save the new path
                    currentDir = tmp[1]

                    -- save the file name
                    if last_char( currentDir ) = '\\' then
                        fileName = currentDir & tmp[2]
                    else
                        fileName = currentDir & "\\" & tmp[2]
                    end if

                elsif compare( tmp[3], "" ) = 0
                and   compare( get_attrib( 1, DATA ), "" ) != 0 then

                    -- save the file name
                    if last_char( currentDir ) = '\\' then
                        fileName = currentDir
                            & get_attrib( 1, DATA )
                    else
                        fileName = currentDir & "\\"
                            & get_attrib( 1, DATA )
                    end if

                    -- set to upper case
                    fileName = upper( fileName )


                elsif compare( tmp[3], "d") = 0 then

                    -- it's a directory

                    -- change the current directory
                    currentDir = tmp[1]

                    -- new directory
                    set_attrib( 2, REDRAW, 1 )
                    set_attrib( 2, TITLE, currentDir )

                    -- update the file list                    
                    set_attrib( 3, REDRAW, 1 )
                    set_attrib( 3, ITEMS, file_list( currentDir, "*.*" ) )

                    -- update the directory list
                    set_attrib( 4, REDRAW, 1 )
                    set_attrib( 4, ITEMS, dir_list( currentDir ) )

                    -- update the file text
                    set_attrib( 1, REDRAW, 1 )
                    set_attrib( 1, DATA, tmp[2] )

                    -- draw all the controls
                    draw_controls()

                    -- focus on the first item
                    set_dialog( FOCUS, 1 )
                    give_dialog_key( GET_FOCUS )

                    -- don't exit
                    set_dialog( EXIT_STATE, MORE)

                elsif compare( tmp[3], "m") = 0 then

                    -- it's a mask

                    -- change the current directory
                    currentDir = tmp[1]

                    -- new directory
                    set_attrib( 2, REDRAW, 1 )
                    set_attrib( 2, TITLE, currentDir )

                    -- update the file list
                    set_attrib( 3, REDRAW, 1 )
                    set_attrib( 3, ITEMS, file_list( currentDir, tmp[2] ) )


                    -- update the directory list
                    set_attrib( 4, REDRAW, 1 )
                    set_attrib( 4, ITEMS, dir_list( currentDir ) )


                    -- update the file text
                    set_attrib( 1, REDRAW, 1 )
                    set_attrib( 1, DATA, tmp[2] )


                    -- draw all the controls
                    draw_controls()

                    -- focus on the first item
                    set_dialog( FOCUS, 1 )
                    give_dialog_key( GET_FOCUS )

                    -- don't exit
                    set_dialog( EXIT_STATE, MORE )

                else

                    -- it's an error
                    message_box( "Save File", "Invalid filename or path." )

                    -- don't exit
                    set_dialog( EXIT_STATE, MORE )

                end if


            end if
        end if

    end while


    -- get exit state
    result = get_dialog( EXIT_STATE )

    -- restore screen
    hide_dialog()

    -- take action? (not cancelled)
    if result = OKAYED then

        -- does a file by that name already exist?

        -- look for file
        if not atom( dir( fileName ) ) then

            -- file exists
            result = yes_no_dialog( "Save File", "Replace existing file?" )

            -- no/cancelled?
            if result != 1 then
                return -1
            end if

        end if

    else

        -- cancelled
        return -1

    end if

    -- return file name
    return { fileName, currentDir }

end function

