-- Euphoria Editor Tools - Common Dialogs
--      Find Dialog
--      Replace Dialog
-- Find and Replace Dialogs & Functions


include dialog.e        -- dialog controls
include xle.e           -- xle routines
include d_msgbox.e      -- message box
with trace
without warning

-- used by mle_find_next()
global constant     FIND_MODE       = 1,
                    REPLACE_MODE    = 2

-----------------------------------------------------------------------------
global function find_dialog()

        integer keyPress
        object result


        -- put stuff in the box
        new_dialog( "Find", 0, 0, 0, 0, {
                "                                                ",
                "                                                ",
                "                                                ",
                "                                                ",
                "                                                ",
                "                                                ",
                "                                                ",
                "                                                ",
                "                                                ",
                "                                                " } )

        add_control( 1, SLE, "Fi&nd What:", 2, 14, 31, 0, {} )
        add_control( 2, CHECK, "Match &Whole Word Only", 4, 2, 0, 0, {} )
        add_control( 3, CHECK, "Match &Case", 5, 2, 0, 0, {} )
        add_control( 4, BUTTON, "&OK", 8, 3, 8, 0, {DEFAULT_BUTTON} )
        add_control( 5, BUTTON, "&Cancel", 8, 17, 10, 0, {CANCEL_BUTTON} )
        add_control( 6, BUTTON, "&Help", 8, 35, 8, 0, {} )

        -- default values
        set_attrib( 1, DATA, searchFor )
        set_attrib( 2, CHECKED, searchWhole )
        set_attrib( 3, CHECKED, searchCase )

    
    -- set up and display the dialog    
    show_dialog()      

    -- run the dialog
    while get_dialog( EXIT_STATE ) = MORE do    -- until the user exits
        edit_dialog()                               -- process keys
        
        if  get_dialog( FOCUS ) = 6                 -- help key?
        and get_dialog( EXIT_STATE ) = OKAYED then
            message_box( "Find", {
            "Finds text in a file.                                          ",
            "                                                               ",
            " To avoid substring matches, select 'Match Whole Word Only.'  ",
            "                                                               ",
            " To match text exactly, with the same upper and lowercase     ",
            "  letters, select 'Match Case'.                                "} )

                set_dialog( EXIT_STATE, MORE )      -- stay in loop
        end if

    end while

    -- look at exit state
    if get_dialog( EXIT_STATE ) = OKAYED then
            -- save values
            searchFor   = get_attrib( 1, DATA )
            searchWhole = get_attrib( 2, CHECKED )
            searchCase  = get_attrib( 3, CHECKED )
            replaceWith = ""
            searchCount = 0
            result = 1
    else
            result = 0
    end if

    -- remove dialog
    hide_dialog()

    return result

end function

-----------------------------------------------------------------------------
global function replace_dialog()

    sequence cursorAt, helpText
        integer keyPress
        object result

    helpText = {
        "Finds specified text and replaces it with new text.            ",
        "                                                               ",
        " To confirm each replacement, click the Replace button.       ",
        " To replace all occurrences of the text at once, click the    ",
        "  Replace All button.                                          ",
        " To find text that is an entire word and not part of a longer ",
        "  word, click Match Whole Word Only.                           ",
        " To find text with the same combination of uppercase and      ",
        "  lowercase letters as the find text, click Match Case.        ",
        "                                                               ",
        "The first occurrence of the text is selected, and you are      ",
        "prompted to either Replace this match, or to Skip this match   ",
        "without replacing it, or to Cancel the search.                 ",
        "You can also press ESC at any time to cancel the operation.    "}

        -- put stuff in the box
        new_dialog( "Replace", 0, 0, 0, 0, {
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         ",
            "                                                         " } )
    add_control( 1, SLE, "Fi&nd What:   ", 2, 18, 36, 0, {} )
    add_control( 2, SLE, "Re&place With:", 4, 18, 36, 0, {} )
    add_control( 3, CHECK, "Match &Whole Word Only", 6, 4, 0, 0, {} )
    add_control( 4, CHECK, "Match &Case", 7, 4, 0, 0, {} )
    add_control( 5, BUTTON, "&Replace", 10, 2, 11, 0, {DEFAULT_BUTTON, FOCUS} )
    add_control( 6, BUTTON, "Replace &All", 10, 16, 15, 0, {} )
    add_control( 7, BUTTON, "&Cancel", 10, 34, 10, 0, {CANCEL_BUTTON} )
    add_control( 8, BUTTON, "&Help", 10, 47, 8, 0, {} )

    -- default values
    set_attrib( 1, DATA, searchFor )
    set_attrib( 2, DATA, replaceWith )
    set_attrib( 3, CHECKED, searchWhole )
    set_attrib( 4, CHECKED, searchCase )

    -- set up and display the dialog    
    show_dialog()      

    -- run the dialog
    while get_dialog( EXIT_STATE ) = MORE do    -- until the user exits
        edit_dialog()                               -- process keys
        
        if  get_dialog( FOCUS ) = 8                 -- help key?
        and get_dialog( EXIT_STATE ) = OKAYED then
                message_box( "Replace", helpText)       -- help dialog
                set_dialog( EXIT_STATE, MORE )          -- stay in loop
        end if

    end while


    -- look at exit state   
    if get_dialog( EXIT_STATE ) = OKAYED then
    
        -- save values
        searchFor   = get_attrib( 1, DATA )
        replaceWith = get_attrib( 2, DATA )
        searchWhole = get_attrib( 3, CHECKED )
        searchCase  = get_attrib( 4, CHECKED )
        searchCount = 0

        -- what type of replace?
        if  get_dialog( FOCUS ) = 6  then result = 2    -- all
        else                              result = 1    -- single
        end if
        
    else
        result = 0
        
    end if

    -- remove dialog
    hide_dialog()

    return result

end function


-----------------------------------------------------------------------------
global function replace_occurance_dialog( integer row )

    object result

    -- create the dialog
    new_dialog( "Replace", row, 0, 45, 6, {} )

    -- define controls
    add_control( 1, LABEL,  "Replace this occurance?", 2, 12, 0, 0, {} )
    add_control( 2, BUTTON, "&Replace", 4, 3, 9, 0, {} )
    add_control( 3, BUTTON, "&Skip", 4, 17, 8, 0, {} )
    add_control( 4, BUTTON, "&Cancel", 4, 31, 10, 0, {CANCEL_BUTTON} )

    -- run dialog
    run_dialog()
    
    -- determine result    
    if    get_dialog( EXIT_STATE ) = CANCELLED  then  result = 0  -- Cancel
    elsif get_dialog( FOCUS ) = 2               then  result = 1  -- Replace
    else                                              result = 2  -- Skip
    end if

    -- remove dialog
    hide_dialog()

    return result
                                                
end function


-----------------------------------------------------------------------------
function mid( sequence s, integer n )

    -- return a slice of s, from n to the end of s

    if length(s) = 0    then return ""
    elsif n = 0         then return ""
    elsif n > length(s) then return ""
    else                     return s[n..length(s)]
    end if

end function


-----------------------------------------------------------------------------
function word_boundary( integer n )

    -- returns true is n is a word boundary character

    -- case n
    if    n >= 'A' and n <= 'Z' then return 0
    elsif n >= 'a' and n <= 'z' then return 0
    elsif n >= '0' and n <= '9' then return 0
    else                             return 1
    end if

end function

-----------------------------------------------------------------------------
global function mle_find_next( sequence mle, integer mode )

        -- find next occurance of searchFor in mle

        integer thisCol, thisLine, matchLine, matchCol, tmp, goodMatch
        sequence text, key
        -- alias mle

        -- skip if blank
        if compare( searchFor, "" ) = 0 then
                return mle
        end if

        -- starting line and column
        thisLine =  mle[TOP_LINE] + mle[ITEM] - 1
        thisCol = mle[FIRST_CHAR] + mle[CURSOR] - 1
        text = mle[ITEMS][thisLine]

        -- on selected item already?
        if  mle[ANCHOR_COL] != 0 then   -- something selected

            -- already on match? use case insensitive match
            if match( upper( searchFor ),
                    upper( mid( text, thisCol ) ) ) = 1 then

                -- move ahead
                thisCol = thisCol + 1      -- go to next column

                -- past end of line?
                if thisCol > length( text ) then
                    thisLine = thisLine + 1
                    thisCol = 1
                end if

            end if

        end if    

        -- save some conversion
        if searchCase then
                key = searchFor
        else
                key = upper( searchFor )        
        end if


        -- look for match

        matchLine = 0    -- clear flag
        while thisLine <= length( mle[ITEMS] ) do       -- examine lines

        -- status line
        position( SCREEN_DEEP, SCREEN_WIDE-34 )
        put_colors( OF_STATUS, 
                sprintf("  Line:%g   ",  {thisLine} ) )

        -- get text
        text = mid( mle[ITEMS][thisLine], thisCol )

        -- case sensitive?
        if not searchCase then
                text = upper( text )
        end if

        -- do match
         matchCol = match( key, text )


        -- add offset
        if matchCol = 0 then
            goodMatch = 0
        else
            goodMatch = 1
            matchCol = matchCol + thisCol - 1            
        end if

            -- word check
            if  matchCol != 0       -- match
            and searchWhole then    -- whole word match

                -- check left boundary
                if matchCol = 1 then
                    -- left boundary fine
                    
                elsif not word_boundary( mle[ITEMS][thisLine][matchCol-1] ) 
                then
                    -- no match
                    thisCol = matchCol + 1
                    goodMatch = 0

                end if

                -- check right boundary
                tmp = matchCol+length( searchFor )
                if tmp > length( mle[ITEMS][thisLine] ) then
                        -- right boundary fine

                elsif not word_boundary( mle[ITEMS][thisLine][tmp] ) 
                then
                    -- no match
                    thisCol = matchCol + 1
                    goodMatch = 0
                                
                end if

            end if

            -- exit on a match
            if goodMatch then
                matchLine = thisLine
                exit
            end if

            -- move ahead?
            if thisCol > length(text)   -- end of current line
            or matchCol = 0 then        -- no matches on this line
                    
                thisLine = thisLine + 1     -- go to next line
                thisCol = 1                 -- beginning column
            
            end if

        end while


        -- match?
        if not goodMatch then

            -- no match

            if searchCount = 0
            and mode = FIND_MODE then
                -- find, no matches
                message_box( "Find", "Find was unable to find a match." )

            elsif searchCount = 0
            and mode = REPLACE_MODE then
                -- replace, no matches
                message_box( "Replace", "Replace was unable to find a match." )

            elsif  mode = REPLACE_MODE then
                -- replace, matches
                message_box( "Replace", 
                        sprintf( "%g replacements.", {searchCount} ) )

            else
                -- find, no more matches
                message_box( "Find", "No more matches found." )

            end if

            -- remove selection
            mle[ANCHOR_COL] = 0

            -- return mle
            return mle

        else

            -- match

            -- increment search count
            searchCount = searchCount + 1

            -- get the line on the screen
            while matchLine > mle[TOP_LINE]+mle[DEEP]-1 do
                -- page down
                mle[TOP_LINE] = mle [ TOP_LINE]+ mle[DEEP]
            end while

            -- focus on the line
            mle[ITEM] = matchLine - mle[TOP_LINE] + 1

            -- move to left, if on left side
            mle[CURSOR] = 1
            if matchCol < mle[FIRST_CHAR] then
                mle[FIRST_CHAR] = 1
            end if

            -- move to right, until on screen
            while matchCol > mle[FIRST_CHAR]+mle[WIDE]-1 do
                mle[FIRST_CHAR] = mle [ FIRST_CHAR]+ mle[WIDE]
            end while

            -- focus on column
            mle[CURSOR] = matchCol - mle[FIRST_CHAR] + 1

            -- select text
            mle[ANCHOR_LINE] = matchLine
            mle[ANCHOR_COL] = matchCol + length( searchFor )

            -- display?
            if mode = FIND_MODE then
                return edit_xle( mle, REFRESH )
            else
                return mle
            end if

        end if


end function


-----------------------------------------------------------------------------
global function mle_find( sequence mle )

    -- similar to mle_find_next, but brings up dialog

    if find_dialog() then
        -- do match
        return mle_find_next( mle, FIND_MODE )

    end if

    return mle

end function


-----------------------------------------------------------------------------
global function mle_replace( sequence mle )

    -- get values to replace from user
    -- search through the MLE, displaying matches
    -- until no more matches or user exit

    integer result, row
    integer goBack, topLine, item, cursor, firstChar
    
    -- save the current values
    topLine = mle[TOP_LINE]
    item = mle[ITEM]
    cursor = mle[CURSOR]
    firstChar = mle[FIRST_CHAR]

    -- go back is false
    goBack = 0

    -- get values from user
    result = replace_dialog()

    if result = 1 then

        -- selective replace
        while 1 do

            -- do match
            mle = mle_find_next( mle, REPLACE_MODE )

            -- match?
            if mle[ANCHOR_COL] then

                -- display match
                mle = draw_mle_items( mle )

                -- position of dialog
                if mle[ROW]+mle[ITEM] > SCREEN_DEEP / 2 then
                    row = floor( SCREEN_DEEP / 4 ) - 3
                else
                    row = floor( 3 * SCREEN_DEEP / 4 ) - 3
                end if

                -- display dialog
                result = replace_occurance_dialog( row )
                
                if  result = 1 then
                    -- yes
                    mle = xle_cut_text( mle )
                    mle = xle_paste_text( mle, replaceWith )

                elsif result = 2 then
                    -- skip
                    
                else
                    -- cancel
                    exit
                end if
                
            else
                -- leave loop
                goBack = 1
                exit
                
            end if

        end while

    elsif result = 2 then

        -- replace all

        while 1 do

            -- do match
            mle = mle_find_next( mle, REPLACE_MODE )

            -- match?
            if mle[ANCHOR_COL] then
                -- replace
                mle = xle_cut_text( mle )
                mle = xle_paste_text( mle, replaceWith )
            else
                goBack = 1
                exit
            end if
            
        end while
                    
    end if

    -- return to start if goBack was set
    if goBack then

            -- go to line number
            mle[TOP_LINE] = topLine
            mle[ITEM] = item
            mle[CURSOR] = cursor
            mle[FIRST_CHAR] = firstChar
            mle[ANCHOR_COL] = 0

    end if
    
    -- display mle
    return edit_xle( mle, REFRESH )

end function


