-- Settings dialog
-- Euphoria Editor

include dialog.e

-----------------------------------------------------------------------------
global procedure settings_dialog()

    sequence cursorAt
    integer keyPress, lpp
    object result

    -- find the current lines per page
    result = video_config()
    lpp = result[VC_LINES]
    
    -- put stuff in the box
    new_dialog( "Settings", 0, 0, 0, 0, {
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                ",
        "                                                " } )


    add_control( 1, SLE, "&Tab Stops:   ", 2, 20, 9, 0, {} )
    add_control( 2, LABEL, "&Printer Port:", 4, 5, 0, 0, {} )
    add_control( 3, RADIO, "LPT", 4, 19, 0, 0, {CHECKED} )
    add_control( 4, RADIO, "COM1", 5, 19, 0, 0, {} )
    add_control( 5, RADIO, "COM2", 6, 19, 0, 0, {} )
    add_control( 6, LABEL, "&Lines:", 8, 12, 0, 0, {} )
    add_control( 7, RADIO, "25", 8, 19, 0, 0, {} )
    add_control( 8, RADIO, "28", 9, 19, 0, 0, {} )
    add_control( 9, RADIO, "43", 8, 32, 0, 0, {} )
    add_control( 10, RADIO, "50", 9, 32, 0, 0, {} )
    add_control( 11, BUTTON, "&OK", 12, 5, 8, 0, {DEFAULT_BUTTON, FOCUS} )
    add_control( 12, BUTTON, "&Cancel", 12, 18, 10, 0, {CANCEL_BUTTON} )
    add_control( 13, BUTTON, "&Help", 12, 34, 8, 0, {} )

    -- set tab size
    set_attrib( 1, DATA, sprintf( "%d", TAB_SIZE ) )

    -- set printer port to control group #1
        set_attrib( 3, GROUP, 1 )
        set_attrib( 4, GROUP, 1 )
        set_attrib( 5, GROUP, 1 )

    -- check appropriate buttons for printer port
        set_attrib( 3, CHECKED, ( compare( PRINTER_PORT, "LPT" ) = 0 ) )
        set_attrib( 4, CHECKED, ( compare( PRINTER_PORT, "COM1" ) = 0 ) )
        set_attrib( 5, CHECKED, ( compare( PRINTER_PORT, "COM2" ) = 0 ) )


    -- set screen lines to control group #2
        set_attrib( 7, GROUP, 2 )
        set_attrib( 8, GROUP, 2 )
        set_attrib( 9, GROUP, 2 )
        set_attrib( 10, GROUP, 2 )

    -- can't set in VGA mode
        set_attrib( 7, DIM, VGA_MODE )
        set_attrib( 8, DIM, VGA_MODE )
        set_attrib( 9, DIM, VGA_MODE )
        set_attrib( 10, DIM, VGA_MODE )
        
    -- check button for lines per page
        set_attrib( 7, CHECKED, ( lpp = 25 ) )
        set_attrib( 8, CHECKED, ( lpp = 28 ) )
        set_attrib( 9, CHECKED, ( lpp = 43 ) )
        set_attrib( 10, CHECKED, ( lpp = 50 ) )

    -- set up and display the dialog    
    show_dialog()      

    -- run the dialog
    while get_dialog( EXIT_STATE ) = MORE do    -- until the user exits
        edit_dialog()                           -- process keys
        
        if  get_dialog( FOCUS ) = 13                -- help key?
        and get_dialog( EXIT_STATE ) = OKAYED then
            message_box( "Settings", {
            "Determines settings for Edit.                                  ",
            "                                                               ",
            " 'Tab Stops' sets the distance between tab stops.             ",
            "                                                               ",
            " 'Printer Port' is used by File Print to determine where to   ",
            "   send the file to be printed.                                ",
            "                                                               ",
            " 'Lines' sets the number of lines to disply on the screen.    ",
            "  This feature is only available in Text mode.                 "} )

                set_dialog( EXIT_STATE, MORE )          -- stay in loop
        end if

    end while


    -- check for change in tab size
    result = get_attrib( 1, DATA )
    -- non-blank?
    if length( result ) > 0 then
        -- attempt to convert to a number
        result = value( result )
        if result[1] = GET_SUCCESS then
            -- must be an integer
            if integer( result[2] ) then
                -- must be between 1 and 256
                if result[2] > 0 
                and result[2] < 256 then    
                    TAB_SIZE = result[2]
                end if
            end if
        end if        
    end if

    -- check for a change in resolution
    if get_attrib( 7, CHECKED ) and lpp != 25 then
        -- new res is 25    
        lpp = 25

    elsif get_attrib( 8, CHECKED ) and lpp != 28 then
        -- new res is 28
        lpp = 28

    elsif get_attrib( 9, CHECKED ) and lpp != 43 then
        -- new res is 43
        lpp = 43

    elsif get_attrib( 10, CHECKED ) and lpp != 50 then
        -- new res is 50
        lpp = 50
    else
        lpp = 0
    end if


    -- check for value of printer port
    if    get_attrib( 3, CHECKED )      then PRINTER_PORT = "LPT"
    elsif get_attrib( 4, CHECKED )      then PRINTER_PORT = "COM1"
    elsif get_attrib( 5, CHECKED )      then PRINTER_PORT = "COM2"
    end if


    -- restore screen
    hide_dialog()

    -- change res
    if lpp != 0 then

        -- change res
        lpp = text_rows( lpp )
        result = video_config()
        SCREEN_WIDE = result[VC_COLUMNS]
        SCREEN_DEEP = result[VC_LINES]

        -- for each file being edited
        for i = 1 to length( edited ) do

            -- change mle edit window
            edited[i][DEEP] = SCREEN_DEEP-3
            edited[i][WIDE] = SCREEN_WIDE-2
    
            -- make sure elevator isn't off screen
            edited[i][ELEVATOR] = 0
            
            -- move current list
            if edited[i][ITEM] > edited[i][DEEP] then
                edited[i][TOP_LINE] = edited[i][ITEM] - edited[i][DEEP]
                edited[i][ITEM] = edited[i][DEEP]
            end if

        end for

        -- activate intense colors again
        toggle_blinking( 0 )

        -- re-activate the mouse
        mouse_pointer( 1 )    

        -- refresh
        redraw_flag = 1
            
    end if

end procedure

