-- Win95-like DOS Editor Menu & Dialogs
-- handles multiple files

without warning
without trace

-- generic libraries
    include files.e                     -- file utilities
    include get.e                       -- value function

-- generic dialogs
    include dialog.e                    -- controls
    include menu.e                      -- menu
    include d_yesno.e                   -- ask user yes/no question
    include d_msgbox.e                  -- message box
    include d_ask.e                     -- ask user dialog
    include d_help.e                    -- display help
    include d_open.e                    -- open file
    include d_saveas.e                  -- save as dialog

global sequence dialog              -- global dialog
global sequence currentDir          -- current directory
global integer fNum                 -- index of current file open
global sequence edited              -- contents of files being edited
global integer redraw_flag          -- flag to redraw entire window

-- command line arguments
sequence commandLine                -- command line args
    commandLine = ""

-- output screen
sequence userScreen                 -- run program output screen
sequence userVC                     -- video state

-- status message
sequence statusMessage              -- default status message
    statusMessage = "Euphoria Editor  ALT activates menu"

-- help file contents
sequence help                       -- built with create_help()
    help = {}

-- euphoria file extensions
constant EU_EXTENSIONS = { ".EX", ".E", ".EXW", ".EW", ".PP", ".P",
                                ".GTK", ".WIN" }

-- overwrite default color
    OF_TEXT = { WHITE, BLUE }

-- printer port
    global sequence  PRINTER_PORT
    PRINTER_PORT = "COM1"

-- other
    include eprint.e                    -- printing
    include ini.e                       -- initialization file  


-- macros
integer
    macroState,             -- state of the macro
    macroAt                 -- current key being played in the macro
    
sequence
    macroKeys               -- recording of macro

constant
    Off         = 0,            -- macro not running
    Play        = 1,            -- macro playing
    Record      = 2,            -- macro recording
    SHIFT_F12   = 392,          -- append to macro
    CTRL_F12    = 394           -- record macro



-----------------------------------------------------------------------------
global procedure display_pp_line_number( integer l )

    -- used by pre-processor to show current line number

    -- status line
    position( SCREEN_DEEP, SCREEN_WIDE-26 )
    put_colors( OF_STATUS, sprintf("%g   ",  {l} ) )

end procedure



-- Reverse to include/exclude Pre-processor option.
-- include no_pp.e
-- include pp.e
include dot_pp.e


-----------------------------------------------------------------------------

-- Menu Bar

    -- File
    m_bar( 1, "&File" )
    m_item( 1, 1, "&New            ", "Creates a new file" )
    m_item( 1, 2, "&Open           ", "Loads an existing file from disk" )
    m_item( 1, 3, "&Save           ", "Saves current file" )
    m_item( 1, 4, "Save &As...     ", "Saves current file with new name" )
    m_item( 1, 5, "&Close          ", "Closes current file"  )
    m_item( 1, 6, "-", "" )
    m_item( 1, 7, "&Print...       ", ""  )
    m_item( 1, 8, "-", "" )
    m_item( 1, 9, "E&xit           ", "Exits the editor" )

    -- Edit
    m_bar( 2, "&Edit" )
    m_item( 2, 1, "Cu&t       Ctrl+X", "Deletes selected text and copies it to buffer" )
    m_item( 2, 2, "&Copy      Ctrl+C", "Copies selected text to buffer" )
    m_item( 2, 3, "&Paste     Ctrl+V", "Inserts buffer contents at current location" )
    m_item( 2, 4, "Cl&ear     Del   ", "Deletes selected text without copying it to buffer")

    -- Search
    m_bar( 3, "&Search" )
    m_item( 3, 1, "&Find...               ", "Finds specified text" )
    m_item( 3, 2, "Repeat &Last Find    F3", "Finds next occurance of text specified in previous search" )
    m_item( 3, 3, "&Replace...            ", "Finds and changes specified text" )

    -- View
    m_bar( 4, "&View" )
    m_item( 4, 1, "&Subs                F2", "Displays routines in program" )
    m_item( 4, 2, "&Goto Line...          ", "Jump to line number" )
    m_item( 4, 3, "Goto &Error            ", "Jump to line with error" )
    m_item( 4, 4, "&User Screen         F4", "Displays output screen" )
    m_item( 4, 5, "-", "" )
    m_item( 4, 6, "&1 Untitled      Alt+1", "Selects a file to view" )

    -- Run
    m_bar( 5, "&Run" )
    m_item( 5,  1, "&Execute               F5", "Runs current file" )
    m_item( 5,  2, "&Pre-processor           ", "Convert .PP/.P files to .EX/.E" )
    m_item( 5,  3, "&Bind File               ", "Bind current file into a .EXE file" )
    m_item( 5,  4, "-", "" )
    m_item( 5,  5, "&Command Line...         ", "Set command line arguments" )
    m_item( 5,  6, "-", "" )
    m_item( 5,  7, "Play &Macro           F12", "Run a recorded macro." )
    m_item( 5,  8, "&Record Macro    Ctrl+F12", "Start/Stop recording keystrokes as macro." )
    m_item( 5,  9, "&Append Macro   Shift+F12", "Append keystrokes to existing macro." )
    m_item( 5, 10, "-", "" )    
    m_item( 5, 11, "&DOS Prompt              ", "Shell to DOS prompt" )

    -- Options
    m_bar( 6, "&Options" )
    m_item( 6, 1, "&Settings...       ", "Changes editor settings" )
    m_item( 6, 2, "&Colors...         ", "Changes editor screen colors" )
    m_item( 6, 3, "-", "" )
    m_item( 6, 4, "S&yntax Coloring   ", "Turns syntax coloring on and off" )
    m_item( 6, 5, "&Auto Complete     ", "Turns code auto completion on and off" )

    -- Help
    m_bar( 7, "&Help" )
    m_item( 7, 1, "&Commands...      ", "Help on edit commands" )
    m_item( 7, 2, "A&SCII Chart      ", "Display ASCII Codes" )
    m_item( 7, 3, "-", "" )    
    m_item( 7, 4, "&Reference Manual ", "Load REFMAN.DOC" )
    m_item( 7, 5, "&Library          ", "Load LIBRARY.DOC" )
    m_item( 7, 6, "-", "" )    
    m_item( 7, 7, "&About...         ", "About Editor" )

----------------------------------------------------------------------------
procedure status_message( sequence s )

    -- place a message in the status line

    -- clear status line
    position( SCREEN_DEEP, 1 )
    put_colors( OF_STATUS, repeat(  ' ', SCREEN_WIDE ) )

    -- message
    position( SCREEN_DEEP, 2 )
    put_colors( OF_STATUS, s )

end procedure

----------------------------------------------------------------------------
global function pad( sequence s, integer n )
    
    -- return a string padded to a given length

    if length( s ) > n then
        return s[1..n]
    else
        return s & repeat(  ' ', n - length(s) )
    end if

end function


----------------------------------------------------------------------------
global function full_path_name( sequence file )

    -- appends file name to current directory

    -- append the path to the file name
    if last_char( currentDir ) = '\\' then
        return currentDir & file
    else
        return currentDir & "\\" & file
    end if

end function

-----------------------------------------------------------------------------
function strip_path( sequence s )

    -- strip the path from the file name

    integer n

    while 1 do

        -- \ divides nodes
        n = find( '\\', s )

        -- no more nodes?
        if n = 0 then
            exit
        else
            -- remove node
            s = s[n+1..length(s)]
        end if

    end while

    return s

end function

-----------------------------------------------------------------------------
procedure build_file_list()       

    -- rebuild the file list in View

    integer altKey
    sequence shortName

    -- remove prior items
    MAIN_MENU[4][ITEMS] = MAIN_MENU[4][ITEMS][1..6]
    
    -- change the text
    for i = 1 to length( edited ) do
        
        -- get short file name
        if length( edited[i][TITLE] ) = 0 then
            shortName = "Untitled"
        else
            shortName = strip_path( edited[i][TITLE] )
        end if
        
        -- get alt key number
        if i = 10 then
            altKey = 0
        else
            altKey = i
        end if

        -- clone #6 in View menu
        if i > 1 then
            MAIN_MENU[4][ITEMS] = append( MAIN_MENU[4][ITEMS],
                MAIN_MENU[4][ITEMS][6] )
        end if            
        
        -- change text
        MAIN_MENU[4][ITEMS][5+i][TITLE] = sprintf( "%g %s Alt+%g", 
            {altKey, pad( shortName, 14 ), altKey } )
        
    
    end for

end procedure


-----------------------------------------------------------------------------
procedure open_untitled_file()

    -- create an initialized mle and append it to the files list
    -- also, make it the current file

    sequence mle

    mle = new_control( MLE )     -- create the MLE

    mle[FIRST_CHAR] = 1           -- first column displayed
    mle[CURSOR] = 1               -- cursor position
    mle[ROW] = 3                  -- row of MLE
    mle[COL] = 2                  -- column of MLE
    mle[DEEP] = SCREEN_DEEP-3     -- depth, less menu/border/status
    mle[WIDE] = SCREEN_WIDE-2     -- size of screen, less border
    mle[ITEM] = 1                 -- relative position of line edited
    mle[TOP_LINE] = 1             -- first line displayed
    mle[ELEVATOR] = 0             -- elevator on scroll bar
    mle[ITEMS] = { "" }           -- empty file
    mle[COLOR_SYNTAX] = 0         -- colored syntax
    mle[AUTOCOMPLETE] = 0         -- autocompletion
    mle[EXIT_STATE] = MORE        -- exit state flag
    mle[TITLE] = ""               -- file name

    -- append it
    edited = append( edited, mle )

    -- make current
    fNum = length( edited )

end procedure


----------------------------------------------------------------------------
function max_files_open()

    -- return true if too many files are open

    -- more than 10 file open?
    if length( edited ) > 9 then
        -- should cause an error
        message_box( "Editor", "Only 10 files can be open at a time." )
        return 1
    else
        return 0
    end if

end function


----------------------------------------------------------------------------
global procedure read_file( sequence file )

    object line
    sequence buffer, fullName, text
    integer file_num, index

    -- can another file be opened?
    if max_files_open() then
        return
    end if

    -- get full path name
    fullName = full_path_name( file )
    
    -- open the file
    file_num = open( fullName, "r")
    if file_num = -1 then
        -- should cause an error
        message_box( "Editor", "Unable to open \"" & fullName & "\"." )
        return
    end if

    -- is the file already open?
    for i = 1 to length( edited ) do
        -- name match?
        if compare( edited[i][TITLE], fullName ) = 0 then
            -- make current
            fNum = i

            -- redraw flag
            redraw_flag = 1

            -- exit
            return

        end if
    end for

    -- status line
    status_message( "Reading " & fullName )

    -- read a text file into a sequence
    hide_cursor()
    buffer = {}

    -- status line
    position( SCREEN_DEEP, SCREEN_WIDE-34 )
    put_colors( OF_STATUS, "  Loading File..." )

    while 1 do
        
        -- status line
--        position( SCREEN_DEEP, SCREEN_WIDE-26 )
--        put_colors( OF_STATUS, 
--            sprintf("%g   ",  {length( buffer)} ) )

        line = gets(file_num)
        if atom(line) then
            exit   -- end of file
        end if

        buffer = append( buffer, detab_text( line ) )

    end while

    -- close the file
    close( file_num )

    -- check for an empty file. this happens with length 0 files
    if length(buffer) = 0 then       
        -- add an empty line
        buffer = append( buffer, "" )
    end if

    -- open a file
    open_untitled_file()

    -- save full name
    edited[fNum][TITLE] = fullName
    
    -- save information
    edited[fNum][ITEMS] = buffer

    -- add to list
    build_file_list()

    -- syntax coloring?
    for i = 1 to length( EU_EXTENSIONS ) do
        if match( EU_EXTENSIONS[i], upper( file ) ) then
            -- use colored syntax
            edited[fNum][COLOR_SYNTAX] = 1
            
            -- use auto complete
            edited[fNum][AUTOCOMPLETE] = 1            
            
            -- exit loop
            exit
        end if
    end for


    -- redraw flag
    redraw_flag = 1

end procedure

----------------------------------------------------------------------------
global function write_file()

    -- write out current file
    -- return -1 if an error happens

    integer handle

    -- is there a file name to write?
    if length( edited[fNum][TITLE] ) = 0 then
        message_box( "Editor", "File must be named first." )
        return -1
    end if

    -- open the file
    handle = open( edited[fNum][TITLE], "w" )
    if handle = -1 then
        message_box( "Editor", "Unable to open file." )
        return -1
    end if

    -- status line
    status_message( "Writing " & edited[fNum][TITLE] )

    -- write file out
    hide_cursor()

    -- status line
--    position( SCREEN_DEEP, SCREEN_WIDE-34 )
--    put_colors( OF_STATUS, "  Line:    " )

    -- status line
    position( SCREEN_DEEP, SCREEN_WIDE-34 )
    put_colors( OF_STATUS, "  Saving File..." )


    for line = 1 to length(edited[fNum][ITEMS]) do
        
        -- status line
--        position( SCREEN_DEEP, SCREEN_WIDE-26 )
--        put_colors( OF_STATUS, 
--            sprintf("%g   ",  {line} ) )

        -- write to file
        -- puts( handle, retab_text( line ) )
        puts( handle, edited[fNum][ITEMS][line] & "\n" )

    end for

    -- close the file
    close( handle )

    -- no changes to save
    edited[fNum][DIRTY] = 0

    -- status line
    status_message( statusMessage )

    -- return good
    return 0

end function


-----------------------------------------------------------------------------
function save_dirty_file( sequence title )

    -- check if file should be saved
    -- returns 1 to continue, 0 to cancel process

    sequence fName
    object result

    -- skip if not dirty
    if not edited[fNum][DIRTY] then
        return 1
    end if

    -- file name
    if length( edited[fNum][TITLE] ) = 0 then
        fName = "UNTITLED"

    else
        fName = edited[fNum][TITLE]

    end if

    -- ask user to save
    result = yes_no_dialog( title,
        { "The file", fName, "has not been saved. Save it now?" } )

    if result = 1 then

        -- does the file have a name?
        if length( edited[fNum][TITLE] ) = 0 then
            -- try to save it
            result = save_as_dialog( strip_path( edited[fNum][TITLE] ), 
                                        currentDir )
            if atom( result ) then

                -- no save
                return 0

            else

                -- save name
                edited[fNum][TITLE] = result[1]
                currentDir = result[2]

                -- save_file
                result = write_file()

                -- need to put up new file name    
                redraw_flag = 1

                -- saved
                return 1

            end if

        else

            -- save it
            result = write_file()
            return 1

        end if

    elsif result = 0 then

        -- don't save it
        return 1

    else

        -- cancel
        return 0

    end if

end function



-----------------------------------------------------------------------------
global procedure editor_new_file()

    -- can another file be opened?
    if max_files_open() then
        return
    end if

    -- open an empty file
    open_untitled_file()
    
    -- add to list
    build_file_list()  
    
    -- redraw flag
    redraw_flag = 1    
    
end procedure
    

-----------------------------------------------------------------------------
global procedure editor_open_file()

    -- open a file

    object result

    -- can another file be opened?
    if max_files_open() then
        return
    end if

    -- get new file
    result = open_file_dialog( currentDir )
    if atom( result ) then
        -- fail
        return
    else
        currentDir = result[2]
        read_file( result[1] )
    end if

end procedure


-----------------------------------------------------------------------------
procedure select_file_number( integer index )

    -- make requested file current

    -- in range?
    if index <= length( edited ) then
        -- change current file
        fNum = index

        -- refresh
        redraw_flag = 1    
    
    end if
        

end procedure

-----------------------------------------------------------------------------
global procedure editor_save_as()

    -- save current file under new name

    object result

    result = save_as_dialog( strip_path( edited[fNum][TITLE] ), currentDir )
    if atom( result ) then
        -- cancelled
        return
    else
        -- already open?
        for i = 1 to length( edited ) do
            -- name match?
            if compare( edited[i][TITLE], result[1] ) = 0 then

                -- error message
                message_box( "Editor", { "The file", result[1],
                    "is already open." } )

                -- leave
                return

            end if
        end for

        -- name and dir
        edited[fNum][TITLE] = result[1]
        currentDir = result[2]
        result = write_file()

        -- update the file list
        build_file_list()

        -- show new file name
        redraw_flag = 1

    end if

end procedure


-----------------------------------------------------------------------------
global procedure editor_save_file()

    -- save current file under current name

    object result

    if length( edited[fNum][TITLE] ) = 0 then
        -- save as
        result = save_as_dialog( strip_path( edited[fNum][TITLE] ), currentDir )
        return

    else
        -- write it
        result = write_file()
    end if

end procedure


-----------------------------------------------------------------------------
global procedure editor_close()

    -- close current file

    if save_dirty_file( "Close File" ) = 1 then

        -- remove from edited
        edited = edited[1..fNum-1] & edited[fNum+1..length(edited)]

        -- current file is #1
        fNum = 1

        -- if empty, add a file
        if length( edited ) = 0 then
            open_untitled_file()
        end if

        -- add to list
        build_file_list()
        
        -- redraw flag
        redraw_flag = 1
        
    end if

end procedure


-----------------------------------------------------------------------------
global procedure editor_print()

    integer keyPress, cursorAt
    sequence portLabel

    -- port label
    portLabel = "Print to the printer on " & PRINTER_PORT

    -- put stuff in the box
    new_dialog( "Print", 0, 0, 42, 12, {})    
    add_control( 1, LABEL, portLabel, 2, 2, 0, 0, {} )
    add_control( 2, RADIO, "&Selected Text Only", 4, 4, 0, 0, {DIM} )
    add_control( 3, RADIO, "&Complete Document", 5, 4, 0, 0, {CHECKED} )

    add_control( 4, CHECK, "&Syntax Coloring", 7, 4, 0, 0, {} )
    add_control( 5, CHECK, "Co&ndensed Text", 8, 4, 0, 0, {} )
    
    add_control( 6, BUTTON, "OK", 10, 4, 8, 0, {DEFAULT_BUTTON} )
    add_control( 7, BUTTON, "&Cancel", 10, 16, 10, 0, {CANCEL_BUTTON} )
    add_control( 8, BUTTON, "&Help", 10, 30, 8, 0, {} )

    -- syntax coloring?
    set_attrib( 4, CHECKED, edited[fNum][COLOR_SYNTAX] )

    -- dialog edit loop

    show_dialog()      -- set up and display the dialog
    while get_dialog( EXIT_STATE ) = MORE do    -- until the user exits
        edit_dialog()                           -- process keys
        
        if  get_dialog( FOCUS ) = 8                 -- help key?
        and get_dialog( EXIT_STATE ) = OKAYED then

            message_box( "Print", {
            "Prints the selected file you are working on.                   ",
            "                                                               ",
            " To print a selection of the file, highlight the portion and  ",
            "  select 'Selected Text Only'.                                 ",
            "                                                               ",        
            " To print the whole file, select 'Complete Document'.         ",
            "                                                               ",
            " Select 'Syntax Coloring' to color code Euphoria keywords.    ",
            "                                                               ",
            " Select Condensed Text to print 2 pages per document.         "} )

            set_dialog( EXIT_STATE, MORE )          -- stay in loop

        end if

    end while

    -- determine result        
    if  get_dialog( FOCUS ) = 6                     -- OK
    and get_dialog( EXIT_STATE ) = OKAYED then      -- pressed
        
        -- print the file
        eprint( edited[fNum][ITEMS],    -- file
            get_attrib( 4, CHECKED ),   -- syntax coloring
            get_attrib( 5, CHECKED ) )  -- condensed

    end if             


    -- restore screen
    hide_dialog()

end procedure


-----------------------------------------------------------------------------
procedure editor_exit()

    -- save all the files an exit
    -- if the user cancels, return

    integer displayed
    object result
    
    
    -- remember which file is on the screen
    displayed = fNum
    
    -- save the open files
    for i = 1 to length( edited ) do

        -- make current
        fNum = i

        -- save it?
        if not save_dirty_file( "Exit" ) then
            
            -- cancelled. 
            -- need to redraw screen if on different file
            if displayed != fNum then

                -- redraw flag
                redraw_flag = 1

            end if

            -- return from function
            return

        end if

    end for

    -- restore screen
        mouse_pointer(0)
        bk_color( BLACK )
        text_color( WHITE )
        clear_screen()
        result = text_rows( 25 )
        result = graphics_mode( -1 )
        cursor( UNDERLINE_CURSOR )
        toggle_blinking( 1 )

    -- write ini file
        write_ini_file()

    -- exit
        abort( 0 )

end procedure

-----------------------------------------------------------------------------
global procedure goto_line()

    -- get a line number, and jump to it

    sequence s
    object result


    result = ask_user( "Goto Line", "Enter Line Number", "" )
    if not atom( result ) then

        -- convert to a number
        s = value( result )

        if s[1] != GET_SUCCESS then

            -- huh, buddy?
            message_box( "Edit", "Invalid Line Number." )

        elsif s[2] > length( edited[fNum][ITEMS] ) then

            -- huh, buddy?
            message_box( "Edit", "Line Past End of File." )

        else

            -- go to line number
            edited[fNum][TOP_LINE] = s[2]
            edited[fNum][ITEM] = 1

            -- remove anchor, if any
            edited[fNum][ANCHOR_COL] = 0

            -- col 1
            edited[fNum][CURSOR] = 1
            edited[fNum][FIRST_CHAR] = 1

            -- refresh
            redraw_flag = 1

        end if

    end if

end procedure


-----------------------------------------------------------------------------
function instr( integer n, sequence s1, sequence s2 )

    -- return position of s1 in s2, starting at position n

    integer at

    at = match( s1, s2[n..length(s2)] )
    if at != 0 then
        return at + n - 1
    else
        return 0
    end if

end function

-----------------------------------------------------------------------------
function read_error_file()

    -- read ex.err, returns:
    --  { file, message, line, col }
    --  or -1 if can't read file
    
    integer handle, eStart, eEnd, errLine, errCol
    sequence text, errFile, errText
    object result

    -- try to open the error file
    handle = open( "ex.err", "r" )

    -- does it exist?
    if handle = -1 then

        -- no error file
        return -1

    else

        -- read first 4 lines
        text = {}
        for i = 1 to 4 do
            text = append( text, gets( handle ) )
        end for

        -- close the file
        close( handle )

    end if

    -- get the first line, with the error line

    -- the line number is "...:nnnn "
    -- the 3 skips past the colon in the drive
    eStart = instr( 3, ":", text[1] )

    -- check the value
    if eStart = 0 then
        -- can't read file; format is bad
        return -1
    end if

    -- file name comes before colon
        errFile = text[1][1..eStart-1]

    -- value to left of space following colon...
    eEnd = instr( eStart, " ", text[1] )

    -- ...or line feed following colon
    if eEnd = 0 then
        eEnd = instr( eStart, {10}, text[1] )
    end if

    -- found end?
    if eEnd = 0 then
        -- can't read file; format is bad
        return -1
    end if

    -- convert to number
    result = value( text[1][eStart+1..eEnd-1] )
    if result[1] != GET_SUCCESS then
        -- can't read number, format is bad
        return -1
    else
        errLine = result[2]
    end if

    -- second line is error message
    errText = text[2][1..length(text[2])-1]

    -- the fourth line points to the position
    errCol = find( '^', text[4] )

    -- return values
    return { errFile, errText, errLine, errCol }

end function


-----------------------------------------------------------------------------
procedure goto_error()

    -- jump to line with error, show error text

    integer keyPress, dotAt
    sequence cursorAt, savedScreen, fileName, extension
    object err

    -- alias edited[fNum]

    -- read error file
        err = read_error_file()
        if atom( err ) then
            -- can't read file
            return
        end if

    -- convert .PP and .P into .EX and .E
    fileName = upper( edited[fNum][TITLE] )
    
    -- split into extension, if any
    dotAt = match( ".", fileName )
    if dotAt != 0 then
        extension = fileName[dotAt..length(fileName)]
        fileName = fileName[1..dotAt-1]
    else
        extension = ""
    end if

    -- convert preprocessor extensions        
    if compare( ".PP", extension ) = 0 then
        fileName = fileName & ".EX"
    elsif compare( ".P", extension ) = 0 then
        fileName = fileName & ".E"
    else
        fileName = fileName & extension
    end if

    -- same file name?
        if compare( fileName, upper( err[1] ) ) != 0 then
            -- try it with the path attached
            if compare( fileName,
                upper( full_path_name( err[1] ) ) ) != 0 then
            
                -- must be a different file
                message_box( "Goto Error",
                    { "The error file EX.ERR refers to ", err[1] } )

                return
            end if
        end if

    -- can go to line?
        if err[3] > length( edited[fNum][ITEMS] ) then

            -- can't go there
            message_box( "Goto Error",
                "EX.ERR refers to a line the beyond end of file." )

            -- go to end of file instead
            err[3] = length( edited[fNum][ITEMS] )

        end if

    -- go to line number
    if  err[3] >= edited[fNum][TOP_LINE]
    and err[3] <  edited[fNum][TOP_LINE] + edited[fNum][DEEP] then

        -- line is currently on screen
        edited[fNum][ITEM] = err[3] - edited[fNum][TOP_LINE] + 1

    elsif err[3] > 4 then

        -- position error line on line 5
        edited[fNum][TOP_LINE] = err[3] - 4
        edited[fNum][ITEM] = 1 + 4

    else

        -- position error line relative to top line
        edited[fNum][TOP_LINE] = 1
        edited[fNum][ITEM] = err[3]

    end if

    -- remove anchor, if any
    edited[fNum][ANCHOR_COL] = 0

    -- set column
    if err[4] = 0 then
        -- column 1
        edited[fNum][CURSOR] = 1
        edited[fNum][FIRST_CHAR] = 1
    else
        -- move to error column
        edited[fNum][FIRST_CHAR] = 1

        -- scroll right, if needed
        while edited[fNum][FIRST_CHAR] + edited[fNum][WIDE] - 1 < err[4] do
            edited[fNum][FIRST_CHAR] = edited [ fNum][FIRST_CHAR]+ edited[fNum][WIDE]
        end while
        
        -- position cursor
        edited[fNum][CURSOR] = err[4] - edited[fNum][FIRST_CHAR] + 1

    end if

    -- show file
        edited[fNum] = edit_xle( edited[fNum], REFRESH )
        cursorAt = get_position()

    -- save screen
        savedScreen = save_clean_image( {1, 1}, {SCREEN_DEEP, SCREEN_WIDE} )

    -- error message below menu
        position( 2, 1 )
        put_colors( { BRIGHT_WHITE, RED }, repeat(  ' ', SCREEN_WIDE ) )

        position( 2, 2 )
        put_colors( { BRIGHT_WHITE, RED }, err[2] )

    -- status message
        status_message( "Press a key to continue..." )

    -- cursor on error line
        position( cursorAt[1], cursorAt[2] )

    -- wait for a key press
        keyPress = watch_keys()

    -- restore screen
        display_clean_image( {1, 1}, savedScreen )
        
end procedure

-----------------------------------------------------------------------------
procedure get_user_screen_shot()

    -- get a screen shot of the user screen

    object result

    -- get current video mode
        userVC = video_config()

    -- text or graphics?
    if userVC[VC_MODE] > 3 then
        -- save graphics screen
        userScreen = save_image( {1, 1},
            { userVC[VC_XPIXELS], userVC[VC_YPIXELS]} )

    else
        -- save text screen
        userScreen = save_text_image( {1, 1},
            { userVC[VC_LINES], userVC[VC_COLUMNS]} )

    end if

end procedure


-----------------------------------------------------------------------------
global procedure show_user_screen()

    -- show the saved screen

    integer keyPress
    sequence savedScreen, vc
    object tmp

    -- hide mouse
    mouse_pointer(0)

    -- save current video mode
        vc = video_config()

    -- save background
        savedScreen = save_clean_image( {1, 1}, {vc[VC_LINES], vc[VC_COLUMNS]} )

    -- restore to saved screen settings
        tmp = graphics_mode( userVC[VC_MODE] )
        tmp = text_rows( userVC[VC_LINES] )

    -- show the screen
    if userVC[VC_MODE] > 3 then
        -- graphics
        display_image( {1, 1}, userScreen )

    else
        -- text
        display_text_image( {1, 1}, userScreen )

    end if


    -- get a key
        keyPress = watch_keys()

    -- restore screen
        tmp = graphics_mode( vc[VC_MODE] )
        tmp = text_rows( vc[VC_LINES] )
        toggle_blinking( 0 )
        mouse_pointer(1)
        display_clean_image( {1, 1}, savedScreen )

end procedure




-----------------------------------------------------------------------------
global procedure dos_shell( sequence cmd, integer waitFlag )

    -- shell out to dos
    -- if runProg is true, a Euphoria program is being run

    integer keyPress
    sequence savedScreen, vc, newVc
    object result

    -- hide mouse
    mouse_pointer(0)

    -- get current video mode
        vc = video_config()

    -- save background
        savedScreen = save_clean_image( {1, 1}, {SCREEN_WIDE, SCREEN_WIDE} )

    -- restore screen to defaults
        bk_color( BLACK )
        text_color( WHITE )
        clear_screen()
        result = text_rows( 25 )        
        cursor( UNDERLINE_CURSOR )
        toggle_blinking( 1 )

    -- run the command
        system( cmd, 2 )

    -- get screen shot
        get_user_screen_shot()

    -- wait?
    if waitFlag then

        -- set status line to current parms
        newVc = video_config()
        SCREEN_WIDE = newVc[VC_COLUMNS]
        SCREEN_DEEP = newVc[VC_LINES]

        -- message
            status_message( "Press a key to continue..." )

        -- wait for a key
            keyPress = watch_keys()
    
    end if

    -- restore video mode
        result = graphics_mode( vc[VC_MODE] )
        result = text_rows( vc[VC_LINES] )                
        SCREEN_WIDE = vc[VC_COLUMNS]
        SCREEN_DEEP = vc[VC_LINES]
        
        
    -- restore screen
        toggle_blinking( 0 )
        mouse_pointer(1)
        mouse_events( LEFT_DOWN + LEFT_UP )
        display_clean_image( {1, 1}, savedScreen )
        wrap(0)

end procedure


-----------------------------------------------------------------------------
global procedure run_program()

    -- run the program

    integer dotAt
    sequence fileName, extension
    object result


    -- save the file if changes have been made
    if edited[fNum][DIRTY] then
        if write_file() = -1 then
            -- error
            return
        end if
    end if

    -- delete old error file
        if sequence( dir("ex.err") ) then
            system( "del ex.err > NUL", 2 )
        end if

    -- get extension, if any
        fileName = upper( edited[fNum][TITLE] )
        dotAt = match( ".", fileName )
        if dotAt != 0 then
            extension = fileName[dotAt..length(fileName)]
        else
            extension = ""
        end if

    -- what kind of file?
        if compare( extension, ".PP" ) = 0
        or compare( extension, ".P" ) = 0 then

            -- status line
            status_message( "Pre-processing..." )

            -- status line
            position( SCREEN_DEEP, SCREEN_WIDE-34 )
            put_colors( OF_STATUS, "  Line:     " )

            -- preprocess the program
            result = preprocess(edited[fNum][TITLE], 0 )

            if result[1] then
                dos_shell( "ex " & result[2] & " " & commandLine, 1 )
            end if
    
            -- status line
            status_message( statusMessage )


        elsif equal( extension, ".EXW" )
        or    equal( extension, ".EW" )
        or    equal( extension, ".WIN" )
        or    equal( extension, ".GTK" ) then
            -- run the windows version
            dos_shell( "exw " & edited[fNum][TITLE] & " " & commandLine, 1 )
        
    
        else
            -- run the dos 32 version
            dos_shell( "ex " & edited[fNum][TITLE] & " " & commandLine, 1 )
        end if
    
    -- goto line with error (if any)
        goto_error()

end procedure


-----------------------------------------------------------------------------
global procedure pp_program()

    -- preprocess the program

    integer dotAt
    sequence fileName, extension
    object result


    -- save the file if changes have been made
    if edited[fNum][DIRTY] then
        if write_file() = -1 then
            -- error
            return
        end if
    end if

    -- delete old error file
        if sequence( dir("ex.err") ) then
            system( "del ex.err > NUL", 2 )
        end if

    -- status line
        status_message( "Pre-processing..." )

        position( SCREEN_DEEP, SCREEN_WIDE-34 )
        put_colors( OF_STATUS, "  Line:     " )

    -- preprocess the program
        result = preprocess(edited[fNum][TITLE], 0 )
        if result[1] then
            message_box( "Editor", "Created " & result[2] & "." )
        else
            message_box( "Editor", "Error in " & edited[fNum][TITLE] & "." )
        end if

    -- status line
        status_message( statusMessage )

    -- goto line with error (if any)
        goto_error()

end procedure

-----------------------------------------------------------------------------
global procedure bind_program()

    -- bind the program

    sequence fileName
    object result

    -- save the file if changes have been made
    if edited[fNum][DIRTY] then
        if write_file() = -1 then
            -- error
            return
        end if
    end if

    -- delete old error file
    if sequence( dir("ex.err") ) then
        system( "del ex.err > NUL", 2 )
    end if

    -- convert .PP and .P into .EX and .E
    fileName = upper( edited[fNum][TITLE] )
    if match( ".EXW", fileName )
    or match( ".EW", fileName ) then
        -- bind windows version
        dos_shell( "bindw -clear_routines " & edited[fNum][TITLE], 0 )

    else
        -- bind win32 version
        dos_shell( "bind -clear_routines " & edited[fNum][TITLE], 0 )
    end if


    -- check success of bind
    if compare( dir("ex.err"), -1 ) = 0 then
        message_box( "Bind File", "Executable created." )
    else         
        message_box( "Bind File", "Unable to bind " &
            edited[fNum][TITLE] & "." )
    end if

    -- goto line with error (if any)
        goto_error()

end procedure




-----------------------------------------------------------------------------
global procedure command_dialog()

    -- get values for command line

    object text

    text = ask_user( "Command Line", "Enter Command Line Arguments:",
            commandLine )

    if not atom( text ) then
        commandLine = text
    end if

end procedure


-----------------------------------------------------------------------------
global procedure context_help()

    -- context sensitive help

    integer file_num
    sequence word, fName
    object env, buffer, line

    -- current word
    word = on_xle_word( edited[fNum] )

    -- no word?
    if length( word ) = 0 then
        return
    end if

    -- remove "(" from word
    if word[length(word)] = '(' then
        word = word[1..length(word)-1]
    end if

    -- look through help topics
    for i = 1 to length( help ) do  
        -- match topic?
        if equal( help[i][1], word ) then                             
            -- show help
            hide_cursor()
            help_dialog( "Editor Help on \"" & word & "\"", help[i][2] )
            return
        end if
    end for

    -- no help
    message_box( "Editor Help",
        "Help is not available on " & word & "." )

end procedure


-----------------------------------------------------------------------------
global procedure list_ascii()

    -- display the ASCII characters for the user

    integer n
    sequence chart

    -- build the list
    chart = {}
    n = 255 / 5
    for i = 1 to n do
        chart = append( chart, 
        sprintf( " %3d: %s      %3d: %s      %3d: %s      %3d: %s      %3d: %s ", 
        {i,i,i+n,i+n,i+(2*n),i+(2*n),i+(3*n),i+(3*n),i+(4*n),i+(4*n)} ) )
    end for
    
    -- display the list
    help_dialog( "ASCII Character Chart", chart )


end procedure

-----------------------------------------------------------------------------
global procedure load_reference( sequence fName )

    -- load reference from DOC directory

    sequence priorDir
    object env

    -- get environment

        env = getenv("EUDIR")
        if atom( env ) then
            message_box( "Editor", "Unable to find EUDIR." )            
            return
        end if   

    -- save current directory

        priorDir = currentDir
        
    -- set euphoria directory as current
    
        currentDir = env & "\\DOC\\"

    -- load file
        
        read_file( fName )
        
    -- restore current directory
    
        currentDir = priorDir
        
end procedure


-----------------------------------------------------------------------------
global procedure commands_dialog()

    -- show editing commands

    sequence commands

    commands = {                                                      
        " Cursor Movement Commands                                        ",
        "                                        ",  
        " Home       - Move to the start of the current line.             ",  
        " End        - Move to the end of the current line.               ",  
        " Ctrl+Up    - Scroll up one line.                                ",  
        " Ctrl+Down  - Scroll down one line.                              ",  
        " PageUp     - Scroll up one screen.                              ",  
        " PageDown   - Scroll down one screen.                            ",  
        " Ctrl+PgUp  - Scroll left one screen.                            ",  
        " Ctrl+PgDn  - Scroll right one screen.                           ",  
        " Ctrl+Home  - Scroll to the start of the document.               ",  
        " Ctrl+End   - Scroll to the end of the document.                 ",  
        " Ctrl+Left  - Move left one word.                                ",  
        " Ctrl+Right - Move right one word.                               ",  
        "                                                                 ",
        " Editing Commands                                                ",
        "                                                ",
        " Enter      - Start a new line.                                  ",
        " Delete     - Delete the character that the cursor is on.        ",
        " Backspace  - Delete the character to the left of the cursor.    ",
        " Tab        - Move the cursor to next tab stop.                  ",
        " Insert     - Switch between insert and overwrite modes.         ",
        " Ctrl+Y     - Delete the current line.                           ",
        " Ctrl+V     - Paste buffer contents into file.                   ",
        "                                                                 ",
        " Commands for Working with Selected Regions                      ",
        "                      ",
        " Shift      - Use the shift key in conjunction with the cursor   ",
        "              movement functions to select regions.              ",
        "                                                                 ",
        " Ctrl+C     - Copy the current selection to buffer.              ",
        " Ctrl+X     - Delete the current selection and copy it to buffer.",
        " Delete     - Delete the current selection.                      ",
        " Tab        - Indent the selected lines.                         ",
        " Shift+Tab  - Unindent the selected lines.                       ",
        "                                                                 ",
        " Commands for Finding and Replacing Text                         ",
        "                   ",
        " F3         - Repeat the last search.                            ",
        "                                                                 ",
        " Window Management Commands                                      ",
        "                                      ",
        " Alt+n      - Switch to window #n.                               ",
        "                                                                 ",
        " Macro Commands                                                  ",
        "                                                  ",
        " F12        - Run the recorded macro.                            ",
        " Ctrl+F12   - Start/Stop recording the macro.                    ",
        " Shift+F12  - Append to existing macro.                          ",
        "                                                                 ",
        "                                                                 ",
        " Other Commands                                                  ",
        "                                                  ",
        " F1         - Display context-sensitive help.                    ",
        "                                                                 ",
        "                                                                 " }

    -- show commands
        help_dialog( "Edit Commands", commands )

end procedure

-----------------------------------------------------------------------------
global procedure about_dialog()

    integer keyPress

    
    -- put stuff in the box
    new_dialog( "About", 0, 0, 0, 0, {
        "                                           ",
        "             Euphoria Editor               ",
        "               Version 1.2b                ",
        "       (c) David Cuny  Jan 28, 2000        ",
        "                                           ",
        " Ŀ ",
        "  Syntax Coloring   : Robert Craig       ",
        "  Mouse Routines    : Jiri Babor         ",
        "  I/O Port Routines : Jacques Deschenes  ",
        "  SVGA Support      : Pete Eberlein      ",
        "  ",
        "                                           ",
        "                                           ",
        "                                           " } )

    add_control( 1, BUTTON, "&OK", 13, 17, 8, 0, {DEFAULT_BUTTON, FOCUS} )

    -- run the dialog
    run_dialog()

    -- restore screen
    hide_dialog()

end procedure

-----------------------------------------------------------------------------
global procedure draw_editor_frame( sequence mle )

    sequence fName

    -- there is no bottom of this box

    -- alias mle

    -- file name
    if compare( edited[fNum][TITLE], "" ) = 0 then
        fName = " Untitled "
    else
        fName = " " & edited[fNum][TITLE] & " "
    end if


    if VGA_MODE then

       position( mle[ROW]-1, mle[COL]-1 )
       put_colors( {OF_BORDER[2], OF_BORDER[1]}, fName )

    else

        -- top of box
       position( mle[ROW]-1, mle[COL]-1 )
       put_colors( OF_BORDER, "" )
       put_colors( {OF_BORDER[2], OF_BORDER[1]}, fName )
       put_colors( OF_BORDER, repeat( '', mle[WIDE]-length(fName)) & "" )

       for i = 1 to mle[DEEP] do
    
           -- display border
           position( mle[ROW]+i-1, mle[COL]-1 )
           put_colors( OF_BORDER, "" )
           put_colors( OF_TEXT, repeat(   ' ', mle[WIDE] ) )
        
            -- right border
            if length( mle[ITEMS] ) > mle[DEEP] then
                -- display scroll bar
                -- case i
                if i = 1 then 
                    put_colors( OF_SCROLL, {24} )
                elsif i = mle[DEEP] then 
                    put_colors( OF_SCROLL, {25} )
                else
                    put_colors( OF_SCROLL, "" )
                end if
            else
                -- no scroll bar
                put_colors( OF_BORDER, "" )
            end if         
       end for
   end if

end procedure


-----------------------------------------------------------------------------
procedure edit_status()

    -- display the edit status
    
    sequence cursorAt
    integer line, col, char

    -- get the cursor
    cursorAt = get_position()

    -- hide cursor
    hide_cursor()

    -- alias edited[fNum]

    -- get current line & col
        line = edited[fNum][TOP_LINE]+edited[fNum][ITEM]-1
        col = edited[fNum][FIRST_CHAR]+edited[fNum][CURSOR]-1

   -- get current char (default is 32)
        char = 32
        -- line exists?
        if line <= length( edited[fNum][ITEMS] ) then
            -- within line?
            if col <= length( edited[fNum][ITEMS][line] ) then
                -- get char
                char = edited[fNum][ITEMS][line][col]
            end if
        end if

    -- line
        position( SCREEN_DEEP, SCREEN_WIDE-34 )
        put_colors( OF_STATUS, sprintf("  Line:%g   ", {line} ) )

    -- column                           
        position( SCREEN_DEEP, SCREEN_WIDE-20 )
        put_colors( OF_STATUS, sprintf("Col:%g  ", {col} ) )

    -- char
        position( SCREEN_DEEP, SCREEN_WIDE-11 )
        put_colors( OF_STATUS, sprintf("Value:%g  ", {char} ) )


    -- macro
        if macroState = Record then               
            -- recording macro
            position( SCREEN_DEEP, SCREEN_WIDE-2 )
            put_colors( OF_STATUS, "MR" )
        elsif length( macroKeys ) then
            -- a macro is recorded
            position( SCREEN_DEEP, SCREEN_WIDE-2 )        
            put_colors( OF_STATUS, "M" )            
        end if
        

    -- re-position the cursor             
        position( cursorAt[1], cursorAt[2] )

end procedure


-----------------------------------------------------------------------------
function redraw_screen( sequence file )

    -- hide cursor
    hide_cursor()
    
    -- clear the flag
        redraw_flag = 0

    -- menu
        draw_menu_bar( MAIN_MENU )

    -- borders
        draw_editor_frame( file )

    -- force elevator bar redraw at next edit
        file[ELEVATOR] = 0
    
    -- status message
       status_message( statusMessage )

    -- show file
        file = edit_xle( file, REFRESH )

    -- status line
    edit_status()

    -- return mle
        return file

end function

-----------------------------------------------------------------------------
procedure get_search_for()
    
    -- set word to search for based on what the cursor is currently on

    sequence selected

    -- is text highlit?
    selected = xle_selected_text( edited[fNum] )
    if length( selected ) > 0 then
        -- single line?
        if not atom( selected[1] ) then
            -- clear it
            selected = ""
        end if
    end if

    -- if no text is highlit, get word cursor is on
    if length( selected ) = 0 then
        selected = on_xle_word( edited[fNum] )
    end if

    -- replace old value
    if length( selected ) > 0 then
        searchFor = selected
    end if

end procedure

-----------------------------------------------------------------------------
function alpha_numeric( integer char )

    -- returns true if character is alpha numeric
    if char >= 'A' and char <= 'Z' then         return 1
    elsif   char >= 'a' and char <= 'z' then    return 1
    elsif   char >= '0' and char <= '9' then    return 1
    else                                        return 0
    end if

end function


-----------------------------------------------------------------------------
function editor_alt_key()

    -- handle the alt key being pressed

    integer keyPress, exitFlag
    object result    

    -- highlight the menu
    draw_lit_menu_bar( MAIN_MENU )

    -- set file toggle flag to false
    exitFlag = 0

    while 1 do

        -- wait for a key
        keyPress = watch_keys()

        -- special key?
        -- case keyPress

        if compare( keyPress, ALT_RELEASED) = 0 then 

            -- key code should be zero. If it is not, then a key was
            -- entered through the key pad.

            if KEY_CODE = 0 then

                -- was file toggle used?
                if exitFlag then

                    -- dim the menu bar
                    draw_menu_bar( MAIN_MENU )

                    -- exit loop
                    exit

                else

                    -- pass it to the menu
                    keyPress = run_menu( MAIN_MENU, keyPress )

                end if

            else

                -- place code in the clipboard
                if KEY_CODE <  ' ' then
                    clipboard = ""
                else
                    clipboard = { KEY_CODE }
                end if
                
                -- dim the menu bar
                draw_menu_bar( MAIN_MENU )

                -- add clipboard character
                keyPress = CTRL_V

            end if
            
            -- exit while loop
            exit

        elsif compare(keyPress, ALT_1) > -1 and compare( keyPress,  ALT_0) < 1 then 

            -- toggle to the selected file

            -- toggle flag is true
            exitFlag = 1

            -- convert to an index
            result = keyPress - ALT_1 + 1

            -- process
            if  fNum != result                  -- not current file
            and fNum <= length( edited ) then   -- exists

                -- select new file
                select_file_number( result )

                -- display it
                edited[fNum] = redraw_screen( edited[fNum] )


                -- is the alt key still pressed?
                if alt_key_state() then

                    -- light the menu bar again
                    draw_lit_menu_bar( MAIN_MENU )

                else

                    -- dim the menu bar
                    draw_menu_bar( MAIN_MENU )

                    -- exit loop
                    exit

                end if

            end if

        else

            if keyPress >= 'A' and keyPress <= 'Z'
            or keyPress >= 'a' and keyPress <= 'z'
            or keyPress >= '0' and keyPress <= '9'
            or keyPress > 255 then

                -- pass it to the menu
                keyPress = run_menu( MAIN_MENU, keyPress )
    
                -- exit while loop
                exit

            else
        
                -- Alt+Gr key?
    
                -- set exit flag
                exitFlag = 1
    
                -- give key to mle
                edited[fNum] = edit_xle( edited[fNum], keyPress )
                
                -- display it
                edited[fNum] = redraw_screen( edited[fNum] )
    
                -- light the menu bar again
                draw_lit_menu_bar( MAIN_MENU )

            end if
        
        end if

    end while

    return keyPress

end function

-----------------------------------------------------------------------------

include d_search.e                  -- mle search/replace routines
include d_subs.e                    -- list subroutines
include d_set.e                     -- settings
include d_colors.e                  -- color picker

sequence priorTime
priorTime = date()
-----------------------------------------------------------------------------
global procedure run_clock()

    -- displays a clock in the menu bar
    
    sequence at, theTime                    
    
    -- get the time
    theTime = date()
    
    -- has it changed?
    if compare( priorTime, theTime ) != 0 then
        -- save cursor position
        at = get_position()
        
        -- locate in menu bar
        position( 1, 71 )     
        
        -- show time
        put_colors( OF_MENUBAR, sprintf("%02d:%02d:%02d", theTime[4..6]))
        
        -- store time
        priorTime = theTime
        
        -- reposition cursor
        position( at[1], at[2] )
        
    end if
    
end procedure

-- set the routine id (defined in KEYS.E)
    editorClockId = routine_id( "run_clock" )
    
-----------------------------------------------------------------------------
procedure create_help()
    
    sequence topic, text
    integer file, s1, s2, isTopic
    object env, line    

    -- help words are in \euphoria\doc\library.doc file
    -- file name is "word.hlp"
    env = getenv("EUDIR")
    if sequence( env ) then
        -- look in the doc directory
        file = open( env & "\\DOC\\LIBRARY.DOC", "r" )
    else
        file = -1
    end if
    
    -- not in doc
    if file = -1 then
        -- try current directory
        file = open( "library.doc", "r" )
    end if

    if file = -1 then
        -- unable to read library.doc file
        return
    end if

    -- initialize
    isTopic = 0         -- still in heading of file
    topic = ""          -- no topic
    text = {}           -- no topic text

    -- read through file        
    while 1 do   
    
        -- read a line from the file
        line = gets(file)
        
        -- end of file?
        if atom(line) then
            exit   -- end of file
        end if

        -- clean up
        line = detab_text( line )

        -- command?
        if length( line ) > 2
        and line[2] = '' then

            -- need to save prior topic?
            if length( topic ) then
                -- add to help
                help &= { {topic, text} }
                
                -- clear
                topic = ""
                text = {}    
                
            end if

            -- trim before '<'
            s1 = find( '<', line ) + 1
            s2 = find( '>', line ) - 1
            -- line[s1..s2]

            -- save topic
            topic = line[s1..s2]
            
            -- in help
            isTopic = 1

        -- topic header?
        elsif isTopic
        and length( line ) > 2
        and line[2] != ' '
        and find( ':', line ) then

            -- 'Syntax:'?
            if match( "Syntax:", line ) then
                -- clear preceeding text; they were blank spaces
                text = {}
                
                -- where does the keyword belong?
                if match( "include", line ) then
                    -- library
                    library &= {topic}
                else
                    builtins &= {topic}
                end if
                
            end if

            -- set to upper case and store as topic
            s1 = find( ':', line )
            text &= { upper( line[2..s1] ) }
        
            -- text following?
            for i = s1+1 to length( line ) do
                -- not a space?
                if line[i] != ' ' then
                    -- split off attached text
                    text &= { line[i..length(line) ] }
                    exit
                end if
            end for

        -- add to topic text?
        elsif isTopic then
            
            -- strip padding?
            if length( line ) > 15 then         
                -- strip
                line = line[15..length(line)]  
            end if
            
            -- add to text
            text &= { line }

        end if            

    end while
    
    -- topic pending?
    if length( topic ) then
        -- add to help
        help &= { {topic, text} }
    end if
    
end procedure




-----------------------------------------------------------------------------


------------------
-- MAIN ROUTINE --
------------------
------------------

integer keyPress, flag
sequence cursorAt
object result, cmd

-- housekeeping
    edited = {}                     -- no files open
    get_user_screen_shot()          -- copy screen
    mouse_pointer( 1 )              -- activate the mouse
    currentDir = current_dir()      -- current directory
    fNum = 1                        -- edited file number    
    toggle_blinking( 0 )            -- set blinking to intense
    
-- macros    
    macroState = Off                -- macro is not running
    macroAt = 0                     -- point to zero
    macroKeys = {}                  -- macro is empty

-- ini file
    read_ini_file()                 -- load settings from ini file

-- help file
    create_help()                   -- create help from LIBRARY.DOC

-- open an empty file for display
    open_untitled_file()

-- was a file entered at the command line?
    cmd = command_line()
    if length( cmd ) > 2 then
        -- should be file name
        result = full_path( upper( cmd[3] ), currentDir )
        -- is it a file/new file?
        if compare( result[3], "f" ) = 0
        or compare( result[3], "n" ) = 0 then
            -- existing or valid file name
            currentDir = result[1]
            edited[fNum][TITLE] = result[2]                
        else
            -- tell the user there's a problem
            message_box( "Editor", "Unable to open the file \""
                & edited[fNum][TITLE] & "\"" ) 
        end if

    end if

-- defer the display
    defer_display()

-- put up blank screen
    edited[fNum] = redraw_screen( edited[fNum] )    

-- display the screen
    show_deferred()

-- read the file?
    if length( edited[fNum][TITLE] ) = 0 then

        -- no file name given. check ex.err
        result = read_error_file()
        if not atom( result ) then
            -- open the file
            read_file( strip_path( upper( result[1] ) ) )

            -- go to the error line
            goto_error()
        end if

    else

        -- does the file exist?
        result = dir( full_path_name( edited[fNum][TITLE] ) )

        if not atom( result ) then
            -- exists. open it
            read_file( edited[fNum][TITLE] )
        end if

    end if


-- if there are two files open, close the first
    if length( edited ) > 1 then

        -- only use the second file
            edited = { edited[2] }
            fNum = 1

        -- rebuild the list
            build_file_list()

    end if


-- cursor at
    cursorAt = get_position()


-- until the user exits
while 1 do

    -- redraw the screen?
    if redraw_flag then
        edited[fNum] = redraw_screen( edited[fNum] )
        cursorAt = get_position()
    end if

    -- status line
    edit_status()

    -- re-position the cursor             
    position( cursorAt[1], cursorAt[2] )
    
    -- make sure the cursor is back on
    if VGA_MODE then
        show_graphic_cursor()
    else
        show_cursor()
    end if

    -- get a key
    if macroState = Play then
        -- move to next key
        macroAt = macroAt + 1
        
        -- run key, if not past the end
        if macroAt > length( macroKeys ) then
            -- end of macro
            macroState = Off
            keyPress = 0
            
        else
            -- get a key from the macro
            keyPress = macroKeys[macroAt]
                    
        end if
        
    else        
        
        -- get a key
        keyPress = watch_keys()

        -- add to macro?        
        if macroState = Record then
        
            -- decide if key should be part of macro
            if keyPress >= 800 then
                -- mouse event. ignore.             
                
            elsif find( keyPress, {F12, SHIFT_F12, CTRL_F12} ) then
                -- macro event. don't record.
                
            else
                -- good key
                macroKeys = macroKeys & keyPress
                
            end if
        end if

    end if    

    -- make sure the cursor is off
    if VGA_MODE then
        hide_graphic_cursor()
    else
        hide_cursor()
    end if

    -- set menu states
    -- only do these if menu might be activated
    if ( MOUSE_CLICK and MOUSE_ROW = 1 )
    or ( keyPress = ALT_PRESSED ) then

        -- dim/set file menu
        -- alias MAIN_MENU[1][ITEMS]
        MAIN_MENU[1][ITEMS][3][DIM] = (length( edited[fNum][TITLE] ) = 0)    -- save needs file name
    
        -- dim/set edit menu
        -- alias MAIN_MENU[2][ITEMS]
        flag = ( edited[fNum][ANCHOR_COL] = 0 )   -- true if no text is selected
        MAIN_MENU[2][ITEMS][1][DIM] = flag                          -- cut needs text selected
        MAIN_MENU[2][ITEMS][2][DIM] = flag                          -- copy needs text selected
        MAIN_MENU[2][ITEMS][3][DIM] = not compare( clipboard, "" )  -- paste needs text in clip
        MAIN_MENU[2][ITEMS][4][DIM] = flag                          -- clear needs text selected
    
        -- dim/set search menu
        -- alias MAIN_MENU[3][ITEMS]
        MAIN_MENU[3][ITEMS][2][DIM] = not compare( searchFor, "" )  -- repeat prior find needs text
    
        -- dim/set view menu
        -- alias MAIN_MENU[4][ITEMS]
        MAIN_MENU[4][ITEMS][3][DIM] = not sequence( dir("ex.err") ) -- needs error file

        -- dim/set run menu
        -- alias MAIN_MENU[5][ITEMS]
        MAIN_MENU[5][ITEMS][1][DIM] = (length( edited[fNum][TITLE] ) = 0)   -- need file name

        MAIN_MENU[5][ITEMS][2][DIM] = 1  -- default is PP not available
        if PP_LOADED then  -- loaded?
            if match( ".P", upper( edited[fNum][TITLE] ) ) then  -- extension?
                MAIN_MENU[5][ITEMS][2][DIM] = 0
            end if
        end if


        -- check options
        -- alias MAIN_MENU[6][ITEMS]
        MAIN_MENU[6][ITEMS][4][CHECKED] = edited[fNum][COLOR_SYNTAX]      -- use color syntax?
        MAIN_MENU[6][ITEMS][5][CHECKED] = edited[fNum][AUTOCOMPLETE]      -- use auto complete?


    end if

    -- handle alt key
    if keyPress = ALT_PRESSED then
        keyPress = editor_alt_key()
    end if

    -- give key to mle
    edited[fNum] = edit_xle( edited[fNum], keyPress )

    -- give key to menu
    keyPress = give_menu( MAIN_MENU, keyPress )

    -- cursor at
    cursorAt = get_position()

   -- menu item selected?
    if keyPress > 1000              -- menu
    or keyPress = F1                -- help
    or keyPress = F2                -- subs/functions
    or keyPress = F3                -- find
    or keyPress = F4                -- output screen
    or keyPress = F5                -- run
    or keyPress = F12               -- run macro
    or keyPress = CTRL_F12          -- start/stop macro
    or keyPress = SHIFT_F12         -- append macro
    then

        -- clear status bar
        status_message( "" )

        -- react to menu selection
        -- case keyPress
        if keyPress =  1101 then    -- new            
            editor_new_file()

        elsif keyPress =  1102 then    -- open
            editor_open_file()

        elsif keyPress =  1103 then    -- save
            editor_save_file()

        elsif keyPress =  1104 then    -- save as
            editor_save_as()

        elsif keyPress =  1105 then    -- close
            editor_close()

        elsif keyPress =  1107 then    -- print
            editor_print()

        elsif keyPress =  1109 then    -- exit
            editor_exit()

        elsif keyPress =  1201 then    -- cut
            edited[fNum] = edit_xle( edited[fNum], CTRL_X )     -- do copy

        elsif keyPress =  1202 then    -- copy
            edited[fNum] = edit_xle( edited[fNum], CTRL_C )     -- do cut

        elsif keyPress =  1203 then    -- paste
            edited[fNum] = edit_xle( edited[fNum], CTRL_V )     -- do paste

        elsif keyPress =  1204 then    -- clear
            edited[fNum] = edit_xle( edited[fNum], DELETE )     -- do delete

        elsif keyPress =  1301 then    -- find
            get_search_for()        -- get word(s) at cursor
            edited[fNum] = mle_find( edited[fNum] )             -- try to match

        elsif keyPress =  1302 or keyPress =  F3 then    -- repeat last find
            edited[fNum] = mle_find_next( edited[fNum], FIND_MODE )

        elsif keyPress =  1303 then    -- replace
            get_search_for()        -- get word(s) at cursor
            edited[fNum] = mle_replace( edited[fNum] )          -- try to replace

        elsif keyPress =  1401 or keyPress =  F2 then    -- output screen
            subs_dialog( edited[fNum][ITEMS], 
                on_xle_word( edited[fNum] ) )                 -- sub list

        elsif keyPress =  1402 then        -- goto line
            goto_line()                                     -- jump to line

        elsif keyPress =  1403 then        -- show error
            goto_error()                                    -- jump to error

        elsif keyPress =  1404 or keyPress =  F4 then    -- sub list
            show_user_screen()                              -- output screen

        elsif keyPress >=  1406 and keyPress <=  1416 then    -- edited file
            select_file_number( keyPress - 1406 + 1 )

        elsif keyPress =  1501 or keyPress =  F5 then    -- run program
            run_program()

        elsif keyPress =  1502 then  -- preprocess program
            pp_program()

        elsif keyPress =  1503 then    -- bind program
            bind_program()

        elsif keyPress =  1505 then    -- command line
            command_dialog()

        elsif keyPress =  1507 or keyPress =  F12 then    -- run macro
            macroState = Play
            macroAt = 1

        elsif keyPress =  1508 or keyPress =  CTRL_F12 then    -- start/stop macro
            if macroState = Off then
                macroState = Record
                macroKeys = {}
            else
                macroState = Off
            end if
        
        elsif keyPress =  1509 or keyPress =  SHIFT_F12 then    -- append to macro
                macroState = Record                

        elsif keyPress =  1511 then    -- dos prompt
            dos_shell( "command.com", 0 )

        elsif keyPress =  1601 then    -- settings
            settings_dialog()

        elsif keyPress =  1602 then    -- colors
            colors_dialog()

        elsif keyPress =  1604 then    -- syntax coloring
            edited[fNum][COLOR_SYNTAX] = not edited[fNum][COLOR_SYNTAX]
            redraw_flag = 1

        elsif keyPress =  1605 then    -- autocompletion
            edited[fNum][AUTOCOMPLETE] = not edited[fNum][AUTOCOMPLETE]

        elsif keyPress =  1701 then    -- commands
            commands_dialog()

        elsif keyPress =  1702 then    -- ascii chart
            list_ascii()

        elsif keyPress =  F1 then      -- help
            context_help()

        elsif keyPress =  1704 then    -- refman
            load_reference( "refman.doc" )

        elsif keyPress =  1705 then    -- library
            load_reference( "library.doc" )

        elsif keyPress =  1707 then    -- about
            about_dialog()

        end if

        -- status line
        edit_status()

        -- position cursor
        edited[fNum] = edit_xle( edited[fNum], 0 )
        cursorAt = get_position()

    end if

end while

arctan

