-- fonter.e
-- major rewrite by Pete Eberlein
-- original code by David Cuny's fonter.ex code
--
-- modified from fonter.ex for use with screen.e for:
--  1. displaying graphic text, and
--  2. loading custom fonts

include image.e

sequence char
char = repeat({}, 256)

-----------------------------------------------------------------------------
global procedure graphic_text( integer row, integer col,
                    integer fColor, integer bColor, sequence chars )

    -- displays chars at position in given colors

    integer x, y, l, dx
    sequence pos, ch, image


    -- get x and y and store in simple variables
    y = (row-1) * cTall
    x = (col-1) * cWide
    pos = {x,y}
    image = repeat(repeat(bColor, length(chars) * cWide), cTall)

    dx = 0
    for c = 1 to length(chars) do
        ch = char[chars[c]]
        for i = 1 to length(ch)-1 by 2 do
            x = ch[i][1] + dx 
            y = ch[i][2]
            l = ch[i+1]
            image[y][x..x+l-1] = repeat(fColor, l)
        end for
        dx = dx + cWide
    end for


    display_image(pos, image)

    end procedure


-----------------------------------------------------------------------------
global procedure loadFont()

    -- load font into font memory

    sequence regs
    integer font_buf, i, on

    -- register array
    regs = repeat(0, 10)

    regs[REG_AX] = #1130
    if cWide = 8 and cTall = 16 then
        regs[REG_BX] = #0600  
    elsif cWide = 8 and cTall = 14 then
        regs[REG_BX] = #0200
    elsif cWide = 8 and cTall = 8 then
        regs[REG_BX] = #0300
    end if

    regs = dos_interrupt(#10, regs)

    font_buf = regs[REG_ES] * #10 + regs[REG_BP]
    for newAscii = 1 to 256 do
        for y = 1 to cTall do
            i = peek(font_buf + (newAscii)*cTall + y-1)
            on = 0
            for x = 1 to 8 do
                if on then
                    if and_bits(i, power(2, 8-x)) then
                        char[newAscii][on] = char[newAscii][on] + 1
                    else
                        on = 0
                    end if
                elsif and_bits(i, power(2, 8-x)) then
                    char[newAscii] = char[newAscii] & {{x,y}, 1}
                    on = length(char[newAscii])
                end if
            end for
        end for
    end for

end procedure


-- Load in new font definitions

without warning
without trace

include machine.e     -- We need to access BIOS interrupts


----------------------------------------------------------------------------
procedure addFonts( integer start, sequence char )
    object    lowMem
    sequence  inReg, outReg

    lowMem = allocate_low( length( char ) )   -- get memory
    poke( lowMem, char )         -- move description to memory

    inReg = repeat( 0, 10 )         -- initialize register list
    inReg[REG_AX] = #1100           -- new char function
    inReg[REG_BX] = 16 * 256        -- 16 bytes/char - for 256 characters
    inReg[REG_CX] = length(char)/16 -- number of characters to change
    inReg[REG_DX] = start           -- first character to change
    inReg[REG_ES] = floor(lowMem / 16)    -- segment of low_mem (low memory)
    inReg[REG_BP] = remainder(lowMem, 16) -- offset of low_mem
    outReg = dos_interrupt( #10, inReg )  -- bios call

    free_low( lowMem )  -- free the memory

end procedure



-----------------------------------------------------------------------------
global procedure loadFontFile( sequence fileName )

    sequence newChar
    object text, handle
    integer startAscii, priorAscii, newAscii, accum
    integer y, on


    -- open the file
    handle = open( fileName, "r" ) -- font file
    if handle = -1 then         
            puts( 1, "Can't open " & fileName & ".\n" )
            abort( 1 )
    end if

    -- initialize
    newChar = {}
    priorAscii = 0
    startAscii = 0

    while 1 = 1 do

        -- read until eof
        text = gets( handle ) -- get text
        if atom( text ) then  -- eof?
            exit -- quit loop
        end if

        if text[1] = ';' then
            -- ignore, comment

        elsif text[1] = '#' or text[1] = '-' then
    
            -- get the character code
            if text[1] = '-' then
                -- get ascii code
                newAscii = text[2]
            else
                -- convert to number
                newAscii = 0
                for i = 2 to length( text ) do
                    if text[i] = ' ' then 
                        exit  
                    end if

                    -- convert to digits
                    newAscii = ( newAscii * 10 ) + ( text[i] - '0' )

                end for
            end if
    
            if VGA_MODE then
                y = 1
            else
                -- if a break in the sequence, display prior
                if newAscii != priorAscii + 1 then
                    -- add if there is something to add
                    if length( newChar ) > 0 then
                        addFonts( startAscii, newChar )
                    end if

                    -- new start of fonts
                    priorAscii = newAscii
                    startAscii = newAscii
                    newChar = {}
                else
                    -- reset prior character
                    priorAscii = newAscii
                end if
            end if
                
        elsif text[1] = '.' or text[1] = 'x' then
            if VGA_MODE then
                if y = 1 then
                    char[newAscii] = {}
                end if
                -- find strips
                on = 0
                for x = 1 to 8 do
                    if on then
                        if text[x] = 'x' then
                            char[newAscii][on] = char[newAscii][on] + 1
                        else
                            on = 0
                        end if
                    elsif text[x] = 'x' then
                        char[newAscii] = char[newAscii] & {{x,y}, 1}
                        on = length(char[newAscii])
                    end if
                end for
                y = y + 1
                
            else
                -- convert to binary
                accum = 0
                if text[1] = 'x' then accum = accum + 128       end if
                if text[2] = 'x' then accum = accum + 64        end if
                if text[3] = 'x' then accum = accum + 32        end if
                if text[4] = 'x' then accum = accum + 16        end if
                if text[5] = 'x' then accum = accum + 8         end if
                if text[6] = 'x' then accum = accum + 4         end if
                if text[7] = 'x' then accum = accum + 2         end if
                if text[8] = 'x' then accum = accum + 1         end if
                
                -- store in newChar
                newChar = newChar & accum
            end if

        else
                -- junk.
        end if



    end while
    
    -- close the file
    close( handle )

end procedure

