-- Create an INI file for the editor

-----------------------------------------------------------------------------
constant colorList = {
        "BLACK", "BLUE", "GREEN", "CYAN", "RED", "MAGENTA", "BROWN", "WHITE",
        "GRAY", "BRIGHT_BLUE", "BRIGHT_GREEN", "BRIGHT_CYAN", "BRIGHT_RED", 
        "BRIGHT_MAGENTA", "YELLOW", "BRIGHT_WHITE" }

-----------------------------------------------------------------------------
procedure write_color_pair( integer fHandle, sequence name, sequence pair )

    -- write the color scheme pair out

    printf( fHandle, "%s=%s,%s\n", 
        {name, colorList[pair[1]+1], colorList[pair[2]+1]} )

end procedure
    

-----------------------------------------------------------------------------
global procedure write_ini_file()

    -- write out the new values for the .ini file
    
    integer fHandle
    object env

    -- get environment
    env = getenv("EUDIR")
    if atom( env ) then
        return
    else
        env = env & "\\BIN\\"
    end if

    -- open the .ini file
    fHandle = open( env & "ee.ini", "w" )
    if fHandle != -1 then

        -- the VGA colors are hard coded, so skip

        if not VGA_MODE then

            -- color schemes
            puts( fHandle, "[Color Scheme]\n" )
            write_color_pair( fHandle, "Text", OF_TEXT )
            write_color_pair( fHandle, "Selected", OF_SELECTED )
            write_color_pair( fHandle, "Border", OF_BORDER )
            write_color_pair( fHandle, "MenuBar", OF_MENUBAR )
            write_color_pair( fHandle, "Status", OF_STATUS )
            write_color_pair( fHandle, "Accel", OF_ACCEL )
            write_color_pair( fHandle, "Disabled", OF_DISABLED )
            write_color_pair( fHandle, "Dialog", OF_DIALOG )
            write_color_pair( fHandle, "Title", OF_DIALOG_TITLE )
            write_color_pair( fHandle, "Button", OF_DIALOG_BUTTON )
            write_color_pair( fHandle, "HotKey", OF_HOTKEY )
            write_color_pair( fHandle, "Scroll", OF_SCROLL )

        end if
        
        -- printer port
        puts( fHandle, "[Printer]\n" )
        printf( fHandle, "Port=%s\n", {PRINTER_PORT} )
        
        -- lines on screen
        puts( fHandle, "[Screen Depth]\n" )
        printf( fHandle, "Lines=%d\n", {SCREEN_DEEP} )


        -- tab size
        puts( fHandle, "[Tab Size]\n" )
        printf( fHandle, "Tab=%d\n", {TAB_SIZE} )


        close( fHandle )

    end if

end procedure


-----------------------------------------------------------------------------
function read_color_pair( sequence s, sequence default )

    integer comma, color1, color2
    sequence color

    -- the VGA colors are hard coded, so skip
    if VGA_MODE then
        return default
    end if    

    comma = find( ',', s )
    if comma > 0 then
        color = s[1..comma-1]    
        color = s[comma+1..length(s)]
        
        color1 = find( s[1..comma-1], colorList ) - 1
        color2 = find( s[comma+1..length(s)], colorList ) - 1
        if  color1 > -1
        and color2 > -1 then
            return { color1, color2 }
        end if
    end if

    -- error. return default
    return default

end function

-----------------------------------------------------------------------------
function read_ini_string( sequence s, sequence default )

        
    if length( s ) > 0 then
        return s
    else
        return default
    end if

end function

-----------------------------------------------------------------------------
function read_ini_number( sequence s, object default )

    object result

    -- number?
    result = value( s )
    -- successful translation?
    if result[1] = GET_SUCCESS then
        return result[2]
    end if
        
    return default

end function

-----------------------------------------------------------------------------
global procedure read_ini_file()
    
    integer fHandle, sep
    sequence attrib, args
    object line, result, env

    -- get environment
    env = getenv("EUDIR")
    if atom( env ) then
        return
    else
        env = env & "\\BIN\\"
    end if

    -- open the file
    fHandle = open( env & "ee.ini", "r")
    if fHandle = -1 then
        -- Warning
        -- message_box( "Editor", "Unable to open Editor .ini file." )
        return
    end if

    while 1 do
        
        line = gets( fHandle )
        if atom( line ) then
            exit   -- end of file
        else

            -- strip off line feed at end of line
            line = line[1..length(line)-1]

            -- not empty?
            if length( line ) > 0 then

                -- look for the arg
                sep = find( '=', line )
                if sep > 0 then
                    attrib = line[1..sep-1]
                    args = line[sep+1..length(line) ]
                    if compare( attrib, "Text" ) = 0 then
                        OF_TEXT = read_color_pair( args, OF_TEXT )

                    elsif compare( attrib, "Selected" ) = 0 then
                        OF_SELECTED = read_color_pair( args, OF_SELECTED )

                    elsif compare( attrib, "Border" ) = 0 then
                        OF_BORDER = read_color_pair( args, OF_BORDER )

                    elsif compare( attrib, "MenuBar" ) = 0 then
                        OF_MENUBAR = read_color_pair( args, OF_MENUBAR )
                    
                    elsif compare( attrib, "Status" ) = 0 then
                        OF_STATUS = read_color_pair( args, OF_STATUS )
                    
                    elsif compare( attrib, "Accel" ) = 0 then
                        OF_ACCEL = read_color_pair( args, OF_ACCEL )

                    elsif compare( attrib, "Disabled" ) = 0 then
                        OF_DISABLED = read_color_pair( args, OF_DISABLED )

                    elsif compare( attrib, "Dialog" ) = 0 then
                        OF_DIALOG = read_color_pair( args, OF_DIALOG )

                    elsif compare( attrib, "Title" ) = 0 then
                        OF_DIALOG_TITLE = read_color_pair( args, OF_DIALOG_TITLE )

                    elsif compare( attrib, "BUTTON" ) = 0 then
                        OF_DIALOG_BUTTON = read_color_pair( args, OF_DIALOG_BUTTON )

                    elsif compare( attrib, "HotKey" ) = 0 then
                        OF_HOTKEY = read_color_pair( args, OF_HOTKEY )

                    elsif compare( attrib, "Scroll" ) = 0 then
                        OF_SCROLL = read_color_pair( args, OF_SCROLL )

                    elsif compare( attrib, "Port" ) = 0 then
                        PRINTER_PORT = read_ini_string( args, PRINTER_PORT )

                    elsif compare( attrib, "Lines" ) = 0 then
                        SCREEN_DEEP = read_ini_number( args, SCREEN_DEEP )

                    elsif compare( attrib, "Tab" ) = 0 then
                        TAB_SIZE = read_ini_number( args, TAB_SIZE )

                    end if
                end if
            end if
        end if
    
    end while

    -- close the file
    close( fHandle )

    -- handle screen resolution

        -- deactivate the mouse
        mouse_pointer( 0 )    

        -- attempt to change res
        SCREEN_DEEP = text_rows( SCREEN_DEEP )
        
        -- activate intense colors again
        toggle_blinking( 0 )

        -- re-activate the mouse
        mouse_pointer( 1 )    

        -- save results
        result = video_config()
        SCREEN_WIDE = result[VC_COLUMNS]
        SCREEN_DEEP = result[VC_LINES]


end procedure


