-- scrolling lists for Euphoria

include keys.e          -- key definitions
include get.e           -- saving screen
include graphics.e      -- colors
include image.e         -- save_text_image()
include wildcard.e      -- upper/lower
include control.e       -- color scheme
include sort.e          -- sorted lists
include scroll.e        -- handle scroll bar

without warning



-----------------------------------------------------------------------------
global procedure draw_list_item( sequence list, integer item )

	sequence colors

    integer this

	-- alias list

	-- empty?
	if length( list[ITEMS] ) = 0 then
		return
	end if

	-- color
    if item = list[ITEM] then
        colors = { OF_DIALOG[2], OF_DIALOG[1] }
    else
        colors = OF_DIALOG
    end if

    -- absolute position
    this = list[TOP_LINE] + item - 1

	-- go to position
    position( list[ROW]+item-1, list[COL] )

	-- will text fit?
    if length( list[ITEMS][this]) > list[WIDE]-2-VGA_MODE  then
	 
		-- truncate
        put_colors( colors, " " & list[ITEMS][this][1..list[WIDE]-2-VGA_MODE] & " " )

	else
	 
		-- text fits
        put_colors( colors, " " & list[ITEMS][this] )
        put_colors( colors, repeat(  ' ',
            list[WIDE]-VGA_MODE-1-length(list[ITEMS][this]) ) )

	end if

    -- run elevator
    list = draw_elevator( list )

	-- place cursor    
    position( list[ROW]+item-1, list[COL]-VGA_MODE )


end procedure


-----------------------------------------------------------------------------
global function draw_list( sequence list )
	
	-- draw a list
	
    integer this
	sequence text

	-- refer to the list
	-- alias list

	-- only draw if flag is set
	if list[REDRAW] then

        if VGA_MODE then

            if list[DIM] then

                graphic_title( list[TITLE], {OF_DISABLED[1], OF_DIALOG[2]},
                            list[ROW]-1, list[COL], list[HOT], 0 )
            
            else

                graphic_title( list[TITLE], OF_DIALOG,
                        list[ROW]-1, list[COL], list[HOT], 0 )

            end if

        else

            -- position title
            position( list[ROW]-2, list[COL] )
     
            if list[DIM] then
                put_colors( { OF_DISABLED[1], OF_DIALOG[2] }, list[TITLE] )
            else
                put_colors( OF_DIALOG, list[TITLE] )
                -- hot key?
                if list[HOT] != 0 then
                    position( list[ROW]-2, list[COL]-1 + list[HOT] )
                    put_colors( { OF_HOTKEY[1], OF_DIALOG[2] }, list[TITLE][list[HOT]] )
                end if
            end if

        end if
    end if


	-- body of list

    if VGA_MODE then

        if list[REDRAW] then

            -- draw frame
            graphic_frame( list[ROW]-1, list[COL]-1, list[WIDE], list[DEEP] )

        end if

    else

		-- position top line
		position( list[ROW]-1, list[COL]-1 )
		put_colors( OF_DIALOG, "" & repeat( '', list[WIDE] ) & "" )

    end if

    for i = 1 to list[DEEP] do

		-- current item
		this = list[TOP_LINE] + i - 1

        -- left hand side
        if VGA_MODE then
            position( list[ROW]+i-1, list[COL] )
        else
            position( list[ROW]+i-1, list[COL]-1 )
            put_colors( OF_DIALOG, "" )
        end if

        put_colors( OF_DIALOG, " " )

		if this <= length( list[ITEMS] ) then
			-- get correct length segment of text
            if length(list[ITEMS][this]) > list[WIDE]-2 then
                text = list[ITEMS][this][1..list[WIDE]-2] & " "
			else
				text = list[ITEMS][this] &
                    repeat(   ' ', list[WIDE]-VGA_MODE-1-length(list[ITEMS][this]) )
			end if

            -- the text
            put_colors( OF_DIALOG, text )

		else
			-- empty
            put_colors( OF_DIALOG, repeat(   ' ', list[WIDE]-VGA_MODE-1 ) )
		end if
				
	end for

	
    if  list[REDRAW]
    and not VGA_MODE then
		-- position top line
        position( list[ROW]+list[DEEP], list[COL]-1 )
		put_colors( OF_DIALOG, "" & repeat( '', list[WIDE] ) & "" )
	end if

    -- run the elevator
    list = draw_elevator( list )

	-- clear the flag    
	list[REDRAW] = 0

	-- check for empty list
	if length( list[ITEMS] ) != 0 then
		-- move cursor to selected item
        position( list[ROW]+list[ITEM]-1, list[COL]-VGA_MODE )
	end if

	-- return modified list
	return list

end function

-----------------------------------------------------------------------------
global function list_selected_index( sequence list )

	-- return text of current item

	-- alias list
	-- alias list

	if length( list[ITEMS] ) = 0 then
		return 0
	else
        return current_line( list )
	end if

end function

-----------------------------------------------------------------------------
global function list_selected_text( sequence list )

	-- return text of current item

	-- alias list

	-- check for empty list
	if length( list[ITEMS] ) = 0 then
		return ""
	else
        return list[ITEMS][current_line( list )]
	end if

end function

-----------------------------------------------------------------------------
global procedure draw_selected_list_item( sequence list )

	-- highlight the selected item on the list

	-- alias list

	draw_list_item( list, list[ITEM] )
	
end procedure

-----------------------------------------------------------------------------
global function focus_on_list_item( sequence list, integer moveTo )
	
	-- alias list

	-- empty?
	if length( list[ITEMS] ) = 0 then
		return list
	end if

	-- ensure that the selected item is visible in the list
	-- case moveTo
	if moveTo< 1 then 
		list[TOP_LINE] = 1
		list[ITEM] = 1
	elsif moveTo< list[TOP_LINE] then 
		list[TOP_LINE] = moveTo
		list[ITEM] = 1
    elsif moveTo> list[TOP_LINE]+list[DEEP]-1 then 
        list[TOP_LINE] = moveTo-list[DEEP]+1
        list[ITEM] = list[DEEP]
	else
		list[ITEM] = moveTo - list[TOP_LINE] + 1
	end if

	-- return new list
	return list

end function

-----------------------------------------------------------------------------
global function add_list_item( sequence list, sequence item )

	-- add an item to a list

	-- alias list

	-- add to list
	list[ITEMS] = append( list[ITEMS], item )

	-- focus on item
	list = focus_on_list_item( list, length( list[ITEMS] ) )

	-- draw list
	return draw_list( list )

end function


-----------------------------------------------------------------------------
global function add_sorted_list_item( sequence list, sequence item )

	-- add an item to a list; sort the list

	-- alias list

	-- add to list
	list[ITEMS] = append( list[ITEMS], item )

	-- sort list
	list[ITEMS] = sort( list[ITEMS] )

	-- find it, and focus on it
	list = focus_on_list_item( list, find( item, list[ITEMS] ) )

	-- draw list
	return draw_list( list )

end function


-----------------------------------------------------------------------------
global function remove_list_item( sequence list, integer index )

	-- remove an item from a list

	-- alias list

	-- don't delete what's not there
	if index > length( list[ITEMS] ) then
		-- no action
		return list
	end if

	-- remove from list
	list[ITEMS] = list[ITEMS][1..index-1] & list[ITEMS][index+1..length(list[ITEMS])]

	-- focus on prior item
	list = focus_on_list_item( list, length( list[ITEMS] ) - 1 )

	-- draw list
	return draw_list( list )

end function


-----------------------------------------------------------------------------
function move_up( sequence list )

	-- move up 1 line

	-- alias list

    if current_line( list ) > 1 then
		if list[ITEM] > 1 then
			list[ITEM] = list [ ITEM]- 1
		else
			-- scroll list up
			list[TOP_LINE] = list [ TOP_LINE]- 1                
		end if
	end if

	
	return list

end function


-----------------------------------------------------------------------------
function move_down( sequence list )

	-- move down 1 line

	-- alias list

    if current_line( list ) < length( list[ITEMS] ) then
	 
        if list[ITEM] < list[DEEP] then
			list[ITEM] = list [ ITEM]+ 1
		else
			-- scroll the list down
			list[TOP_LINE] = list [ TOP_LINE]+ 1                
		end if
	end if

	return list

end function



-----------------------------------------------------------------------------
global function edit_list( sequence list, integer keyPress )

	integer this, matchFlag, priorTop, priorItem

	-- run the list for one key press

	-- alias list

	-- empty?
	if length( list[ITEMS] ) = 0 then
		return list
	end if

	-- this is broken, so i'm patching it here
	list[ELEVATOR] = elevator_at( list )

	-- prior state of list
	priorItem = list[ITEM]
	priorTop = list[TOP_LINE]

	-- current item
    this = current_line( list )

	-- case keyPress

	if keyPress = MOUSE_CLICK or keyPress = MOUSE_DOUBLE or keyPress = MOUSE_HELD then 

        if ( MOUSE_COL = list[COL] + list[WIDE]               -- on scroll bar
             or MOUSE_COL = list[COL] + list[WIDE]-VGA_MODE )  -- on graphic scroll bar 
        and length( list[ITEMS] ) > list[DEEP] then           --  scroll bar exists

                list = scrollbar_clicked( list )

		else

			if MOUSE_ROW = list[ROW]-1 then              -- above list

				-- move up one line
				list = move_up( list )


            elsif MOUSE_ROW = list[ROW]+list[DEEP] then       -- below list

				-- move down one line
				list = move_down( list )

			else

				-- in body of list

				-- focus on the clicked item
				list[ITEM] = MOUSE_ROW - list[ROW] + 1
	
				-- don't select an item that isn't there
                if current_line( list ) > length( list[ITEMS] ) then
					list[ITEM] = length( list[ITEMS] )
				end if

			end if

		end if

	elsif keyPress = UP then 

		list = move_up( list )

	elsif keyPress = DOWN then 

		list = move_down( list )

	elsif keyPress = PAGE_UP then 

		list = page_up( list )

	elsif keyPress = PAGE_DOWN then 
 
		 list = page_down( list )

	elsif keyPress = HOME then 
 
		-- move to top
		list[TOP_LINE] = 1
		list[ITEM] = 1        

	elsif keyPress = END then 

		-- move to bottom
        if length( list[ITEMS] ) <= list[DEEP] then
			-- less than a full list
			list[ITEM] = length( list[ITEMS] )
		else
			-- go to bottom of list
            list[TOP_LINE] = length( list[ITEMS] ) - list[DEEP] + 1
            list[ITEM] = list[DEEP]
		end if

	else

		-- "hot" key?

		-- prevent upper() from choking on a negative key
		if keyPress > 0 and keyPress < 255 then

			-- flag that no matches have been found
			matchFlag = 0

			-- scan from current+1 to end...
            for i = current_line( list )+1 to length( list[ITEMS] ) do
				-- match?
				if compare( upper( {list[ITEMS][i][1]} ), 
							upper( {keyPress} ) ) = 0 then

					-- jump to it
					list = focus_on_list_item( list, i )
					matchFlag = 1
					exit
				end if
			end for

			-- from 1 to current item
			if not matchFlag then
				-- scan
                for i = 1 to current_line( list ) do
					
					if length( list[ITEMS][i] ) > 0 then
						if compare( upper( {list[ITEMS][i][1]} ),
										upper( {keyPress} ) ) = 0 then

							-- jump to it
							list = focus_on_list_item( list, i )
							exit
						end if
					end if
				end for
			end if

		end if

	end if

	-- need to redraw entire list?
	if priorTop != list[TOP_LINE] then
	
		-- redraw list
		list = draw_list( list )
	
		-- draw selected item
		draw_list_item( list, list[ITEM] )

	else

		-- has anthing changed?
		if priorItem != list[ITEM] then

			-- erase prior item
			draw_list_item( list, priorItem )

			-- draw selected item
			draw_list_item( list, list[ITEM] )

		end if

	end if

	return list

end function


