-- edit masks for SLEs
-- notes: MASK was added as an extra attribute in CONTROL.E

-- to do:
-- support overwrite/insert mode

with trace


-----------------------------------------------------------------------------
function masked_to_data( sequence sle )

    -- converts sle back from masked form

    integer pos
    sequence data

    data = "" 
    pos = 0
    for i = 1 to length( sle[MASK] ) do
        if sle[MASK][i] = '9' then
            data = data & sle[DATA][i]
            if sle[CURSOR] = i then
                pos = length( data )
            end if
        end if
    end for
    
    -- if cursor not set, was off end
    if pos = 0 then
        pos = length( data ) + 1
    end if

    sle[DATA] = data
    sle[CURSOR] = pos
    
    return sle
    
end function

-----------------------------------------------------------------------------
function data_to_masked( sequence sle )

    -- convert data to masked format

    integer pos, cursor
    sequence mask, data, masked

    mask = sle[MASK]
    data = sle[DATA] & repeat( ' ', length( mask ) )
    masked = ""        
    pos = 0

    cursor = 0
    for char = 1 to length( mask ) do
        -- data character?
        if mask[char] = '9' then
            pos = pos + 1
            masked = masked & data[pos]
            if pos = sle[CURSOR] then
                cursor = char
            end if
        else
            masked = masked & mask[char]
        end if
    end for

    -- cursor set?
    if cursor = 0 then
        cursor = length( masked ) + 1        
    end if

    sle[CURSOR] = cursor
    sle[DATA] = masked
            
    return sle
    
end function


-----------------------------------------------------------------------------    
procedure correct_mask_focus()
    
    -- focusing on an SLE moves to the first character
    -- on a mask, this may not be the correct. this moves
    -- the cursor to the first mask character

    integer focus, pos
    sequence result, mask

    focus = get_dialog( FOCUS )

    -- leave if not masked    
    if integer( get_attrib( focus, MASK ) ) then
        return
    end if

    -- get attributes
    mask = get_attrib( focus, MASK )
    pos = get_attrib( focus, CURSOR )


    -- position 1?   
    if pos = 1 then
        -- mask char?
        if mask[1] != '9' then
            -- move to mask char
            set_attrib( focus, CURSOR, find( '9', mask ) )
            
            -- kill anchor
            set_attrib( focus, ANCHOR_COL, 0 )
            
            -- redisplay         
            result = draw_sle( get_control( focus ) )

        end if        
    end if
    
end procedure


    
-----------------------------------------------------------------------------
function mask_edit( integer key )

    -- same as edit dialog, but handles masks

    integer pos, control
    sequence sle, data
    object mask

    -- current control        
    control = get_dialog( FOCUS )
    
    -- is there a mask?
    mask = get_attrib( control, MASK )
    if integer( mask ) then
    
        -- no action
        return key
        
    end if

    -- copy the SLE to a local
    sle = get_control( control )
    
    -- convert the SLE to raw form
    sle = masked_to_data( sle )

    -- get the cursor position
    pos = sle[CURSOR]
    
    
    -- the mask edit loop
    if natural_key( key ) then

        -- within data?
        if pos <= length( sle[DATA] ) then

            if  key >= '0' 
            and key <= '9' then        

                -- which mode?
                if editMode = OVERWRITE_MODE then
            
                    -- update data
                    sle[DATA][pos] = key          
                
                else  

                    -- insert character, drop one off end
                    sle[DATA] = sle[DATA][1..pos-1] & 
                        key & sle[DATA][pos..length( sle[DATA] )-1 ]
                        
                end if       
                
                -- move cursor
                sle[CURSOR] = pos + 1
                
            end if
        end if
        
        -- kill the key
        key = 0

    elsif key = END then
        sle[CURSOR] = pos + 1
        
        -- kill the key
        key = 0
        
            
    elsif key = LEFT then

        if pos > 1 then
            sle[CURSOR] = pos - 1
        end if
            
        -- kill the key
        key = 0
        
    elsif key = HOME then
        sle[CURSOR] = 1
        
        -- kill the key
        key = 0
            
    elsif key = RIGHT then               

        -- move right
        if pos <= length( sle[DATA] ) then
            sle[CURSOR] = pos + 1
        end if
            
        -- kill the key
        key = 0

    elsif key = DELETE then               
    
        if pos <= length( sle[DATA] ) then
            sle[DATA] = sle[DATA][1..pos-1] 
                & sle[DATA][pos+1..length(sle[DATA])]
        end if
                        
        key = 0
        
    elsif key = INSERT then
        -- ignore
        key = 0

    elsif key = BACKSPACE then
        
        if pos > 1 then
            sle[DATA] = sle[DATA][1..pos-2] 
                & sle[DATA][pos..length(sle[DATA])]
                
            sle[CURSOR] = pos - 1
        end if

        -- kill the key
        key = 0
    
    end if
    
    -- kill any anchors
    sle[ANCHOR_COL] = 0

    -- convert the data back to masked
    sle = data_to_masked( sle )

    -- normal processing?
    if key != 0 then
        -- regular key handling...
        sle = edit_xle( sle, key )
    end if

    -- make changes permanent
    set_control( control, sle )
    
    -- return key
    return key
    
end function


-----------------------------------------------------------------------------
procedure mask_edit_dialog()

    -- wait for a key for the dialog

    integer keyPressed

    -- fix focus on masked SLE
    correct_mask_focus()
    
    -- display cursor
    if VGA_MODE then
        -- only show if in MLE or SLE
        if get_attrib( focus(), TYPE ) = SLE
        or get_attrib( focus(), TYPE ) = MLE then
            show_graphic_cursor()
        end if
    else
        show_cursor()                               -- display the cursor
    end if

    keyPressed = watch_keys()                       -- wait for a key
    set_dialog( KEY_PRESSED, keyPressed )           -- save it
    
    -- hide cursor
    if VGA_MODE then
        -- only show if in MLE or SLE
        if get_attrib( focus(), TYPE ) = SLE
        or get_attrib( focus(), TYPE ) = MLE then
            hide_graphic_cursor()
        end if
    else
        hide_cursor()
    end if

    -- handle mask, if any
    keyPressed = mask_edit( keyPressed )        
    
    -- give to dialog
    give_dialog_key( keyPressed )

end procedure


-----------------------------------------------------------------------------    
global procedure move_data_into_masks()

    -- move data of masked SLEs into masked format

    integer pos
    
    for control = 1 to length( MAIN_DIALOG[ITEMS] ) do
    
        -- masked SLE?
        if  get_attrib( control, TYPE ) = SLE
        and sequence( get_attrib( control, MASK ) ) then

            -- convert data to masked format
            set_control( control, data_to_masked( get_control( control ) ) )
            
            -- fix the size
            set_attrib( control, WIDE, length( get_attrib( control, MASK ) ) )
            
        end if            
    end for


end procedure

-----------------------------------------------------------------------------    
global procedure move_masks_into_data()

    -- move data of masked SLEs back to raw data

    integer pos
    
    for control = 1 to length( MAIN_DIALOG[ITEMS] ) do
    
        -- masked SLE?
        if  get_attrib( control, TYPE ) = SLE
        and sequence( get_attrib( control, MASK ) ) then

            -- convert data to masked format
            set_control( control, masked_to_data( get_control( control ) ) )
            
        end if            
    end for


end procedure



-----------------------------------------------------------------------------    
global procedure mask_run_dialog()    

    -- dialog run routine, with 

    -- set masks
    move_data_into_masks()

    -- display the objects
    show_dialog()

    -- loop forever
    while get_dialog(EXIT_STATE) = MORE do
        mask_edit_dialog()      -- process keys
    end while

    -- restore masked data to raw form
    move_masks_into_data()

end procedure

