-- Menus for Euphoria

include graphics.e      -- screen display
include get.e           -- screen display
include image.e         -- screen display
include wildcard.e      -- searching
include keys.e          -- extended keys
include control.e       -- control constants
include dialog.e        -- dialog box

----------------------------------------------------------------------------
global procedure draw_menu_bar( sequence menu )

    -- put up a blank bar
    position( 1, 1 )
    put_colors( OF_MENUBAR, repeat(  ' ', SCREEN_WIDE ) )

    -- put up the items
    for item = 1 to length(menu) do
        -- alias item
        if VGA_MODE then
            graphic_title( menu[item][TITLE], OF_MENUBAR, 1, menu[item][COL]+1, menu[item][HOT], 0 )
        else
            position( 1, menu[item][COL]+1 )
            put_colors( OF_MENUBAR, menu[item][TITLE] )
        end if
    end for

end procedure


-----------------------------------------------------------------------------
global procedure draw_lit_menu_bar( sequence menu )    

    -- put up the items
    for item = 1 to length(menu) do
        -- alias item
        if not VGA_MODE then
            position( 1, menu[item][COL]+menu[item][HOT] ) 
            put_colors( { OF_SELECTED[1], OF_MENUBAR[2] }, menu[item][TITLE][menu[item][HOT]] )
        end if
    end for

end procedure


-----------------------------------------------------------------------------
procedure draw_menu_item( sequence menu, integer dropDown, 
                            integer item, integer lit )

    integer col, hot, check
    sequence colors
    -- alias menu[dropDown][ITEMS][item]

    -- the column
    col = menu[dropDown][COL]

    -- check if it's a valid item

    if item < 1 then                                -- nothing selected/bad
        return
    elsif item >
        length( menu[dropDown][ITEMS][item] ) then  -- past end
        return
   elsif menu[dropDown][ITEMS][item][TITLE][1] = '-' then                       -- seperator
        return
    elsif length( menu[dropDown][ITEMS][item][TITLE] ) = 0 then                 -- empty
        return
    end if

    -- the color
    if lit then
        colors = { OF_MENUBAR[2], OF_MENUBAR[1] }        
    else
        if menu[dropDown][ITEMS][item][DIM] then
            -- dim/normal
            colors = { OF_DISABLED[1], OF_MENUBAR[2] }
        else
            -- normal
            colors = OF_MENUBAR
        end if
    end if

    -- the item
    if VGA_MODE then

        -- handle check
        if menu[dropDown][ITEMS][item][CHECKED] then
            check = 04
        else
            check =  ' '
        end if

        -- move hot key over by 1
        if menu[dropDown][ITEMS][item][HOT] then
            hot = menu[dropDown][ITEMS][item][HOT] + 1
        else
            hot = 0
        end if

        -- display
        graphic_title( check & menu[dropDown][ITEMS][item][TITLE]  & " ", colors,
                            item+2-VGA_MODE, col, hot, 0 )

    else

        -- the text
        position( item+2-VGA_MODE, col )

        if menu[dropDown][ITEMS][item][CHECKED] then
            -- with check
            put_colors( colors, {7} & menu[dropDown][ITEMS][item][TITLE] & " " )
        else
            -- without check
            put_colors( colors, " " & menu[dropDown][ITEMS][item][TITLE] & " " )
        end if

        -- hot key
        if  menu[dropDown][ITEMS][item][HOT] != 0
        and not menu[dropDown][ITEMS][item][DIM] then
            -- color
            if lit then
                -- reverse
                colors = { OF_ACCEL[1], OF_MENUBAR[1] }
            else
                -- normal
                colors = { OF_ACCEL[1], OF_MENUBAR[2] }
            end if
    
            -- display
            position( item+2, col+menu[dropDown][ITEMS][item][HOT] )
            put_colors( colors, menu[dropDown][ITEMS][item][TITLE][menu[dropDown][ITEMS][item][HOT]] )
    
        end if
    end if


    -- help, if selected and there is text
    if  lit
    and compare( menu[dropDown][ITEMS][item][HELP_TEXT], "" ) != 0 then

        -- bottom of the screen
        position( SCREEN_DEEP, 1 )
        put_colors( OF_STATUS, repeat(  ' ', SCREEN_WIDE ) )
        position( SCREEN_DEEP, 2 )
        put_colors( OF_STATUS, menu[dropDown][ITEMS][item][HELP_TEXT] )

    end if


end procedure


-----------------------------------------------------------------------------
procedure draw_dropdown( sequence menu, integer dropDown )

    integer r, c, wide, bottomLine
    sequence colors

    -- make "snappier"
    defer_display()
    
    -- menu bar
    draw_menu_bar( menu )

    -- first, take care of the menu bar
    -- alias menu[dropDown]

    -- get row and column
    r = 2
    c = menu[dropDown][COL]-1

    -- highlight menu bar
    if VGA_MODE then        
        graphic_title( " " & menu[dropDown][TITLE] & " ", { OF_MENUBAR[2], OF_MENUBAR[1] },
                            1, menu[dropDown][COL], 0, 0 )
    else
        position( 1, menu[dropDown][COL] ) 
        put_colors( { OF_MENUBAR[2], OF_MENUBAR[1] }, " " & menu[dropDown][TITLE] & " " )
    end if

    -- now, take care of the items
    -- alias menu[dropDown][ITEMS]

    -- get the width of the menu
    wide = length( menu[dropDown][ITEMS][1][TITLE] )

    if VGA_MODE then
        -- background
        graphic_menu( c, r, wide+3, r+length(menu[dropDown][ITEMS])-2 )
    else
        -- top of menu
        position( r, c )
        put_colors( OF_MENUBAR, "" & repeat( '', wide ) & "Ŀ" )
    end if

    for item = 1 to length(menu[dropDown][ITEMS]) do

        -- examine each item
        -- alias item

        -- position
        position( r+item-VGA_MODE, c+VGA_MODE )

        if menu[dropDown][ITEMS][item][TITLE][1] = '-' then            
            -- empty
            if VGA_MODE then
                put_colors( OF_MENUBAR,  repeat( '', wide+2 ) )
            else

                -- seperator
                put_colors( OF_MENUBAR,  "" & repeat( '', wide ) & "Ĵ" )

                -- shadow
                put_shadow( 2 )

            end if
        else
            if not VGA_MODE then

                -- frame
                put_colors( OF_MENUBAR,  " "  & repeat(  ' ', wide ) & " " )

                -- shadow
                put_shadow( 2 )

            end if

            -- menu text
            draw_menu_item( menu, dropDown, item, 0 )

        end if

        -- hot key?
        if menu[dropDown][ITEMS][item][HOT] != 0
        and not VGA_MODE
        and not menu[dropDown][ITEMS][item][DIM] then

            -- character
            position( r+item-VGA_MODE, c+menu[dropDown][ITEMS][item][HOT]+1 )
            put_colors( {OF_ACCEL[1], OF_MENUBAR[2]}, menu[dropDown][ITEMS][item][TITLE][menu[dropDown][ITEMS][item][HOT]] )

        end if


    end for

    if not VGA_MODE then
    
        -- calculate bottom row
        bottomLine = r + length(menu[dropDown][ITEMS]) + 1
    
        -- bottom line
        position( bottomLine, c )
        put_colors( OF_MENUBAR, "" & repeat( '', wide ) & "" )
    
        -- shadow
        put_shadow( 2 )
    
        -- shadow under the box
        position( bottomLine+1, c+2 )
        put_shadow( wide+4 )

    end if

    -- show it
    show_deferred()

end procedure

-----------------------------------------------------------------------------
function focused_menu_dropdown( sequence menu, integer dropDown )

    -- return the index of the menu dropdown clicked, or
    -- zero if it was out of range.

    integer wide

    -- alias menu[dropDown]

    -- get the width of the menu
    wide = length( menu[dropDown][ITEMS][1][TITLE] )
    
    if MOUSE_COL < menu[dropDown][COL]-2 then
        -- to the left of menu
        return 0
    
    elsif MOUSE_COL > menu[dropDown][COL]-2 + wide + 4 then
        -- to the right of the menu
        return 0

    elsif MOUSE_ROW = 1 then
        -- above menu
        return 0

    elsif MOUSE_ROW > length(menu[dropDown][ITEMS]) + 2 - VGA_MODE then
        -- below the menu
        return 0
    else
        -- must be in menu
        return MOUSE_ROW - 2 + VGA_MODE
    end if

end function

-----------------------------------------------------------------------------
function focused_menu_bar( sequence menu )
    -- return menu bar mouse clicked, if any

    if MOUSE_ROW = 1 then
        -- look at each menu item
        for i = 1 to length( menu ) do
            -- alias menu[i]

            -- on item?
            if  MOUSE_COL >= menu[i][COL]
            and MOUSE_COL <= menu[i][COL]+length(menu[i][TITLE]) then

                -- return item number
                return i

            end if
        end for
    end if

    -- no match
    return 0

end function


----------------------------------------------------------------------------
function save_dropdown_image( sequence menu, integer i )

    integer row, col, deep, wide

    -- get size of menu
    row = 1
    col = menu[i][COL]-1
    wide = length( menu[i][ITEMS][1][TITLE] )+5
    deep = length( menu[i][ITEMS] )+3

    -- return the start, and an image
    return { {row, col},
        save_clean_image( {row,col}, {row+deep, col+wide} ) }

end function

----------------------------------------------------------------------------
procedure display_dropdown_image( sequence image )

    -- return the start, and an image
    display_clean_image( image[1], image[2] )

end procedure

-----------------------------------------------------------------------------
function run_dropdown( sequence menu, integer i, integer menuItem )
    sequence cleanScreen
    integer keyPress, priorMenuItem, picked, this


    -- remove highlight from menu bar
    draw_menu_bar( menu )

    -- save the screen
    cleanScreen = save_dropdown_image( menu, i )
    
    -- display the dropdown menu
    draw_dropdown( menu, i )

    -- draw a menu dropdown
    priorMenuItem = 0

    picked = 0

    while not picked do
        
        -- alias menu[i]

        if priorMenuItem != menuItem then
            -- erase prior
            draw_menu_item( menu, i, priorMenuItem, 0 )

            -- draw menu item
            draw_menu_item( menu, i, menuItem, 1 )
            priorMenuItem = menuItem
        end if

        -- wait for a key
        keyPress = watch_keys()

        -- process if
        -- case keyPress

        if keyPress = MOUSE_CLICK or keyPress = MOUSE_DRAG then 
   
            -- clicked another dropdown?
            this = focused_menu_bar( menu )

            if this != 0
            and this != i then
                -- show new dropdown

                -- new dropdown
                i = this

                -- top menu item
                menuItem = 0

                -- restore screen
                display_dropdown_image( cleanScreen )

                -- save background
                cleanScreen = save_dropdown_image( menu, i )

                -- draw new menu
                draw_dropdown( menu, i )


            else

                -- inside current dropdown?
                this = focused_menu_dropdown( menu, i )
                if this = 0 then

                    -- outside dropdown
                    menuItem = 0
                
                else

                    -- pick it
                    menuItem = this

                end if
            end if

        elsif keyPress = MOUSE_UP then 

            -- still on the menu bar?
            if MOUSE_ROW = 1 then
                menuItem = 1
            else
                -- exit if not item is picked (outside menu)
                if menuItem = 0 then
                    picked = 1

                elsif not menu[i][ITEMS][menuItem][DIM] then
                    -- return the item, since it's not dim
                    picked = 1
                end if
            end if   

        elsif keyPress = ENTER then 
            -- can't select if it's dimmed
            if not menu[i][ITEMS][menuItem][DIM] then
                picked = 1
            end if

        elsif keyPress = ESC then 
            picked = 1
            menuItem = 0

        elsif keyPress = RIGHT then 

            -- move to next menu on the right
            i = i + 1
            if i > length( menu ) then
                i = 1
            end if

            -- top menu item
            menuItem = 1

            -- restore screen
            display_dropdown_image( cleanScreen )

            -- save background
            cleanScreen = save_dropdown_image( menu, i )

            -- draw new menu
            draw_dropdown( menu, i )            

            -- force redraw
            priorMenuItem = 0


        elsif keyPress = LEFT then 
                        
            -- restore screen
            display_dropdown_image( cleanScreen )

            -- move to next menu on the left
            i = i - 1
            if i < 1 then
                i = length( menu )
            end if

            -- save background screen
            cleanScreen = save_dropdown_image( menu, i )

            -- draw new dropdown
            draw_dropdown( menu, i )
            

            -- top menu item
            menuItem = 1

            -- force redraw
            priorMenuItem = 0

        elsif keyPress = UP then 

            menuItem = menuItem - 1
            while 1 do 
                if menuItem < 1 then
                    -- move to the bottom
                    menuItem = length( menu[i][ITEMS] )
                    exit

                elsif menu[i][ITEMS][menuItem][TITLE][1] = '-' then

                    -- move up
                    menuItem = menuItem - 1

                else
                    -- on good item
                    exit
                end if
            end while

        elsif keyPress = DOWN then 
                    
            menuItem = menuItem + 1
            while 1 do 
                if menuItem > length( menu[i][ITEMS] ) then
                    -- move to the top
                    menuItem = 1

                elsif menu[i][ITEMS][menuItem][TITLE][1] = '-' then

                    -- move down
                    menuItem = menuItem + 1

                else
                    -- on good item
                    exit
                end if
            end while

        else
            -- hot key?
            for item = 1 to length(menu[i][ITEMS]) do

                -- look at the dropdown menu item
                -- alias item

                -- hot key?
                if menu[i][ITEMS][item][HOT] > 0 then
                    if compare( upper( {menu[i][ITEMS][item][TITLE][menu[i][ITEMS][item][HOT]]} ),
                                upper( {keyPress} ) ) = 0 then
                                    
                        picked = 1
                        menuItem = item

                    end if
                end if

            end for

        end if

    end while

    -- restore screen
    display_dropdown_image( cleanScreen )

    -- return value
    if menuItem = 0 then
        return 0
    else
        return 1000 + (i * 100) + menuItem
    end if

end function


-----------------------------------------------------------------------------
function run_menu_bar( sequence menu )

    sequence cleanScreen
    integer this, prior, keyPress, tmp
    -- alias menu

    -- save menu bar
    cleanScreen = save_clean_image( {1,1},{1,SCREEN_WIDE} )

    this = 1
    prior = 0

    while 1 do 

        -- check for movement to another menu bar
        if this != prior then
    
            -- erase prior?
            if prior != 0 then
    
                -- restore menu bar
                display_clean_image( {1, 1}, cleanScreen )
    
            end if

            -- set prior to current
            prior = this
    
            -- light the current bar
            -- alias menu[this]
    
            -- text
            position( 1, menu[this][COL] )
            put_colors( { OF_MENUBAR[2], OF_MENUBAR[1] },
                 " " & menu[this][TITLE] & " " )
    
            -- hot key
            position( 1, menu[this][COL]+menu[this][HOT] ) 
            put_colors( {OF_ACCEL[1], OF_MENUBAR[1]}, menu[this][TITLE][menu[this][HOT]] )
    
        end if

        -- alias menu                  -- focus on the menu
        keyPress = watch_keys()     -- get a key

        -- process the key
        -- case keyPress

        if keyPress = MOUSE_CLICK then 

            -- menu item clicked?
            tmp = focused_menu_bar( menu )
            if tmp != 0 then
                -- run menu, no selection
                return run_dropdown( menu, tmp, 0 )
            end if

        elsif keyPress = ESC or keyPress = ALT_RELEASED then 
            -- restore menu bar
            draw_menu_bar( menu )

            -- exit
            return 0

        elsif keyPress = LEFT then 
            -- move left
            if this = 1 then
                -- wrap
                this = length( menu )
            else
                -- move left
                this = this - 1
            end if
                
        elsif keyPress = RIGHT then 
            -- move right
            if this = length( menu ) then
                -- wrap
                this = 1 
            else
                -- move right
                this = this + 1
            end if

        elsif keyPress = UP or keyPress = DOWN or keyPress = ENTER then 

            -- run the dropdown menu, first item selected
            return run_dropdown(menu, this, 1 )

        else

            for item = 1 to length(menu) do
                -- alias item
                -- matches key pressed?
                if compare( upper( {menu[item][TITLE][menu[item][HOT]]} ),
                            upper( {keyPress} ) ) = 0 then                    
                    -- return menu number
                    return run_dropdown(menu, item, 1) -- first item list
                end if
            end for

        end if
    end while

end function

-----------------------------------------------------------------------------
global function run_menu( sequence menu, integer keyPress )

    -- run the menu. passing the keypress allows "pre-processing" the
    -- menu for special alt keys (for example, foreign keyboards, or
    -- special Alt keys in programs).

    sequence    cleanScreen, result
    integer     hotKey, this

    -- hide cursor
    hide_cursor()

    -- highlight the menu
    draw_lit_menu_bar( menu )

    -- handle keys
    while 1 do 
        if keyPress = ALT_RELEASED then
            -- run the menu bar
            return run_menu_bar( menu )

        elsif keyPress = MOUSE_CLICK then

            -- menu item clicked?
            this = focused_menu_bar( menu )
            if this != 0 then
                -- dropdown, no item selected
                return run_dropdown( menu, this, 0 )
            end if
            
        else
            -- scan through menu
            for item = 1 to length(menu) do
                -- alias item
                -- matches key pressed?
                if normal_to_alt( menu[item][TITLE][menu[item][HOT]] ) = keyPress then                    
                    -- return menu number
                    return run_dropdown( menu, item, 1 )   -- item 1 lit
                end if
            end for
        end if

        -- wait for a key
        keyPress = watch_keys()

    end while

    -- restore screen
    draw_menu_bar( menu )    

    -- return zero
    return 0

end function

-----------------------------------------------------------------------------
global function give_menu( sequence menu, integer key )

    -- top level menu call
    -- if menu uses key, it will replace it with it's own

    integer pick

    if key = MOUSE_CLICK then

        -- clicked menu bar?
        pick = focused_menu_bar( menu )
        if pick != 0 then
            -- run dropdown, no item picked
            return run_dropdown( menu, pick, 0 )
        end if


    elsif key = ALT_PRESSED then
        -- run menu
        return run_menu( menu, 0 )

    end if

    -- do nothing with key
    return key

end function

