                --------------------
                -- Mouse Routines --
                --------------------
-- You need DOS mouse support for these routines.

-- modified by Pete Eberlein <xseal@harborside.com>
-- added "set_svga_mouse()" for SVGA mouse cursor  

include machine.e
include image.e
include graphics.e

-- Mouse Event Numbers:
global constant MOVE = 1,       -- track every movement of the mouse
                LEFT_DOWN = 2,  -- the rest are button pressing/releasing
                LEFT_UP = 4,
                RIGHT_DOWN = 8,
                RIGHT_UP = 16,
                MIDDLE_DOWN = 32,
                MIDDLE_UP = 64

constant M_GET_MOUSE = 14,
         M_MOUSE_EVENTS = 15,
         M_MOUSE_POINTER = 24

integer SVGA_MOUSE
    SVGA_MOUSE = 0

sequence m_back, m_pos, m_rawpos, m_bord
integer m_visible, m_buts
m_visible = 1  -- not visible
m_buts = 0
constant m_mask = 255*{
                { 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1},
                { 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1},
                { 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1},
                { 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1},
                { 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1},
                { 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1},
                { 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1},
                { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1},
                { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1},
                { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                { 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1},
                { 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1},
                { 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1},
                { 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1},
                { 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1},
                { 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1}}
constant m_map = {
                { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                { 0,15, 0, 0, 0, 0, 0, 0, 0, 0, 0},
                { 0,15,15, 0, 0, 0, 0, 0, 0, 0, 0},
                { 0,15,15,15, 0, 0, 0, 0, 0, 0, 0},
                { 0,15,15,15,15, 0, 0, 0, 0, 0, 0},
                { 0,15,15,15,15,15, 0, 0, 0, 0, 0},
                { 0,15,15,15,15,15,15, 0, 0, 0, 0},
                { 0,15,15,15,15,15,15,15, 0, 0, 0},
                { 0,15,15,15,15,15,15,15,15, 0, 0},
                { 0,15,15,15,15,15, 0, 0, 0, 0, 0},
                { 0,15,15, 0,15,15, 0, 0, 0, 0, 0},
                { 0,15, 0, 0, 0,15,15, 0, 0, 0, 0},
                { 0, 0, 0, 0, 0,15,15, 0, 0, 0, 0},
                { 0, 0, 0, 0, 0, 0,15,15, 0, 0, 0},
                { 0, 0, 0, 0, 0, 0,15,15, 0, 0, 0},
                { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}}

constant mWide = length(m_map[1]),  mTall = length(m_map)


global function get_mouse()
-- report mouse events,
-- returns -1 if no mouse event,
-- otherwise returns {event#, x-coord, y-coord}
    object me
    sequence REGS, temp
    integer dy, dx, buts, bx1, bx2, bx3, bx4, tx1, tx2
    
    if SVGA_MOUSE then
        REGS = repeat(0, 10)
        REGS[REG_AX] = #B              -- get relative motion counters
        REGS = dos_interrupt(#33, REGS)
        temp = {REGS[REG_CX], REGS[REG_DX]}
        m_rawpos = m_rawpos + temp - 65536 * (temp > 32767)
        m_rawpos = m_rawpos * (m_rawpos >= 0) * (m_rawpos <= m_bord)
                 + m_bord * (m_rawpos > m_bord)
        
        REGS[REG_AX] = 3               -- get mouse buttons in BX
        REGS = dos_interrupt(#33, REGS)
                
        me = 0 & m_rawpos
        
        if me[2] != m_pos[1] or me[3] != m_pos[2] then
            me[1] = me[1] + MOVE
            if m_visible = 0 then
                temp = save_image(me[2..3], me[2..3]+{mWide,mTall}-1)
                dx = me[2] - m_pos[1]
                if (me[3] - m_pos[2] > -mTall and me[3] - m_pos[2] < mTall) 
                    and (dx > -mWide and dx < mWide) then
                        if dx >= 0 and dx < mWide then   -- mouse moved to the right
                            tx1 = 1           tx2 = mWide - dx
                            bx1 = 1 + dx      bx2 = mWide
                            bx3 = 1           bx4 = dx
                            dx = 0
                        elsif dx > -mWide and dx < 0 then -- mouse moved to the left
                            tx1 = 1 - dx      tx2 = mWide
                            bx1 = 1           bx2 = mWide + dx
                            bx3 = mWide+1+dx  bx4 = mWide
                            dx = mWide + dx
                        end if
                    for y = 1 to length(m_back) do 
                        dy = (m_pos[2] + y) - me[3]
                        if dy >= 1 and dy <= mTall then
                            temp[dy][tx1..tx2] = m_back[y][bx1..bx2]
                            pixel(m_back[y][bx3..bx4], m_pos+{dx,y-1})
                        else
                            pixel(m_back[y], m_pos+{0,y-1})
                        end if
                    end for
                else  -- new mouse image doesn't overlap old background
                    display_image(m_pos, m_back)
                end if
                display_image(me[2..3], or_bits(and_bits(temp, m_mask), m_map))
                m_back = temp
            end if
            m_pos = me[2..3]
            
        end if
        if REGS[REG_BX] != m_buts then
            buts = xor_bits(REGS[REG_BX], m_buts)
            me[1] = me[1] + 2*(and_bits(buts,1) * (1+and_bits(m_buts,1))  --LEFT
                             + and_bits(buts,2) * (2+and_bits(m_buts,2))  --RIGHT
                             + and_bits(buts,4) * (4+and_bits(m_buts,4))) --MIDDLE
            m_buts = REGS[REG_BX]
        end if
        if me[1] = 0 then
            me = -1
        end if
    else -- not SVGA_MOUSE
        me = machine_func(M_GET_MOUSE, 0)
    end if
    return me
end function

global procedure mouse_events(integer events)
-- select the mouse events to be reported by get_mouse()
-- e.g. mouse_events(LEFT_UP + LEFT_DOWN + RIGHT_DOWN)
    machine_proc(M_MOUSE_EVENTS, events)
end procedure

global procedure mouse_pointer(integer show_it)
-- show (1) or hide (0) the mouse pointer
    if SVGA_MOUSE then
        if show_it then
            if m_visible > 0 then
                m_visible = m_visible - 1
            end if
            if m_visible = 0 then
                m_back = save_image(m_pos, m_pos+{mWide,mTall}-1)
                display_image(m_pos, or_bits(and_bits(m_back, m_mask), m_map))
            end if
        else
            if m_visible = 0 then
                display_image(m_pos, m_back)
            end if
            m_visible = m_visible + 1
        end if
    else
        machine_proc(M_MOUSE_POINTER, show_it)
    end if
end procedure

global procedure set_svga_mouse()
    sequence vparams
    vparams = video_config()
        
    m_bord = {vparams[VC_XPIXELS], vparams[VC_YPIXELS]} - 1

    SVGA_MOUSE = 1
    
    m_pos = {0,0}
    m_rawpos = {0,0}
    m_back = {{}}
    
end procedure

