-- scroll bar routines
-- used by LIST and XLE

include keys.e          -- key definitions
include get.e           -- saving screen
include graphics.e      -- colors
include image.e         -- save_text_image()
include control.e       -- color scheme
include gui.e           -- gui interface

-----------------------------------------------------------------------------
global function current_line( sequence list )

    -- returns current item in list

    -- alias list
    
    return list[TOP_LINE] + list[ITEM] - 1

end function

-----------------------------------------------------------------------------
global function elevator_at( sequence list )

    -- return the position of the elevator car on the scroll bar

    integer this

    -- alias list

    -- answer is zero if length is shorter than depth
    if list[DEEP] < length( list[ITEMS] ) then

        -- current item
        this = current_line( list )

        -- position
        return floor( ( this/length(list[ITEMS]) ) * (list[DEEP]-3) ) + 1

    else

        return 0

    end if

end function


-----------------------------------------------------------------------------
global function draw_elevator( sequence list )
    
    -- draw an elevator
    
    integer elevator

    -- refer to the list
    -- alias list

    -- defer so overwritten thumb will not flash
    defer_display()

    -- position the elevator car
    if list[DEEP] < length( list[ITEMS]) then
        -- position is a ratio to where it is
        elevator = elevator_at( list )
    else
        elevator = 0
    end if

    -- add scroll bar if more items than length of list
    if list[REDRAW] then

        if VGA_MODE then

            -- draw scroll bar
            graphic_scroll_bar( list[ROW]-1, list[COL]+list[WIDE]-1, list[DEEP] )

        else

            for i = 1 to list[DEEP] do

                -- move to position
                position( list[ROW]+i-1, list[COL]+list[WIDE] )
                    
                -- determine what to draw
                if length( list[ITEMS] ) > list[DEEP] then
                    -- display scroll bar
                    -- case i
                    if i = 1 then 
                        put_colors( OF_SCROLL, {24} )
                    elsif i = list[DEEP] then 
                        put_colors( OF_SCROLL, {25} )
                    else
                        put_colors( OF_SCROLL, "" )
                    end if
                else
                    
                    -- no scroll bar
                    if list[TYPE] = MLE then
                        put_colors( OF_BORDER, "" )
                    else
                        put_colors( OF_DIALOG, "" )
                    end if
                                    
                end if
            end for
        end if
    end if

    -- need to display elevator?
    if  elevator != 0 then  -- there is an elevator to draw

        -- has it moved?
        if elevator != list[ELEVATOR]    -- elevator has moved
        or list[REDRAW]                  -- or been overwritten
        or VGA_MODE then            -- no optimization here yet
        
            -- erase prior elevator if moved
            if list[ELEVATOR] != 0 then
                -- clear old
                if VGA_MODE then
                    graphic_thumb( list[ROW]+list[ELEVATOR]-1, list[COL]+list[WIDE]-2, 0 )
                else
                    position( list[ROW]+list[ELEVATOR], list[COL]+list[WIDE] )
                    put_colors( OF_SCROLL, "" )
                end if
            end if

            -- new position
            list[ELEVATOR] = elevator
            if VGA_MODE then
                graphic_thumb( list[ROW]+list[ELEVATOR]-1, list[COL]+list[WIDE]-2, 1 )
            else
                position( list[ROW]+list[ELEVATOR], list[COL]+list[WIDE] )
                put_colors( OF_SCROLL, "" )
            end if

        end if

    end if

    -- draw screen
    show_deferred()

    -- return modified list
    return list

end function

-----------------------------------------------------------------------------
global function page_up( sequence list )

    -- alias list
    
    -- scroll
    list[TOP_LINE] = list [ TOP_LINE]- list[DEEP]
    if list[TOP_LINE] < 1 then
        -- move to top
        list[TOP_LINE] = 1
        list[ITEM] = 1            
    end if

    return list

end function


-----------------------------------------------------------------------------
global function page_down( sequence list )

    -- alias list

    -- need to scroll?        
    list[TOP_LINE] = list [ TOP_LINE]+ list[DEEP]
    if list[TOP_LINE] + list[DEEP] + VGA_MODE > length( list[ITEMS] ) then
        -- move to bottom
            if length( list[ITEMS] ) <= list[DEEP] then
                -- less than a full list; move to end
                list[TOP_LINE] = 1
                list[ITEM] = length( list[ITEMS] )
            else
                -- go to bottom of list
                list[TOP_LINE] = length( list[ITEMS] ) - list[DEEP] + 1
                list[ITEM] = list[DEEP]
            end if
    end if

    return list

end function

-----------------------------------------------------------------------------
function line_up( sequence list )

    -- move page up 1 line

    -- alias list

    if list[TOP_LINE] > 1 then

        -- scroll list down
        list[TOP_LINE] = list [ TOP_LINE]- 1

        -- move current line down
        if list[ITEM] < list[DEEP] then
            list[ITEM] = list [ ITEM]+ 1
        end if

    end if

        
    return list

end function


-----------------------------------------------------------------------------
function line_down( sequence list )

        -- move down 1 line

    -- alias list

    if list[TOP_LINE]+list[DEEP]-1 < length( list[ITEMS] ) then
        -- scroll up 1 line
        list[TOP_LINE] = list [ TOP_LINE]+ 1

        -- move current line
        if list[ITEM] > 1 then
            list[ITEM] = list [ ITEM]- 1
        end if

    end if

    return list

end function


-----------------------------------------------------------------------------
global function scrollbar_clicked( sequence list )

    -- move the elevator in the direction of the mouse

    integer mouseAt, mouseItem

    -- alias list

    -- position of elevator
    list[ELEVATOR] = elevator_at( list )

    -- no elevator to run if less items than size of list
    if length( list[ITEMS]) <= list[DEEP] then
        -- do nothing
        return list
    end if

    -- where is the mouse on the scroll bar
    mouseAt = floor (MOUSE_ROW - list[ROW])

    -- what item is the mouse on?
    mouseItem = floor( length(list[ITEMS]) * ( mouseAt / (list[DEEP] - 3 ) ) )

    -- where is the mouse on the scroll bar?

    if MOUSE_ROW = list[ROW]+list[ELEVATOR] then   -- on elevator

        -- move to top/bottom?
        if MOUSE_ROW = list[ROW]+1 then              -- below up arrow

            -- go to first item
            list = page_up( list )

        elsif MOUSE_ROW = list[ROW]+list[DEEP]-2 then     -- above down arrow

            -- go to last item
            list = page_down( list )

        end if

    elsif MOUSE_ROW = list[ROW] then                 -- up arrow

        -- move up 1 line
        list = line_up( list )

    elsif MOUSE_ROW = list[ROW]+list[DEEP]-1 then     -- down arrow

        -- move down 1 line
        list = line_down( list )

    elsif MOUSE_ROW < list[ROW]+list[ELEVATOR] then   -- above scroll bar

        -- need to scroll?
        if mouseItem < list[TOP_LINE] then
            -- scroll up towards item
            list = page_up( list )
        else
            -- move to item
            list[ITEM] = mouseItem + 1 - list[TOP_LINE]
        end if

    elsif MOUSE_ROW > list[ROW]+list[ELEVATOR] then   -- below scroll bar

        -- need to scroll?
        if mouseItem > list[TOP_LINE]+list[DEEP]-1 then
            -- scroll down towards item
            list = page_down( list )
        else
            -- move to item
            list[ITEM] = mouseItem + 1 - list[TOP_LINE]
        end if

    end if

    return list

end function

