-- svgamode.e
-- by Pete Eberlein
-- support tools for 256 color graphics mode 19

include mouse.e     -- Pete's SVGA mouse support tools

constant
    FONT_NAME = "CHICAGO.FNT"   -- set to blank if no font needed

include fonter.e    -- graphic_text, custom font

object result

-- mode 19 or 257      
constant MODE = 257

-- svga mode
    if graphics_mode(MODE) then
        puts( 1, "Unable to set SVGA mode on this machine!\n" )
        abort(0)
    end if

-- fonts for mode
    if MODE = 19 then
        -- mode 19 font size
            cTall = 8
            cWide = 8

    else
    
        -- mode 257 font size
            cTall = 16
            cWide = 8

    end if

-- svga mode
    VGA_MODE = 1
    set_svga_mouse()

-- load graphic font from RAM
    loadFont()

-- optional graphic font file
--    if length( FONT_NAME ) then
        loadFontFile(FONT_NAME)
--    end if

-- what's the configuration?
    result = video_config()

-- screen size
    SCREEN_WIDE = result[VC_COLUMNS]
    SCREEN_DEEP = result[VC_LINES] 
    
-- start of video ram
    if result[VC_COLOR] then
        SCREEN_MEMORY = #B8000
    else
        SCREEN_MEMORY = #B0000
    end if


----------------------------------------------------------------------------
global procedure defer_display()
    -- not used in graphic mode
end procedure


----------------------------------------------------------------------------
global procedure show_deferred()
    -- not used in graphic mode
end procedure


----------------------------------------------------------------------------
global procedure hide_mouse()
    -- hide the mouse in graphic mode
    -- used before a screen update

    mouse_pointer(0)

end procedure

----------------------------------------------------------------------------
global procedure show_mouse()
    -- display the mouse cursor, if it was hidden by hide_mouse()

    mouse_pointer(1)

end procedure


-----------------------------------------------------------------------------
global procedure put_graphic_colors( sequence colors, object text )

    -- emulate put_colors for graphics colors display

    integer l1, c1, l2, c2
    sequence cursor

    -- transform atoms into lists
    if atom( text ) then
        text = { text }
    end if

    if length( text ) = 0 then
        return
    end if
    
    -- transform sequences to sequences of 
    if atom(text[1]) then
        text = { text }
    end if

    cursor = get_position()

    l1 = (cursor[1]-1)*cTall
    c1 = (cursor[2]-1)*cWide
    l2 = l1 + cTall - 1
    c2 = c1 + ( length( text[1] ) * cWide ) - 1

    hide_mouse()

    for i = 1 to length( text ) do        
        -- text
        graphic_text( cursor[1]+i-1, cursor[2], 
            colors[1], colors[2], text[i] )
    end for
    
    if  cursor[1]+length( text ) <= SCREEN_DEEP
    and cursor[2]+length( text[1] ) <= SCREEN_WIDE then
        position( cursor[1]+length( text )-1, cursor[2]+length(text[1]))
    end if

    show_mouse()

end procedure



----------------------------------------------------------------------------
global procedure put_colors( sequence colors, object text )
        
    -- given a line of text, or sequence of lines, writes them to the
    -- screen with the given attributes

    sequence cursorAt
    integer thisRow, thisCol, bpl
    integer attribute, pointer

    -- vga gets another drive
    if VGA_MODE then
        put_graphic_colors( colors, text )
        return
    end if

    -- convert to a list if text is a single byte
    -- this is to save me from myself
    if atom( text ) then
            text = { text }
    end if

    -- skip if empty
    if length( text ) = 0 then
            return
    end if

    -- get cursor location
    cursorAt = get_position()
    thisRow = cursorAt[1]
    thisCol = cursorAt[2]

    -- if single line, convert to a list
    if atom( text[1] ) then

        -- make it an item in a list
        text = { text }

    end if

    -- combine the colors into a single attribute byte
    attribute = ( colors[2] * #10 ) + colors[1]

    -- video is located at HB8000 for color, and HB0000 for monochrome

    -- hide the mouse
    hide_mouse()

    -- bytes per line
    bpl = SCREEN_WIDE*2

    for row = 1 to length( text ) do
        -- calculate offset onto screen (line, col)
        pointer = SCREEN_MEMORY             -- start of video
            + ( (thisRow+row-2) * bpl )     --  bytes per line
            + ( (thisCol-1) * 2 )           -- 2 bytes per line

        for col = 1 to length( text[row] ) do
            poke( pointer, text[row][col] )
            poke( pointer+1, attribute )
            pointer = pointer + 2
        end for
    end for

    -- show the mouse
    show_mouse()

    -- new row
    thisRow = thisRow + length(text) - 1
    if thisRow > SCREEN_DEEP then 
        thisRow = SCREEN_DEEP 
    end if

    -- new column
    thisCol = thisCol + length( text[length(text)] )
    if thisCol > SCREEN_WIDE then 
        thisCol = SCREEN_WIDE 
    end if

    -- reposition cursor
    position( thisRow, thisCol )


end procedure


----------------------------------------------------------------------------
global function save_clean_image( sequence pos1, sequence pos2 )

    -- save as display_text_image(), but hides mouse cursor

    sequence s

    hide_mouse()
        pos1 = ({pos1[2]-(pos1[2]>1), pos1[1]-(pos1[1]>1)} - 1) * {cWide,cTall}
        pos2 = ({pos2[2], pos2[1]} * {cWide,cTall}) - 1
        s = save_image(pos1, pos2)
    show_mouse()

    return s

end function



----------------------------------------------------------------------------
global procedure display_clean_image( sequence pos, sequence img )

    -- save as display_text_image(), but hides mouse cursor

    hide_mouse()
        pos = ({pos[2]-(pos[2]>1), pos[1]-(pos[1]>1)} - 1) * {cWide,cTall}
        display_image(pos, img)
    show_mouse()

end procedure


----------------------------------------------------------------------------
global procedure put_shadow( integer i )
    -- not used in graphic modes
end procedure


----------------------------------------------------------------------------
global procedure toggle_blinking( integer flag )
    -- not used in graphic modes
end procedure

