-- textmode.e
-- support tools for text mode

without warning
with trace

----------------------------------------------------------------------------
-- Tools for deferred displaying

object result

-- font size
    cTall = 16
    cWide = 8

-- VGA mode flag
    VGA_MODE = 0
    
-- attempt to set to 25 lines
    result = text_rows( 25 )
            
-- what's the configuration?
    result = video_config()

-- screen size
    SCREEN_WIDE = result[VC_COLUMNS]
    SCREEN_DEEP = result[VC_LINES] 
    
-- start of video ram
    if result[VC_COLOR] then
        -- color VGA
        SCREEN_MEMORY = #B8000
    else
        -- mono VGA
        SCREEN_MEMORY = #B0000
    end if


----------------------------------------------------------------------------
global procedure defer_display()

    -- route writes to the buffer, not the display

    -- set flag
    defer_display_flag = defer_display_flag + 1

    -- skip the rest if already deferred
    if defer_display_flag = 1 then
    
        -- hide cursor for the remainder of the screen writes
        hide_cursor()
    
        -- hide the mouse
        mouse_pointer(0)
    
        -- copy the screen to the buffer
        defer_display_buffer =
            save_text_image( {1, 1}, {SCREEN_DEEP, SCREEN_WIDE} )
    
        -- restore the mouse
        mouse_pointer(1)

    end if


end procedure


----------------------------------------------------------------------------
global procedure show_deferred()

    -- show deferred screen

    -- anything deferred?
    if defer_display_flag > 0 then
        -- pop a level
        defer_display_flag = defer_display_flag - 1
    else
        -- nothing deferred
        return
    end if

    -- at bottom level?
    if defer_display_flag = 0 then           

        -- clear the flag
        defer_display_flag = 0

        -- hide the mouse
        mouse_pointer(0)

        -- restore screen
        display_text_image( {1, 1}, defer_display_buffer )

        -- show mouse
        mouse_pointer(1)

        -- release memory
        defer_display_buffer = ""

    end if

end procedure


----------------------------------------------------------------------------
global procedure put_deferred( integer attribute, sequence text )

    -- write the text to the buffer, not the screen

    sequence cursorAt
    integer row, col, ptr

    -- get cursor location
    cursorAt = get_position()
    row = cursorAt[1]
    col = cursorAt[2]

    for i = 1 to length( text ) do
        ptr = (col*2)-1
        for j = 1 to length( text[i] ) do

            -- char
            defer_display_buffer[row+i-1][ptr] = text[i][j]

            -- attribute
            defer_display_buffer[row+i-1][ptr+1] = attribute

            -- advance pointer
            ptr = ptr + 2

        end for
    end for    

end procedure

----------------------------------------------------------------------------
global procedure hide_mouse()
    -- only used in graphics mode

end procedure

----------------------------------------------------------------------------
global procedure show_mouse()
    -- only used in graphics mode

end procedure


----------------------------------------------------------------------------
global procedure put_colors( sequence colors, object text )
        
    -- given a line of text, or sequence of lines, writes them to the
    -- screen with the given attributes

    sequence cursorAt
    integer thisRow, thisCol, bpl
    integer attribute, pointer

    -- convert to a list if text is a single byte
    -- this is to save me from myself
    if atom( text ) then
            text = { text }
    end if

    -- skip if empty
    if length( text ) = 0 then
            return
    end if

    -- get cursor location
    cursorAt = get_position()
    thisRow = cursorAt[1]
    thisCol = cursorAt[2]

    -- turn off screen wrapping
    wrap( 0 )

    -- if single line, convert to a list
    if atom( text[1] ) then

        -- make it an item in a list
        text = { text }

    end if

    -- combine the colors into a single attribute byte
    attribute = ( colors[2] * #10 ) + colors[1]

    -- defer, or write directly to video
    if defer_display_flag then

        -- write to buffer
        put_deferred( attribute, text )

    else
        -- video is located at HB8000 for color, and HB0000 for monochrome

        -- hide the mouse
        mouse_pointer(0)

        -- bytes per line
        bpl = SCREEN_WIDE*2

        for row = 1 to length( text ) do
            -- calculate offset onto screen (line, col)
            pointer = SCREEN_MEMORY             -- start of video
                + ( (thisRow+row-2) * bpl )     --  bytes per line
                + ( (thisCol-1) * 2 )           -- 2 bytes per line

            for col = 1 to length( text[row] ) do
                poke( pointer, text[row][col] )
                poke( pointer+1, attribute )
                pointer = pointer + 2
            end for
        end for

        -- show the mouse
        mouse_pointer(1)

    end if

    -- new row
    thisRow = thisRow + length(text) - 1
    if thisRow > SCREEN_DEEP then 
        thisRow = SCREEN_DEEP 
    end if

    -- new column
    thisCol = thisCol + length( text[length(text)] )
    if thisCol > SCREEN_WIDE then 
        thisCol = SCREEN_WIDE 
    end if

    -- reposition cursor
    position( thisRow, thisCol )


end procedure

-----------------------------------------------------------------------------
global procedure put_shadow( integer wide )
        
    -- draw a shadow at the current position

    sequence cursorAt
    integer thisRow, thisCol, bpl
    integer attribute, pointer

    -- combine the colors into a single attribute byte
    attribute = ( BLACK * #10 ) + GRAY

    -- bytes per line 
    bpl = SCREEN_WIDE * 2

    -- get cursor location
    cursorAt = get_position()
    thisRow = cursorAt[1]
    thisCol = cursorAt[2]


    if defer_display_flag then

        -- write to deferred buffer

        pointer = (thisCol*2)
        for i = 1 to wide do
                                            
            -- attribute
            defer_display_buffer[thisRow][pointer] = attribute

            -- advance pointer
            pointer = pointer + 2

        end for
            
    else

        -- poke into video

        -- calculate offset onto screen (line, col)
        pointer = SCREEN_MEMORY             -- start of video
            + ( (thisRow-1) * bpl )         -- SCREEN_WIDE*2 bytes per line
            + ( (thisCol-1) * 2 )           -- 2 bytes per line

        -- hide the mouse
        mouse_pointer(0)
                                    
        for col = 1 to wide do
            poke( pointer+1, attribute )
            pointer = pointer + 2
        end for

        -- show the mouse
        mouse_pointer(1)

    end if

    -- new column
    thisCol = thisCol + wide
    if thisCol > SCREEN_WIDE then thisCol = SCREEN_WIDE end if

    -- reposition cursor
    position( thisRow, thisCol )

end procedure

----------------------------------------------------------------------------
global function save_clean_image( sequence pos1, sequence pos2 )

    -- save as display_text_image(), but hides mouse cursor

    sequence s

    mouse_pointer(0)                    -- hide mouse pointer
    s = save_text_image( pos1, pos2 )   -- get text image
    mouse_pointer(1)                    -- restore mouse pointer

    return s

end function



----------------------------------------------------------------------------
global procedure display_clean_image( sequence pos, sequence img )

    -- save as display_text_image(), but hides mouse cursor

    mouse_pointer(0)                    -- hide mouse pointer
    display_text_image( pos, img )      -- draw
    mouse_pointer(1)                    -- restore mouse pointer

end procedure


----------------------------------------------------------------------------
global procedure toggle_blinking( integer flag )
        sequence  inReg, outReg

        inReg = repeat( 0, 10 )         -- initialize register list
        inReg[REG_AX] = #1003           -- function 10, subfunction 03
        inReg[REG_BX] = flag            -- enable/disable intensity
        outReg = dos_interrupt( #10, inReg )  -- bios call

end procedure

----------------------------------------------------------------------------
global procedure graphic_text( atom r, atom c, atom c1, atom c2, sequence s )

    -- dummied for compatibility
    
end procedure

-- turn on intense colors
        toggle_blinking(0)

