-- 16color.e
-- support tools for 16 color graphics mode

without warning
with trace

constant VGA_ADDRESS = #A0000   -- vga port address

object result

-- mode 18
    
-- font size
    cTall = 16
    cWide = 8
    
-- vga mode
    if graphics_mode(18) then
        puts( 1, "Unable to set VGA mode on this machine!\n" )
        abort(0)
    end if
    
    VGA_MODE = 1        

            
-- what's the configuration?
    result = video_config()

-- screen size
    SCREEN_WIDE = result[VC_COLUMNS]
    SCREEN_DEEP = result[VC_LINES] 
    
-- start of video ram
    if result[VC_COLOR] then
        SCREEN_MEMORY = #B8000
    else
        SCREEN_MEMORY = #B0000
    end if


----------------------------------------------------------------------------
global procedure defer_display()
    -- not needed in graphics mode
end procedure


----------------------------------------------------------------------------
global procedure show_deferred()
    -- not needed in graphics mode
end procedure

----------------------------------------------------------------------------
global procedure hide_mouse()
    -- hide the mouse in graphic mode
    -- used before a screen update

    mouse_pointer(0)

end procedure

----------------------------------------------------------------------------
global procedure show_mouse()
    -- display the mouse cursor, if it was hidden by hide_mouse()

    mouse_pointer(1)

end procedure



-----------------------------------------------------------------------------
global procedure graphic_text( integer row, integer col,
                    integer fColor, integer bColor, sequence char )

    -- display string(s) on screen in given fore/background colors

    object    lowMem
    integer   len
    sequence  inReg, outReg, cursor
    atom      ptr

    hide_mouse()

    len = length( char )
    lowMem = allocate_low( len )            -- get memory    

    inReg = repeat( 0, 10 )                 -- initialize register list

    -- clear prior text
    poke( lowMem, repeat( 219, len ) )

    inReg[REG_AX] = #1300                   -- tty string write
    inReg[REG_BX] = bColor                  -- video page/color
    inReg[REG_CX] = len                     -- number of characters to print
    inReg[REG_DX] = ((row-1)*#100)+(col-1)  -- position of cursor
    inReg[REG_ES] = floor(lowMem / 16)      -- segment of low_mem (low memory)
    inReg[REG_BP] = remainder(lowMem, 16)   -- offset of low_mem
    outReg = dos_interrupt( #10, inReg )    -- bios call

    -- OR it with the background to mask the letters to BLACK
    poke( lowMem, char )

    inReg[REG_BX] = #F0 + bColor            -- video page/color
    outReg = dos_interrupt( #10, inReg )    -- bios call
    
    -- next, logical OR it with the black text for the
    -- correct color
    inReg[REG_BX] = #F0 + fColor            -- video page/color
    outReg = dos_interrupt( #10, inReg )    -- bios call
    

    free_low( lowMem )  -- free the memory

    show_mouse()

end procedure



-----------------------------------------------------------------------------
global procedure put_colors( sequence colors, object text )

    -- display text using given colors

    integer l1, c1, l2, c2, trimFlag, char
    sequence cursor, reg

    -- transform atoms into lists
    if atom( text ) then
        text = { text }
    end if

    if length( text ) = 0 then
        return
    end if
    
    -- transform sequences to sequences of 
    if atom(text[1]) then
        text = { text }
    end if

    cursor = get_position()

    l1 = (cursor[1]-1)*cTall
    c1 = (cursor[2]-1)*cWide
    l2 = l1 + cTall - 1
    c2 = c1 + ( length( text[1] ) * cWide ) - 1
    
    for i = 1 to length( text ) do

        -- trim text?
        if cursor[2] + length( text[i] )-1 >= SCREEN_WIDE then
            text[i] = text[i][1..SCREEN_WIDE-cursor[2]+1]
            trimFlag = 1
        else
            trimFlag = 0
        end if

        if  cursor[1]+i-1 = SCREEN_DEEP
        and trimFlag=1 then

            -- display text, less 1 character
            graphic_text( cursor[1]+i-1, cursor[2],
                colors[1], colors[2], text[i][1..length( text[i] )-1 ] )

            -- get last character                
            char = text[i][length( text[i] )]                

            -- put char with right color in last position            
            wrap(0)
            position( SCREEN_DEEP, SCREEN_WIDE )
            text_color( colors[2] )
            puts( 1, {219} )        
        
        else

            -- display text
            graphic_text( cursor[1]+i-1, cursor[2],
                colors[1], colors[2], text[i] )
                
        end if
                        
    end for

    -- new row
    cursor[1] = cursor[1]+length( text )-1
    if cursor[1] > SCREEN_DEEP then
        cursor[1] = SCREEN_DEEP
    end if

    -- new col
    cursor[2] = cursor[2]+length(text[1])
    if cursor[2] > SCREEN_WIDE then
        cursor[2] = SCREEN_WIDE
    end if

    position( cursor[1], cursor[2] )

end procedure


----------------------------------------------------------------------------
global function save_clean_image( sequence pos1, sequence pos2 )

    -- save a VGA graphic screen, return pointer to buffer

    -- This reads the graphic planes one at a time, and stores them
    -- to a memory location.

    -- Calls to the Graphics Controller Registers are made by  writing 
    -- to the Graphics Address Register (port 3CE) with an index value 
    -- indicating what register is to be accessed.
    -- Next, values can be written to that register by writing to the 
    -- Graphics Data Register (port 3CF).

    -- This function goes (more or less) as follows: 
    --
    -- 1. The amount of memory needed to save the screen is calculated.
    --    On a slow machine, a full screen restore is noticably slower
    --    than a partial screen restore.
    -- 
    -- 2. The Graphic Address Register is loaded with an index value of 4. 
    --    This selects the Read Map Select option, allowing one color plane 
    --    of memory to be read.
    --
    -- 3. The Read Map Select Register is written to with the value of
    --    the plane to be accessed.
    --
    -- 4. The plane is read an copied into memory.
    --
    -- 5. The process is repeated for each of the 4 color planes.
    --

    object ptr, offset, bytes, index

    hide_mouse()

    -- add one more line if not first line
    -- this takes care of title bars
    if pos1[1] > 1 then
        pos1[1] = pos1[1] - 1
    end if

    -- offset
    offset = (pos1[1]-1) * 80 * cTall

    -- bytes
    bytes = ((pos2[1]+1) * 80 * cTall ) - offset

    -- allocate memory for 4 planes
        ptr = allocate( bytes * 4 )

    -- prepare video memory for reading and read
    index = 0
    for i = 0 to 3 do
        Output( 4, #3CE )                       -- read planes
        Output( i, #3CF )                       -- which plane
        mem_copy( ptr+index, VGA_ADDRESS+offset, bytes )
        index = index + bytes
    end for

    show_mouse()

    return { ptr, offset, bytes }

end function



----------------------------------------------------------------------------
global procedure display_clean_image( sequence pos, sequence img )

    -- restore graphic screen from buffer, and release memory

    -- This is the complement to the graphic_screen_save. It uses
    -- the Sequencer Registers to talk to the VGA. First, the
    -- Sequencer Address Register (port 3C4) is written to specify the
    -- desired function. Next, the Sequence Data Register (port 3C5) is
    -- written to to specify parameters.

    -- For this procedure, the following steps are taken:
    --
    -- 1. The Sequencer Address Register is sent a 2, indicating a Mask Mode
    --    request.
    --
    -- 2. The Sequence Address Register is sent a bit pattern indicating
    --    what video memory color plane(s) to activate for the following
    --    memory move operations. Bits 0-3 correspond the the planes 0-3.
    --
    -- 3. A plane is selected, and the stored color plane is written to the
    --    VGA. This is repeated for each plane.
    --
    -- 4. 

    object ptr, offset, bytes, index

    hide_mouse()

    ptr = img[1]
    offset = img[2]
    bytes = img[3]

    -- prepare video memory for writing
    index = 0
    for i = 0 to 3 do
        Output( 2, #3C4 )           -- mask plane mode
        Output( power(2,i), #3C5 )  -- plane to activate
        mem_copy( VGA_ADDRESS+offset, ptr+index, bytes )
        index = index + bytes
    end for

    -- activate all the planes
    Output( 2, #3C4 )           -- mask planes
    Output( 1+2+4+8, #3C5 )     -- activate all planes
    
    free( ptr )

    show_mouse()

end procedure


----------------------------------------------------------------------------
global procedure put_shadow( integer i )
    -- not used in graphic modes
end procedure



----------------------------------------------------------------------------
global procedure toggle_blinking( integer flag )
    -- not used in graphic modes
end procedure





