-- multiple/single line edit for Euphoria
-- this combines the mle and sle code into one module

include keys.e          -- key definitions
include graphics.e      -- colors
include wildcard.e      -- upper/lower
include control.e
include color.e
include scroll.e        -- scroll bars


without warning
with trace

-- tab size    
    global integer TAB_SIZE
    TAB_SIZE = 4        -- logical tab size

-- used to determine if scrolling is needed
    integer oldFirstChar        -- prior value of column 1
    integer oldFirstLine        -- prior first line on screen
    integer redrawFlag          -- used to force screen redraw

-- for auto completion
-- from ED.EX
constant 
    expand_word = {"if", "for", "while", "elsif",
                   "procedure", "type", "function",
                   "global procedure", "global type", "global function" },

     expand_text = {" then",  " do",  " do",  " then",
                    "",  "",  "" ,
                    "", "", "" },

    expand_end  = {"end if", "end for", "end while", "",
                   "end procedure", "end type", "end function",
                   "end procedure", "end type", "end function" }
                       
integer 
    completingLine      -- true if line has been auto completed

    completingLine = 0                       


-- HELPER FUNCTIONS

----------------------------------------------------------------------------
global function detab_text( sequence line )

    -- clean up the tabs and carriage return from line

    integer at

    -- check for length zero
    if length( line ) = 0 then
        return line
    end if
    
    -- remove line feed, if on end of line
    if line[length(line)] = '\n' then
        line = line[1..length(line)-1]
    end if
    
    -- replace tabs
    while 1 do   
        -- look for tab
        at = find( '\t', line )
        
        -- found tab?
        if at then         
            -- replace with spaces
            line = line[1..at-1] & "        " & line[at+1..length(line)]
        else  
            -- no more tabs
            exit
        end if
    end while

    return line

end function

----------------------------------------------------------------------------
global function leading_spaces( sequence line )

    -- return count of spaces in front of line

    integer spaces

    -- quick test
    if length( line ) = 0 then
        return 0
    end if

    -- count spaces
    spaces = 0
    for char = 1 to length(line) do
        if line[char] !=  ' ' then exit        end if
        spaces = char
    end for

    return spaces

end function


----------------------------------------------------------------------------
global function retab_text( sequence line )

    -- convert leading spaces to tabs, add CR to end of line

    integer spaces, tabs
    sequence text

    -- count leading spaces
    spaces = leading_spaces( line )

    -- convert
    if spaces > TAB_SIZE then

        -- convert tabs to spaces
        tabs = floor( spaces / TAB_SIZE )

        -- remove spaces
        line = line[tabs*TAB_SIZE+1..length(line)]

        -- prepend tabs
        line = repeat( '\t', tabs ) & line

    end if

    -- add newline
    return line & "\n"
                
end function




-----------------------------------------------------------------------------
procedure mle_scroll( sequence mle, integer direction )

    integer offset, shift

    -- alias mle

    if direction = 1 then

        offset = 0
        shift = 1

    else

        offset = 1
        shift = -1

    end if

    display_clean_image( {mle[ROW]+offset+shift, mle[COL] },
         save_clean_image( {mle[ROW]+offset, mle[COL] },
            { mle[ROW]+mle[DEEP]-2+offset, mle[COL]+mle[WIDE]-1-VGA_MODE } ) )

end procedure


-----------------------------------------------------------------------------
function relative_anchor_position( sequence mle )

    -- return relative position of anchor in mle
    
            
    -- calculate anchor's relative position
                
    if mle[ANCHOR_COL] = 0 then
        return 0

    elsif mle[ANCHOR_COL] < mle[FIRST_CHAR] then
        -- anchor is off the left hand edge
        return 1

    elsif mle[ANCHOR_COL] > mle[FIRST_CHAR]+mle[WIDE] then
        -- anchor is off the right hand edge
        return mle[WIDE]

    else
        -- anchor is within the mle
        return mle[ANCHOR_COL] - mle[FIRST_CHAR] + 1

    end if

end function


-----------------------------------------------------------------------------
global function current_col( sequence mle )

    -- return absolute position of cursor
    
    return mle[FIRST_CHAR]+mle[CURSOR]-1

end function


-----------------------------------------------------------------------------
function after_full_tab( sequence mle )

    -- true if cursor is sitting on space delimited tab

    integer line, col

    -- alias mle

    -- tab stop?
    col = current_col( mle )
    if remainder( col-1, TAB_SIZE ) != 0 then
        return 0
    end if

    -- too small for tab stops?
    if col < TAB_SIZE then
        return 0
    end if

    -- past end of file?
    line = current_line( mle )
    if line > length( mle[ITEMS] ) then
        return 0
    end if

    -- past end of line? the -1 allows cursor 1 past end
    if col-1 > length( mle[ITEMS][line] ) then
        return 0
    end if

    -- preceeded by spaces?
    if compare( mle[ITEMS][line][col-TAB_SIZE..col-1],
        repeat(  ' ', TAB_SIZE ) ) = 0 then

        -- it's a tab stop
        return 1

    end if

    -- nope
    return 0

end function



-----------------------------------------------------------------------------
global function xle_selected_text( sequence mle )

    -- return the selected segment of text

    sequence text
    integer this

    -- nothing selected?
    if mle[ANCHOR_COL] = 0 then
            return ""
    end if

    -- current line
    this = current_line( mle )

    -- multi-line?
    if this != mle[ANCHOR_LINE] then
        if this > mle[ANCHOR_LINE] then
            -- cursor above line
            return mle[ITEMS][this..mle[ANCHOR_LINE]-1]
        else
            -- cursor below the line
            return mle[ITEMS][mle[ANCHOR_LINE]..this-1]
        end if
    end if
    
    -- text in question
    text = mle[ITEMS][this]

    -- selected on this line?
    if mle[ANCHOR_LINE] != this then
        -- not this line
        return ""
    end if

    -- no text selected?
    if mle[ANCHOR_COL] = 0
    or relative_anchor_position( mle ) = mle[CURSOR] then
                    
        -- nothing selected
        return ""

    end if
    
    -- figure out the segment from the boundaries
    if mle[ANCHOR_COL] < current_col( mle ) then
                                    
        return text[mle[ANCHOR_COL]..current_col( mle )-1]

    else 
                            
        return text[current_col( mle )..mle[ANCHOR_COL]-1]

    end if

end function


-----------------------------------------------------------------------------
function sep( integer i )

    -- true if i is a seperator

    if i = '_'                      then return 0
    elsif i >= 'A' and i <= 'Z'     then return 0
    elsif i >= 'a' and i <= 'z'     then return 0
    elsif i >= '0' and i <= '9'     then return 0
    else                            return 1
    end if

end function



-----------------------------------------------------------------------------
global function on_xle_word( sequence xle )

    -- get the word the the cursor is on

    integer line, col, at, wStart, wEnd
    sequence text

    -- alias xle

    -- current values
    line = current_line( xle )
    col = current_col( xle )
    text = xle[ITEMS][line]
    at = xle[CURSOR] + xle[FIRST_CHAR] - 1

    -- outside text?
    if at > length( text ) then
        return ""
    end if

    -- find start of word
        wStart = at
        while 1 do    
            if wStart = 1 then            
                exit
            elsif sep( text[wStart] ) then
                -- back up, but only if position moved
                if wStart < at then
                    wStart = wStart + 1
                end if
                exit
            else
                wStart = wStart - 1
            end if    
        end while

    -- find end of word
    wEnd = at
    while 1 do
        if wEnd = length(text) then
            exit
        elsif sep( text[wEnd] ) then
            -- back up, but only if position moved
            if wEnd > at then
                wEnd = wEnd - 1
            end if
            exit
        else
            wEnd = wEnd + 1
        end if
    end while

    return text[wStart..wEnd]

end function


-----------------------------------------------------------------------------
global procedure draw_mle_item( sequence mle, integer line )

    -- color coded display function
    
    integer     item, anchor, this
    sequence    outText, leftSeg, midSeg, rightSeg, color
    integer     sColor, sStart, sEnd, sWide
    sequence    sText, segs

    -- alias mle

    -- get the actual item
    item = mle[TOP_LINE]+line-1

    -- current line
    this = current_line( mle )

    -- move to position
    position( mle[ROW]+line-1, mle[COL]-VGA_MODE )

    -- selected text
    if item > length( mle[ITEMS] ) then
        outText = ""
    else
        outText = mle[ITEMS][item]
    end if

    -- pad the text to maximum size
    sWide = mle[FIRST_CHAR] + mle[WIDE] - 1
    if length( outText ) < sWide then
        outText = outText & repeat(  ' ', sWide - length( outText ) + 1 )
    end if



    -- tests to see if we can skip the colored syntax
    -- this can happen if colorization is off, or the line is
    -- entirely highlit

    -- current line is above anchor, and line's below it
    if  mle[ANCHOR_COL] != 0
    and  this > mle[ANCHOR_LINE]
    and  item >= mle[ANCHOR_LINE]
    and  item < this
    then
        -- highlight
        color = OF_SELECTED

    -- current line is below anchor, and line's below that?
    elsif  mle[ANCHOR_COL] != 0
    and    this < mle[ANCHOR_LINE]
    and    item < mle[ANCHOR_LINE]
    and    item >= this
    then
        -- highlight
         color = OF_SELECTED

    elsif not mle[COLOR_SYNTAX] then
        color = OF_TEXT

    else
        -- colorized, not selected
        color = ""
    end if


    -- display whole line in one color, or colorize?

    if length( color ) != 0 then

        -- text displayed is just one color
        put_colors( color, outText[mle[FIRST_CHAR]..sWide] )

    else

        -- multi-colored text

        -- get the colored segments
        segs = BuildColorLine( outText )

        -- display the segments
        sEnd = 0
        for i_1 = 1 to length(segs) do
        
            -- get color and text
            sText = segs[i_1][2]

            -- position
            sStart = sEnd + 1
            sEnd = sStart + length( sText ) - 1

            -- slice segment
            if sEnd < mle[FIRST_CHAR] then
                -- skip this segment, not visible
                
            elsif sStart > sWide then
                -- finished, past end of line
                exit                        
                
            else

                -- trims are reversed so slice positions can be used

                -- right trim
                if sEnd > sWide then
                    sText = sText[1..(sWide-sStart)+1]
                end if

                -- left trim
                if sStart < mle[FIRST_CHAR] then
                    sText = sText[mle[FIRST_CHAR]-sStart+1..length(sText)]
                end if
        
                -- display
                put_colors( { segs[i_1][1], OF_TEXT[2] }, sText )

            end if
            
        end for

    end if

    -- skip this if there is no segment to display
    if mle[ANCHOR_COL] = 0 then
        return
    end if

    -- calculate anchor's relative position
    anchor = relative_anchor_position( mle )

    -- anchor set; on anchored line; cursor is not on anchor
    if      mle[ANCHOR_LINE] = item
    and     mle[ANCHOR_LINE] = this
    and     anchor != mle[CURSOR]
    then

        -- figure out the segments
        if anchor < mle[CURSOR] then
                                        
            -- segment is <left--><anchor--cursor><--right>

            -- move to position
            position( mle[ROW]+line-1, mle[COL]+anchor-1-VGA_MODE )
            put_colors( OF_SELECTED, outText[anchor+mle[FIRST_CHAR]-1..
                mle[CURSOR]+mle[FIRST_CHAR]-2] )
                                    
        else 

            -- segment is <left--><cursor--anchor><--right>

            -- move to position
            position( mle[ROW]+line-1, mle[COL]+mle[CURSOR]-1-VGA_MODE )
            put_colors( OF_SELECTED, outText[mle[CURSOR]+mle[FIRST_CHAR]-1..
                anchor+mle[FIRST_CHAR]-2] )

        end if

    end if

end procedure


-----------------------------------------------------------------------------
global function draw_mle_items( sequence mle )

    -- draw all the items in the mle

    for i = 1 to mle[DEEP] do
        draw_mle_item( mle, i )
    end for

    -- draw the elevator
    mle = draw_elevator( mle )

    return mle

end function


-----------------------------------------------------------------------------
global function draw_sle( sequence sle )

    integer     anchorAt, maxSize, fillChar
    sequence    outText, leftSeg, midSeg, rightSeg, color

    -- alias sle

    -- color
    if sle[DIM] then
        color = {OF_DISABLED[1], OF_DIALOG[2]}
    else
        color = OF_DIALOG
    end if


    -- skip title if no refresh
    if  sle[REDRAW]
    and compare( sle[TITLE], "" ) != 0 then

        if VGA_MODE then
    
            -- display
            graphic_title( sle[TITLE], color,
                 sle[ROW], sle[COL] - length( sle[TITLE] ) - 2 , sle[HOT], 0 )
    
        else

            -- position title
            position( sle[ROW], sle[COL] - length( sle[TITLE] ) - 2 )
 
            if sle[DIM] then
                put_colors( color, sle[TITLE] )
            else
                put_colors( color, sle[TITLE] )
                -- hot key?
                if sle[HOT] != 0 then
                    position( sle[ROW], sle[COL] - length( sle[TITLE] ) - 3 + sle[HOT] )
                    put_colors( { OF_HOTKEY[1], OF_DIALOG[2] }, sle[TITLE][sle[HOT]] )
                end if
            end if
        end if
        end if


    -- frame
    if VGA_MODE then

        -- only on refresh
        if sle[REDRAW] then
            if VGA_MODE then
                graphic_frame( sle[ROW]-1, sle[COL]-2, sle[WIDE]+1, 1 )
            end if
        end if

    else

        -- move to start of sle
        position( sle[ROW], sle[COL]-1 )

        -- left side
        put_colors( color, "[" )

        -- right side
        position( sle[ROW], sle[COL]+sle[WIDE] )
        put_colors( color, "]" )

    end if

    -- move to start of sle
    position( sle[ROW], sle[COL]-VGA_MODE )

    -- text to display
    outText = sle[DATA]

    -- pad to maximum size
    maxSize = sle[FIRST_CHAR] + sle[WIDE]
    
    if length( outText ) < maxSize then

        -- fill char replaces nulls
        if VGA_MODE then
            fillChar =  ' '
        else
            fillChar = ''
        end if

        -- build text
        outText = outText & repeat( fillChar, maxSize - length( sle[DATA] ) )

    end if

    -- left trim text
    outText = outText[sle[FIRST_CHAR]..length( outText )]

    -- right trim text
    if length( outText ) > sle[WIDE] then
        outText = outText[1..sle[WIDE]]
    end if

    -- calculate anchor's relative position
    anchorAt = relative_anchor_position( sle )
    
    -- text to highlight?
    if sle[DIM] then

        -- display text
        put_colors( {OF_DISABLED[1], OF_DIALOG[2]}, outText )

        -- remainder
        put_colors( {OF_DISABLED[1], OF_TEXT[2]},
        repeat(   ' ', sle[WIDE] - length( outText ) ) )

    elsif anchorAt = 0 or anchorAt = sle[CURSOR] then
                    
        -- display text
        put_colors( OF_DIALOG, outText )

        -- remainder
        put_colors( OF_DIALOG,
        repeat(   ' ', sle[WIDE] - length( outText ) ) )

    else

        -- figure out the segments
        if anchorAt < sle[CURSOR] then
                                            
            -- segment is <left--><anchor--cursor><--right>
            leftSeg = outText[1..anchorAt-1]
            midSeg = outText[anchorAt..sle[CURSOR]-1]
            rightSeg = outText[sle[CURSOR]..sle[WIDE]]
                                    
        else 
                                                                    
            -- segment is <left--><cursor--anchor><--right>
            leftSeg = outText[1..sle[CURSOR]-1]
            midSeg = outText[sle[CURSOR]..anchorAt-1]
            rightSeg = outText[anchorAt..sle[WIDE]]

        end if

        -- display left segment
        put_colors( OF_DIALOG, leftSeg )

        -- display selected segment
        put_colors( OF_SELECTED, midSeg )

        -- display right segment
        put_colors( OF_DIALOG, rightSeg )

    end if

    -- position cursor
    position( sle[ROW], sle[COL]+sle[CURSOR]-1-VGA_MODE )

    -- return sle
    return sle

end function



-----------------------------------------------------------------------------
function move_cursor_into_mle( sequence mle, integer here )

    -- move cursor to HERE (absolute position)
    -- make sure it is visible there
    
    if here < mle[FIRST_CHAR] then
        -- position is to the left mle
        while here < mle[FIRST_CHAR] do
                mle[FIRST_CHAR] = mle[FIRST_CHAR] - 1
        end while

        -- don't fall off the edge
        if mle[FIRST_CHAR] < 1 then
            mle[FIRST_CHAR] = 1
            here = 1
        end if
    
    elsif here > mle[FIRST_CHAR]+mle[WIDE]-1 then
        -- position is to the right of the mle
        while here > mle[FIRST_CHAR]+mle[WIDE]-1 do
            mle[FIRST_CHAR] = mle[FIRST_CHAR] + 1
        end while

    end if    
            
    -- calculate cursor position
    mle[CURSOR] = here-mle[FIRST_CHAR]+1

    -- return mle; FIRST_CHAR and CURSOR are altered
    return mle

end function

-----------------------------------------------------------------------------
function pad_to_cursor( sequence mle )

    -- add padding to current line if the cursor lies beyond the end of
    -- the text

    integer this, cAt
    sequence text

    cAt = current_col( mle )
    this = current_line( mle )
    text = mle[ITEMS][this]

    -- does it need padding?
    if cAt > length( text ) then

        -- add padding
        mle[ITEMS][this] = text &
            repeat(  ' ', cAt - length(text) - 1 )

    end if

    return mle

end function

-----------------------------------------------------------------------------
global function xle_cut_text( sequence mle )

    -- remove selected text      

    integer this, rightmost

    -- object is mle
    -- alias mle

    -- exit if nothing selected
    if mle[ANCHOR_COL] = 0 then
            return mle
    end if

    -- set as dirty
    mle[DIRTY] = 1

    -- current line
    this = current_line( mle )

    -- multiple lines?
    if this != mle[ANCHOR_LINE] then

        -- which way is the selection?
        if this < mle[ANCHOR_LINE] then

            -- cursor is above anchor line
            mle[ITEMS] = mle[ITEMS][1..this-1]
                & mle[ITEMS][mle[ANCHOR_LINE]..length(mle[ITEMS])]

        else

            -- cursor is below the anchor line
            mle[ITEMS] = mle[ITEMS][1..mle[ANCHOR_LINE]-1]
               & mle[ITEMS][this..length(mle[ITEMS])]

            -- adjust cursor position
            if mle[ANCHOR_LINE] < mle[TOP_LINE] then

                -- move to top of page
                mle[TOP_LINE] = mle[ANCHOR_LINE]
                mle[ITEM] = 1

            else

                -- move to anchor
                mle[ITEM] = mle[ANCHOR_LINE] - mle[TOP_LINE] + 1

            end if
            
        end if
            
        -- clear selection
        mle[ANCHOR_COL] = 0
            
        -- return mle
        return mle
            
    end if
    

    -- single line

    -- is the cursor/anchor beyond the current text?

    -- find rightmost position
    rightmost = current_col( mle ) 
    if mle[ANCHOR_COL] > rightmost then
        rightmost = mle[ANCHOR_COL]
    end if
            
    -- add padding if needed
    if rightmost > length( mle[ITEMS][this] ) then
        -- add padding
        mle[ITEMS][this] = mle[ITEMS][this] & 
            repeat(   ' ', rightmost - length( mle[ITEMS][this] ) - 1 )
    end if

    -- cut segment
    if mle[ANCHOR_COL] < current_col( mle ) then
                            
        -- replace between anchor and cursor
        mle[ITEMS][this] = 
            mle[ITEMS][this][1..mle[ANCHOR_COL]-1] 
            & mle[ITEMS][this][current_col(mle)..length(mle[ITEMS][this])]

        -- move to cursor at anchor
        mle = move_cursor_into_mle( mle, mle[ANCHOR_COL] )

    else 
                    
        -- replace between cursor and anchor
        mle[ITEMS][this] = 
            mle[ITEMS][this][1..current_col( mle )-1]
            & mle[ITEMS][this][mle[ANCHOR_COL]..length(mle[ITEMS][this])]
                                                             
        -- move to cursor at new cursor
        mle = move_cursor_into_mle( mle, current_col( mle ) )

    end if

    -- clear the anchor
    mle[ANCHOR_COL] = 0
    
    -- return mle
    return mle


end function

-----------------------------------------------------------------------------
global procedure xle_copy_text( sequence mle )

    -- copies selected text to clipboard

    sequence text
    integer this, rightmost

    -- object is mle
    -- alias mle

    -- exit if nothing selected
    if mle[ANCHOR_COL] = 0 then
            return
    end if

    -- current line
    this = current_line( mle )

    -- multiple lines?
    if this != mle[ANCHOR_LINE] then

        -- which way is the selection?
        if this < mle[ANCHOR_LINE] then
            -- cursor is above anchor line
            clipboard = mle[ITEMS][this..mle[ANCHOR_LINE]-1]

        else
            -- cursor is below the anchor line
            clipboard = mle[ITEMS][mle[ANCHOR_LINE]..this-1]

        end if

        -- leave procedure
        return

    end if

    -- single line

    -- is the cursor/anchor beyond the current text?

    -- find rightmost position
    rightmost = current_col( mle ) 
    if mle[ANCHOR_COL] > rightmost then
            rightmost = mle[ANCHOR_COL]
    end if
            
    -- copy to local variable
    text = mle[ITEMS][this]

    -- add padding if needed
    if rightmost > length( text ) then
        -- add padding
        text = text & 
            repeat(  ' ', rightmost - length( text ) - 1 )
    end if

    -- cut segment
    if mle[ANCHOR_COL] < current_col( mle ) then
        -- replace between anchor and cursor
        clipboard = text[mle[ANCHOR_COL]..current_col(mle)-1]

    else 
        clipboard = text[current_col(mle)..mle[ANCHOR_COL]-1]

    end if

end procedure

-----------------------------------------------------------------------------
global function xle_paste_text( sequence mle, sequence text )

    -- paste text at cursor position
    
    integer this, cursor

    -- object is mle
    -- alias mle

    -- current line
    this = current_line( mle )

    -- absolute cursor position
    cursor = current_col( mle ) 

    -- nothing to paste?
    if length( text ) = 0 then
            return mle
    end if

    -- mark as dirty
    mle[DIRTY] = 1

    -- single line?
    if atom( text[1] ) then

        -- pad, if needed
        mle = pad_to_cursor( mle )

        -- insert at position
        mle[ITEMS][this] = 
            mle[ITEMS][this][1..cursor-1] 
            & text 
            & mle[ITEMS][this][cursor..length(mle[ITEMS][this])]

        -- adjust cursor
        mle = move_cursor_into_mle( mle, cursor + length( text ) )

        -- return mle
        return mle

    end if

    -- skip if sle
    if mle[TYPE] = SLE then
        -- can't paste multiple lines into an sle
        return mle
    end if

    -- multiple lines
    mle[ITEMS] =
        mle[ITEMS][1..this-1]
        & text
        &mle[ITEMS][this..length(mle[ITEMS])]

    return mle

end function


-----------------------------------------------------------------------------
function move_up( sequence mle )

    -- move up by one line
                            
    -- move up
    if mle[ITEM] > 1 then
        -- move to prior line
        mle[ITEM] = mle[ITEM] - 1
    
        -- draw the line we moved from
        draw_mle_item( mle, mle[ITEM] + 1 )

    else
        -- can we scroll?
        if mle[TOP_LINE] > 1 then

            if VGA_MODE then
                -- force redraw
                mle[TOP_LINE] = mle[TOP_LINE] - 1
    
            else
                -- physical scroll
                mle_scroll( mle, 1 )
    
                -- adjust settings
                mle[TOP_LINE] = mle[TOP_LINE] - 1
    
                -- prevent redraw of screen
                oldFirstLine = mle[TOP_LINE]
    
            end if
        end if
    end if

    -- return mle
    return mle

end function


-----------------------------------------------------------------------------
function move_down( sequence mle )

    -- move down by one line

    if mle[ITEM] < mle[DEEP] then
        -- is there a line to scroll to?
        if current_line( mle ) < length( mle[ITEMS] ) then
            -- move to next line
            mle[ITEM] = mle[ITEM] + 1

            -- draw the line we moved from
            draw_mle_item( mle, mle[ITEM] - 1 )

        end if    
        
    else
        -- can we scroll?
        if mle[TOP_LINE]+mle[DEEP]-1 < length( mle[ITEMS] ) then

            if VGA_MODE then
                -- force redraw
                mle[TOP_LINE] = mle[TOP_LINE] + 1

            else
                -- physical scroll
                mle_scroll( mle, -1 )

                -- adjust settings
                mle[TOP_LINE] = mle[TOP_LINE] + 1

                -- draw the line we moved from
                draw_mle_item( mle, mle[ITEM] - 1 )

                -- prevent redraw of screen
                oldFirstLine = mle[TOP_LINE]

            end if

        end if
        
    end if

    -- return mle
    return mle


end function


-----------------------------------------------------------------------------
function set_anchor_state( sequence mle, integer keyPress )

    -- check for anchor being set.
    -- clear, if state changes

    -- clear anchor if shift is *released*
    if shift_key_state() = 0 then
        if mle[ANCHOR_COL] != 0 then
            -- clear it
            mle[ANCHOR_COL] = 0

            -- redraw if multiple lines selected
            if current_line( mle ) != mle[ANCHOR_LINE] then
                -- force redraw
                redrawFlag = 1
            end if

        end if

    else

        -- shift is *pressed*
        -- set anchor if not already set
        if mle[ANCHOR_COL] = 0 then
            -- set cursor
            mle[ANCHOR_COL] = current_col( mle )
            mle[ANCHOR_LINE] = current_line( mle )

        end if

        -- redraw if not handled as special case
        if keyPress = HOME          
        or keyPress = END
        or keyPress = CTRL_HOME     
        or keyPress = CTRL_END then

            -- force screen redraw
            redrawFlag = 1

        end if

    end if

    -- updated mle
    return mle

end function


-----------------------------------------------------------------------------
function find_next_word( sequence mle, integer dir )

    -- move to prior/next delimited alpha

    integer line, col

    -- alias mle

    -- current values    
    col = current_col( mle )
    line = current_line( mle )

    -- clip
    if  col > length( mle[ITEMS][line] ) then
        col = length( mle[ITEMS][line] )
    end if

    -- add direction to column
    col = col + dir


    -- loop while within array
    while line <= length( mle[ITEMS] )
    and   line > 0 do

        -- bounds check for line
        if  col <= length( mle[ITEMS][line] )
        and col > 0 then

            -- can we check prior char?
            if  col > 1 then
                -- can check prior char
                if  sep( mle[ITEMS][line][col-1] )         -- prior is sep
                and not sep( mle[ITEMS][line][col] ) then  -- this is not
                    exit
                end if

            -- first char on line
            elsif col <= length( mle[ITEMS][line] ) then    -- in text
                if not sep( mle[ITEMS][line][col] ) then    -- not a sep
                    exit
                end if
            end if
        end if

        -- move to next/prior char
        col = col + dir

        -- past bounds?
        if col > length( mle[ITEMS][line] )
        or col < 1 then

            if dir = 1 then
                -- next line                    
                line = line + dir

                -- first column
                col = 1

            else
                -- prior line
                line = line + dir

                -- end of line (if in range)
                if line > 0 then
                    col = length( mle[ITEMS][line] )
                end if

            end if

        end if

    end while

    -- still within bounds?
    if  line <= length( mle[ITEMS] )
    and line > 0 then

        -- scroll?
        if line > mle[TOP_LINE] + mle[DEEP] - 1
        or line < mle[TOP_LINE] then
            mle[TOP_LINE] = line
        end if

        -- move to line
        mle[ITEM] = line - mle[TOP_LINE] + 1

        -- position in column
        mle = move_cursor_into_mle( mle, col )

    end if

    return mle

end function


-----------------------------------------------------------------------------
global function is_cut_key( integer keyPress )

    -- returns true if key pressed is the "cut" key

    return  keyPress = CTRL_X
            or ( keyPress = DELETE and shift_key_state() )
            
end function


-----------------------------------------------------------------------------
global function is_copy_key( integer keyPress )

    -- returns true if key pressed is the "copy" key

    return  keyPress = CTRL_C
            or ( keyPress = CTRL_INSERT )
            
end function
                            

-----------------------------------------------------------------------------
global function is_paste_key( integer keyPress )

    -- returns true if key pressed is the "paste" key

    return  keyPress = CTRL_V
            or ( keyPress = INSERT and shift_key_state() )
            
end function



-- MAIN EDIT ROUTINE

-----------------------------------------------------------------------------
global function edit_xle( sequence mle, integer keyPress )

    -- edit text in an mle/sle

    integer this, that, dir, count, at
    sequence text, indent

    -- key handling loop

    -- alias mle

    -- discard the key if it's a mouse event and not in the MLE
    if keyPress >= 800 and keyPress <= 900 then
        if MOUSE_COL < mle[COL]             -- left column
        or MOUSE_COL > mle[COL] + mle[WIDE]     -- scroll bar
        or MOUSE_ROW < mle[ROW]-1           -- top border
        or MOUSE_ROW > mle[ROW] + mle[DEEP]     -- bottom border
        then

            -- ignore mouse
            return mle

        end if
    end if

    -- make the sle look like a 1 line mle
    if mle[TYPE] = SLE then
        mle[ITEMS] = { mle[DATA] }
    end if

    -- for various reasons, it's helpful to ensure there is a line past the
    -- "last" line in the file

    -- append an empty line to the end of an MLE if there isn't one
    if compare( mle[ITEMS][length(mle[ITEMS])], "" ) != 0
    and mle[TYPE] = MLE then
        mle[ITEMS] = append( mle[ITEMS], "" )
    end if

    -- get the old first char (to tell if the lines have been shifted)
    oldFirstChar = mle[FIRST_CHAR]

    -- get the old first line (to tell if the lines have been scrolled)
    oldFirstLine = mle[TOP_LINE]

    -- get the count (to see if it grew or shrank)
    count = length( mle[ITEMS] )

    -- don't redraw the whole screen
    redrawFlag = 0

    -- auto complete ---------------------------------------------------
    if  mle[AUTOCOMPLETE] then

        -- get current line
        this = current_line( mle )
  
        -- completion flag set?
        if completingLine then 
            -- are we on a different line?
            if completingLine != this then
                -- turn off the flag
                completingLine = 0
            end if
        end if


        -- pressed enter while auto completing?
        if keyPress = ENTER   
        and completingLine then

            -- proper behavior is to move to end of next line
            -- do not break the current line
            
            -- is there a line to go to?
            if this < length( mle[ITEMS] ) then
            
                -- move to the next line
                mle = move_down( mle )
        
                -- move to end of that line
                count = length( mle[ITEMS][current_line(mle)] ) + 1
                mle = move_cursor_into_mle( mle, count )
               
                -- kill keystroke
                keyPress = 0
                
            end if
            
            -- clear flag
            completingLine = 0
            
        end if

        -- autocomplete code on space and no text selected
        if  keyPress =  ' '
        and completingLine = 0
        and mle[ANCHOR_COL] = 0 then
        
            -- check to see if the prior word if an auto complete word
        
            -- trim off leading spaces
            -- place them in 'indent'
            text = mle[ITEMS][this]
            for i = 1 to length( text ) do
                if text[i] !=  ' ' then
                    text = text[i..length(text)]
                    indent = repeat(  ' ', i-1 )            
                    exit
                end if
            end for
        
            -- trim off trailing spaces
            for i = length( text ) to 1 by -1 do
                if text[i] !=  ' ' then
                    text = text[1..i]
                    exit
                end if
            end for
        
            -- look in the replacement list
            at = find( text, expand_word )
            if at != 0 then             
                
                -- replace with the expansion
                mle[ITEMS][this] = indent & text & expand_text[at]
            
                -- insert a indented line
                mle[ITEMS] = mle[ITEMS][1..this]
                    & { indent & repeat(   ' ', TAB_SIZE ) }
                    & mle[ITEMS][this+1..length(mle[ITEMS])] 

                -- insert an end... line?
                if length( expand_end[at] ) then
                    mle[ITEMS] = mle[ITEMS][1..this+1]
                        & { indent & expand_end[at] }
                        & mle[ITEMS][this+2..length(mle[ITEMS])] 
                end if
                
        
                -- paste in the space, and move cursor ahead
                mle = xle_paste_text( mle, " " )
            
                -- force redraw
                redrawFlag = 1
        
                -- set flag to handle enter key
                completingLine = this
        
                -- kill the key
                keyPress = 0
            
            end if
                
        end if

    end if
    

    -- translate key --------------------------------------------------
        
    if natural_key( keyPress )
    and   edit_mode() = INSERT_MODE
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1    

        -- text selected?
        if mle[ANCHOR_COL] != 0 then

            -- force redraw, if several lines effected
            if mle[ANCHOR_LINE] != current_line( mle ) then
                -- force redraw
                redrawFlag = 1
            end if

            -- cut any pending text
            mle = xle_cut_text( mle )

        end if

        -- key is a normal key; insert mode or insert
        mle = xle_paste_text( mle, {keyPress} )


    elsif natural_key( keyPress )
    and   edit_mode() = OVERWRITE_MODE
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1

        -- text selected?
        if mle[ANCHOR_COL] != 0 then

            -- force redraw, if several lines effected
            if mle[ANCHOR_LINE] != current_line( mle ) then
                    -- force redraw
                    redrawFlag = 1
            end if

            -- cut any pending text
            mle = xle_cut_text( mle )

        end if

        -- select next character & overwrite
        mle[ANCHOR_COL] = current_col( mle ) + 1
        mle[ANCHOR_LINE] = current_line( mle )

        -- overwrite
        mle = xle_cut_text( mle )
        mle = xle_paste_text( mle, {keyPress} )


    elsif keyPress = BACKSPACE
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1

        -- turn off anchor
        mle[ANCHOR_COL] = 0

        -- current line
        this = current_line( mle )

        -- placed within left side of text?
        if current_col( mle ) > 1 then

            -- placed within right side of text?
            if current_col( mle ) > length( mle[ITEMS][this] ) + 1 then

                -- beyond end of text, just move cursor
                mle[CURSOR] = mle [ CURSOR]- 1

            else

                -- select tab before cursor & replace with null
                mle[ANCHOR_COL] = current_col( mle )
                mle[ANCHOR_LINE] = current_line( mle )

                -- on tab?
                if after_full_tab( mle ) then
                    -- select tab
                    mle[CURSOR] = mle [ CURSOR]- TAB_SIZE
                else
                    -- selct character
                    mle[CURSOR] = mle [ CURSOR]- 1
                end if

                -- cut the text
                mle = xle_cut_text( mle )

            end if


        else

            -- append this line to prior line

            -- don't go past first line
            if this > 1 then

                -- place cursor at end of prior line
                mle = move_cursor_into_mle( mle,
                        length( mle[ITEMS][this-1] ) + 1 )

                -- append line to prior line
                mle[ITEMS][this-1] = mle [ ITEMS][this-1]& mle[ITEMS][this]

                -- remove the prior line
                if this = length( mle[ITEMS] ) then
                    -- drop off end
                    mle[ITEMS] = mle[ITEMS][1..this-1]
                else
                    -- remove from middle
                    mle[ITEMS] =
                        mle[ITEMS][1..this-1] &
                        mle[ITEMS][this+1..length(mle[ITEMS])]
                end if

                -- move current line up
                if mle[ITEM] = 1 then
                    -- change top line
                    mle[TOP_LINE] = mle [ TOP_LINE]- 1
                else
                    -- move relative line
                    mle[ITEM] = mle [ ITEM]- 1
                end if


                -- force redraw
                redrawFlag = 1

            end if

        end if


    elsif ( keyPress = ENTER or keyPress = CTRL_ENTER )
    and   mle[TYPE] = MLE
    and   not mle[READ_ONLY] then

        -- split the line

        -- mark as dirty
        mle[DIRTY] = 1

        -- clear the anchor
        mle[ANCHOR_COL] = 0

        -- current line
        this = current_line( mle )

        -- at end of file?
        if this = length( mle[ITEMS] ) then

            -- add empty line at end
            mle[ITEMS] = append( mle[ITEMS], "" )

        else

            -- split the line                

            -- cursor position
            that = current_col( mle )

            -- add padding, if needed
            mle = pad_to_cursor( mle )

            -- split at the cursor
            mle[ITEMS] = mle[ITEMS][1..this-1] &
                    { mle[ITEMS][this][1..that-1] } &
                    { mle[ITEMS][this][that..length(mle[ITEMS][this])] } &
                    mle[ITEMS][this+1..length(mle[ITEMS])]

        end if

        -- add padding
        that = leading_spaces( mle[ITEMS][this] )
        mle[ITEMS][this+1] =
                repeat(  ' ',  that )
                & mle[ITEMS][this+1]

        -- move the cursor
        if mle[ITEM] >= mle[DEEP] then
            -- move first line down
            mle[TOP_LINE] = mle[TOP_LINE] + 1

            -- make sure we don't go past the bottom line
            mle[ITEM] = mle[DEEP]

        else

            -- move to the next line
            mle[ITEM] = mle[ITEM] + 1

        end if


        -- position cursor past leading spaces
        mle = move_cursor_into_mle( mle, that+1 )

        -- force redraw
        redrawFlag = 1


    elsif keyPress = LEFT then

        -- handle anchors             
        mle = set_anchor_state( mle, keyPress )

        -- need to scroll left?
        mle = move_cursor_into_mle( mle, current_col( mle )-1 )
                                

    elsif keyPress = RIGHT then

        -- handle anchors             
        mle = set_anchor_state( mle, keyPress )

        -- move to the right
        mle = move_cursor_into_mle( mle, current_col( mle )+1 )


    elsif keyPress = TAB
    and   mle[TYPE] = MLE
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1

        -- text selected?
        if mle[ANCHOR_COL] = 0 then            

            -- no text selected. just tab

            -- calculate next tab stop
            this = remainder( current_col( mle )-1, TAB_SIZE )
            if this = 0 then
                -- full tab
                this = TAB_SIZE
            else
                -- partial tab
                this = TAB_SIZE - this
            end if

            mle = xle_paste_text( mle, repeat(  ' ', this ) )

        else

            -- add tab to all the selected items

            -- get range, from this to that
            this = current_line( mle )
            if mle[ANCHOR_LINE] > this then
                that = mle[ANCHOR_LINE]
            else
                that = this
                this = mle[ANCHOR_LINE]
            end if

            -- append indent
            indent = repeat(  ' ', TAB_SIZE )
            for i = this to that-1 do
                mle[ITEMS][i] = indent & mle[ITEMS][i]
            end for

            -- force redraw
            redrawFlag = 1

        end if

    elsif keyPress = SHIFT_TAB
    and   mle[TYPE] = MLE
    and   mle[ANCHOR_COL] != 0
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1

        -- remove 1 tab to all the selected items

        -- get range, from this to that
        this = current_line( mle )
        if mle[ANCHOR_LINE] > this then
            that = mle[ANCHOR_LINE]
        else
            that = this
            this = mle[ANCHOR_LINE]
        end if

        -- force redraw
        redrawFlag = 1

        -- build an indent for comparison
        indent = repeat(  ' ', TAB_SIZE )

        -- examine each line
        for i = this to that-1 do
            -- room for a tab?
            if length( mle[ITEMS][i] ) >= TAB_SIZE then
                -- spaces?
                if compare( mle[ITEMS][i][1..TAB_SIZE], indent ) = 0 then
                        -- remove spaces
                        mle[ITEMS][i] = mle[ITEMS][i][TAB_SIZE+1..length(mle[ITEMS][i])]
                end if
            end if
        end for


    elsif keyPress = CTRL_LEFT then

        -- move to prior word
        mle = find_next_word( mle , -1 )


    elsif keyPress = CTRL_RIGHT then

        -- move to next word
        mle = find_next_word( mle , 1 )


    elsif keyPress = UP
    and   mle[TYPE] = MLE then

        mle = set_anchor_state( mle, keyPress )     -- handle anchors
        mle = move_up( mle )                        -- move up 1 line

    elsif keyPress = DOWN
    and   mle[TYPE] = MLE then

        mle = set_anchor_state( mle, keyPress )     -- handle anchors
        mle = move_down( mle )                      -- move down 1 line

    elsif keyPress = MOUSE_CLICK               -- mouse event
    or keyPress = MOUSE_HELD            
    or keyPress = MOUSE_DRAG
    and MOUSE_ROW = mle[ROW] - 1 then           -- above mle

        -- handle mouse events above the mle

        if keyPress = MOUSE_CLICK then              -- not part of drag
            mle = set_anchor_state( mle, keyPress ) -- handle anchors
        end if

        mle = move_up( mle )                        -- scroll up one line

    elsif keyPress = MOUSE_CLICK            -- mouse event
    or    keyPress = MOUSE_HELD
    or    keyPress = MOUSE_DRAG
    and   MOUSE_ROW = mle[ROW] + mle[DEEP] then     -- under mle

        -- handle mouse events below the mle

        if keyPress = MOUSE_CLICK then                  -- not part of drag
            mle = set_anchor_state( mle, keyPress )     -- handle anchors
        end if

        mle = move_down( mle )              -- scroll down one line

    elsif keyPress = MOUSE_CLICK            -- mouse event
    or    keyPress = MOUSE_HELD
    or    keyPress = MOUSE_DRAG
    and   (    MOUSE_COL = mle[COL] + mle[WIDE]                 -- on scroll bar
            or MOUSE_COL = mle[COL] + mle[WIDE] - VGA_MODE )    -- graphic
    and   length( mle[ITEMS] ) > mle[DEEP] then     -- scroll bar exists

        -- handle anchors
        mle = set_anchor_state( mle, keyPress )

        -- handle scroll bar
        mle = scrollbar_clicked( mle )

    elsif keyPress = MOUSE_CLICK then

        -- set cursor
        mle[ITEM] =  MOUSE_ROW - mle[ROW] + 1
        mle[CURSOR] = MOUSE_COL - mle[COL] + 1

        -- handle anchors
        mle = set_anchor_state( mle, keyPress )

        -- set anchor
        mle[ANCHOR_COL] = current_col( mle )
        mle[ANCHOR_LINE] = current_line( mle )

        -- past end of file?
        this = current_line( mle ) - length( mle[ITEMS] )
        if this > 0 then 
            -- add extra lines
            mle[ITEMS] = mle[ITEMS] & repeat( "", this )
        end if

    elsif keyPress = MOUSE_DRAG then

        -- set cursor
        mle[ITEM] =  MOUSE_ROW - mle[ROW] + 1
        mle[CURSOR] = MOUSE_COL - mle[COL] + 1
    
        -- past end of file?
        this = current_line( mle ) - length( mle[ITEMS] )
        if this > 0 then 
            -- add extra lines
            mle[ITEMS] = mle[ITEMS] & repeat( "", this )
        end if

        -- force redraw
        oldFirstLine =0


    elsif keyPress = MOUSE_UP then

        -- clear anchor if no text selected
        if      mle[ANCHOR_COL] = current_col( mle )
        and     mle[ANCHOR_LINE] = current_line( mle ) then

            mle[ANCHOR_COL] = 0     -- clear anchor

        end if

    elsif keyPress = MOUSE_HELD then
        
        -- this can cause a lot of busy work to happen
        -- instead, just ignore it
        
        return mle


    elsif keyPress = CTRL_UP then

        if mle[TOP_LINE] > 1 then  -- can scroll up?
            mle[TOP_LINE] = mle [ TOP_LINE]- 1                              -- if not at top
            if mle[ITEM] < mle[DEEP] then                       -- can move down?
                mle[ITEM] = mle [ ITEM]+ 1                              -- move down
            end if            
        end if
                                
    elsif keyPress = CTRL_DOWN then

        if mle[TOP_LINE]+mle[DEEP] < length( mle[ITEMS] ) then  -- if not at end
            mle[TOP_LINE] = mle [ TOP_LINE]+ 1                          -- scroll up 1
            if mle[ITEM] > 1 then                       -- can move up?
                mle[ITEM] = mle [ ITEM]- 1                          -- move up
            end if            
        end if


    elsif keyPress = HOME then

        mle = set_anchor_state( mle, keyPress )     -- handle anchors
        mle = move_cursor_into_mle( mle, 1 )        -- move to beginning

                    
    elsif keyPress = END then

        -- handle anchors        
        mle = set_anchor_state( mle, keyPress )

        -- move to end                
        mle = move_cursor_into_mle( mle, 
                length( mle[ITEMS][current_line( mle )] ) + 1 )
                        
    elsif keyPress = CTRL_HOME
    and   mle[TYPE] = MLE then

        -- handle anchors        
        mle = set_anchor_state( mle, keyPress )

        -- jump to top
        mle[CURSOR] = 1
        mle[ITEM] = 1
        mle[TOP_LINE] = 1

        -- force redraw
        redrawFlag = 1


    elsif keyPress = CTRL_END
    and   mle[TYPE] = MLE then

        -- handle anchors        
        mle = set_anchor_state( mle, keyPress )

        -- need to scroll?
        if length( mle[ITEMS] ) > mle[DEEP] then

            -- jump to the bottom
            mle[TOP_LINE] = ( length( mle[ITEMS] ) - mle[DEEP] ) + 1
            mle[ITEM] = mle[DEEP]
                
        else

            -- jump to last item
            mle[ITEM] = length( mle[ITEMS] )

        end if

        -- force redraw
        redrawFlag = 1


    elsif keyPress = PAGE_UP
    and   mle[TYPE] = MLE then

        mle = set_anchor_state( mle, keyPress )     -- handle anchors        
        mle = page_up( mle )                        -- move up 1 page


    elsif keyPress = PAGE_DOWN
    and   mle[TYPE] = MLE then

        mle = set_anchor_state( mle, keyPress )     -- handle anchors        
        mle = page_down( mle )                      -- move down 1 page

    elsif keyPress = CTRL_PAGE_UP then

        -- move one page to the right
        mle[FIRST_CHAR] = mle [ FIRST_CHAR]+ mle[WIDE]


    elsif keyPress = CTRL_PAGE_DOWN then

        -- move one page to the left
        mle[FIRST_CHAR] = mle [ FIRST_CHAR]- mle[WIDE]
        if mle[FIRST_CHAR] < 1 then
            mle[FIRST_CHAR] = 1
        end if


    elsif is_copy_key( keyPress ) then

        -- copy selected text
        xle_copy_text( mle )

        -- remove selection
        mle[ANCHOR_COL] = 0        

        -- force redraw
        redrawFlag = 1

    elsif is_cut_key( keyPress )
    and   mle[ANCHOR_COL] != 0
    and   not mle[READ_ONLY]    then
        
        -- mark as dirty
        mle[DIRTY] = 1

        -- copy selected text
        xle_copy_text( mle )

        -- delete selected text
        mle = xle_cut_text( mle )

        -- force redraw
        redrawFlag = 1

    elsif keyPress = DELETE
    and   mle[ANCHOR_COL] != 0
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1

        -- delete selected text
        mle = xle_cut_text( mle )
                        
        -- force redraw
        redrawFlag = 1

                    
    elsif keyPress = DELETE
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1

        -- save calculation
        this = current_line( mle )

        -- outside range (includes blank lines!)
        if current_col( mle ) > length( mle[ITEMS][this] ) then

            -- append the prior line to the current cursor position

            -- line underneath?
            if this < length( mle[ITEMS] ) then

                -- add padding, if not in position 1
                if current_col( mle ) != 1 then
                        mle = pad_to_cursor( mle )
                end if

                -- append following line
                mle[ITEMS][this] = mle [ ITEMS][this]& mle[ITEMS][this+1]

                -- remove the line just appended
                if this+1 = length( mle[ITEMS] ) then
                    -- drop off end
                    mle[ITEMS] = mle[ITEMS][1..this]
                else
                    -- remove from middle
                    mle[ITEMS] = 
                        mle[ITEMS][1..this] &
                        mle[ITEMS][this+2..length(mle[ITEMS])]
                end if

                -- force redraw
                redrawFlag = 1

            end if

        else

            -- select next character & delete
            mle[ANCHOR_COL] = current_col( mle ) + 1
            mle[ANCHOR_LINE] = current_line( mle )
            mle = xle_cut_text( mle )
                         
        end if

    elsif keyPress = CTRL_Y
    and   mle[TYPE] = MLE
    and   not mle[READ_ONLY] then

        -- yank line

        -- mark as dirty
        mle[DIRTY] = 1

        -- remove selection
        mle[ANCHOR_COL] = 0

        -- current line
        this = current_line( mle )

        -- copy line to buffer
        clipboard = mle[ITEMS][this]

        -- only line?
        if length( mle[ITEMS] ) = 1 then
            -- clear line
            mle[ITEMS][1] = ""

        -- last line?
        elsif this = length( mle[ITEMS] ) then
            -- do nothing, it's the null line

        else
            -- delete line
            mle[ITEMS] = mle[ITEMS][1..this-1] 
                & mle[ITEMS][this+1..length(mle[ITEMS])]
        end if

        -- force redraw
        redrawFlag = 1


    elsif keyPress = ESC then
    
        -- clear anchors
        mle[ANCHOR_COL] = 0
                
        -- redraw if multiple lines selected
        if current_line( mle ) != mle[ANCHOR_LINE] then
            redrawFlag = 1
        end if

    elsif is_paste_key( keyPress )
    and   not mle[READ_ONLY] then

        -- mark as dirty
        mle[DIRTY] = 1

        -- overwrite selected text
        mle = xle_cut_text( mle )
        mle = xle_paste_text( mle, clipboard )

        -- force redraw
        redrawFlag = 1

    elsif keyPress = REFRESH then

        -- set redraw to true
        mle[REDRAW] = 1

    end if
           
    -- hide the cursor
    hide_cursor()

    -- update display

    if  mle[TYPE] = MLE then
        
        -- draw the text
        if oldFirstChar != mle[FIRST_CHAR]              -- left/right scroll
        or oldFirstLine != mle[TOP_LINE]                -- up/down scroll
        or mle[REDRAW]                                  -- redraw flag
        or redrawFlag then                          -- set to redraw
        
            -- redraw the whole list
            mle = draw_mle_items( mle )
                                 
            -- reset the values
            oldFirstChar = mle[FIRST_CHAR]
            oldFirstLine = mle[TOP_LINE]
        
        else
        
            -- just redraw the current line
            draw_mle_item( mle, mle[ITEM] )
        
        end if

        -- check for growth
        if count != length( mle[ITEMS] ) then
            if  count-1 < mle[DEEP]
            and length( mle[ITEMS] ) > mle[DEEP] then

                -- redraw scroll bar
                mle[REDRAW] = 1

            elsif  count-1 > mle[DEEP]
            and length( mle[ITEMS] ) < mle[DEEP] then

                -- remove scroll bar
                mle[REDRAW] = 1

            end if

        end if

    
        -- handle the elevator
        mle = draw_elevator( mle )  

        -- position cursor
        position( mle[ROW]+mle[ITEM]-1, mle[COL]+mle[CURSOR]-1-VGA_MODE )

    else

        -- redisplay
        mle = draw_sle( mle )

    end if


    -- return the SLE data to DATA
    if mle[TYPE] = SLE then
        mle[DATA] = mle[ITEMS][1]
    end if

    -- clear redraw, if set
    mle[REDRAW] = 0

    -- return results of edit
    return( mle )


end function


-----------------------------------------------------------------------------
global procedure draw_mle_frame( sequence mle )
   
   -- display an mle
   
   -- top of box

   -- alias mle

    if VGA_MODE then

        -- frame
        graphic_frame( mle[ROW]-1, mle[COL]-2, mle[WIDE]+1, mle[DEEP] )

    else

       position( mle[ROW]-1, mle[COL]-1 )
       put_colors( OF_DIALOG, "" & repeat( '', mle[WIDE] ) & "" )


       for i = 1 to mle[DEEP] do
    
           -- display border
           position( mle[ROW]+i-1, mle[COL]-1 )
           put_colors( OF_DIALOG, "" )
           put_colors( OF_TEXT, repeat(   ' ', mle[WIDE] - VGA_MODE ) )
    
            -- right hand side

            if  length( mle[ITEMS] ) > mle[DEEP] then
                -- display scroll bar
                if    i = 1         then    put_colors( OF_SCROLL, {24} )
                elsif i = mle[DEEP]     then    put_colors( OF_SCROLL, {25} )
                else                        put_colors( OF_SCROLL, "" )
                end if
            else
                -- no scroll bar
                put_colors( OF_DIALOG, "" )
            end if

       end for
   
       -- bottom of box
       position( mle[ROW]+mle[DEEP], mle[COL]-1 )
       put_colors( OF_DIALOG, "" & repeat( '', mle[WIDE] ) & "" )

    end if

end procedure


-----------------------------------------------------------------------------
global function draw_mle( sequence mle )

        -- draw the mle frame and items

        draw_mle_frame( mle )
    mle = draw_mle_items( mle )

        return mle

end function

-----------------------------------------------------------------------------
global function focus_on_sle( sequence sle )
        -- select all the text in the sle
        -- used when the sle first gets focus

        -- alias sle    
        
        -- cursor is at start of text
        sle[FIRST_CHAR] = 1
        sle[CURSOR] = 1

        -- don't select if there is no text
        if length( sle[DATA] ) = 0 then
                sle[ANCHOR_COL] = 0
        else
                -- anchor is just beyond end
                sle[ANCHOR_COL] = length( sle[DATA] ) + 1
                sle[ANCHOR_LINE] = 1
        end if

        -- display sle
        sle = draw_sle( sle )

        -- return sle
        return sle

end function

-----------------------------------------------------------------------------
global function focus_off_sle( sequence sle )
        -- select all the text in the sle
        -- used when the sle first gets focus

        -- alias sle

        -- cursor is at start of text
        sle[FIRST_CHAR] = 1
        sle[CURSOR] = 1

        -- anchor cleared
        sle[ANCHOR_COL] = 0

        -- display sle
        sle = draw_sle( sle )

        -- return sle
        return sle

end function

